C********************************************************************
C     **** CHEAP ... CHEAP ROW ORDERING MINIMIZING NONZEROS OF L ****
C********************************************************************
C
      SUBROUTINE CHEAP(AATPAT,AATPNT,CLIQS,MAXNZL,MAXM,M,NZL,
     X PERM,INVP,DGHEAD,LINKFD,LINKBK,
     X RWLIST,LSTCLQ,MARKER,TEMP,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,M,NZL,IOLOG
      INTEGER*4 AATPNT(MAXM+1),CLIQS(MAXNZL)
      INTEGER*4 LSTCLQ(MAXM),MARKER(MAXM),TEMP(MAXM),RWLIST(MAXM)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 DGHEAD(MAXM),LINKFD(MAXM),LINKBK(MAXM)
      INTEGER*2 AATPAT(MAXNZL),PERM(MAXM),INVP(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IROW,K,ELROWS,DEGREE
      CHARACTER*100 BUFFER
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     AATPAT  Sparsity pattern of  A*Atransp handled as
C             a collection of sparse row vectors (diagonal
C             elements are excluded from the list).
C     AATPNT  Pointers to rows of  A*Atransp.
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the matrix to be decomposed.
C     M       Dimension of the matrix to be decomposed.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     NZL     Number of nonzero entries in Cholesky factor.
C     PERM    Permutation resulting from the minimum degree ordering.
C     INVP    Inverse permutation.
C
C     WORK ARRAYS:
C     CLIQS   Cliques of the pivotal rows (linked lists).
C     DGHEAD  Headers of the forward linked lists of rows (nodes)
C             with the same degree.
C     LINKFD  Forward linked lists of rows with the same degree.
C     LINKBK  Backward linked lists of rows with the same degree.
C     LSTCLQ  A list of headers to different pivotal cliques
C             that are still active i.e. that have not yet been
C             merged with any pivotal row.
C     RWLIST  A list of nonzero positions of a row that is
C             involved in a current step of elimination.
C     MARKER  Array used to mark already reordered rows.
C     TEMP    Temporary array used for merging lists.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT
C
C
C *** PURPOSE:
C     This routine implements a simple heuristic producing an
C     ordering of rows of A that is expected to reasonably minimize 
C     the number of nonzero entries in a Cholesky matrix.
C     It is significantly cheaper than a minimum degree ordering
C     (MDO) but in some cases it may produce considerably more
C     fill-in in the Cholesky factor.
C
C
C *** NOTES:
C     1. This routine assumes that the matrix  A*Atransp is
C        positive definite i.e. that pivoting in the numerical
C        phase will not be required.
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter 10.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  September 19, 1992
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (CHEAP) ***
C
C
C
C     Zero headers to the linked lists of rows
C     with the same degree.
      DO 20 IROW=1,M
         DGHEAD(IROW)=0
   20 CONTINUE
C
C     Set the linked lists of rows with the same degree (recall
C     that diagonal elements are not stored in the sparsity pattern).
      DO 40 IROW=1,M
         DEGREE=AATPNT(IROW+1)-AATPNT(IROW)+1
         LINKFD(IROW)=DGHEAD(DEGREE)
         DGHEAD(DEGREE)=IROW
   40 CONTINUE
C
C *** DEBUGGING
C     DO 42 IROW=1,M
C     WRITE(BUFFER,41) IROW,DGHEAD(IROW),LINKFD(IROW)
C  41 FORMAT(1X,'CHEAP: row',I6,'  header=',I6,'  linkfd=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C  42 CONTINUE
C
C
C
C     Scan linked lists of rows in order of increasing number
C     of nonzero entries.
C     ELROWS  is the number of already eliminated rows + 1.
      ELROWS=1
      DO 200 DEGREE=1,M
C
         IROW=DGHEAD(DEGREE)
  100    IF(IROW.EQ.0) GO TO 200
C        WRITE(BUFFER,101) IROW,DEGREE
C 101    FORMAT(1X,'CHEAP: row',I6,' has degree=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
C
C     Eliminate row IROW (save its position in a permuted matrix).
         INVP(IROW)=ELROWS
         ELROWS=ELROWS+1
         IROW=LINKFD(IROW)
         GO TO 100
C
  200 CONTINUE
C
C
C
C     The heuristic is completed.
C     Set the permutation vector.
      DO 300 I=1,M
         K=INVP(I)
         PERM(K)=I
  300 CONTINUE
C
C
C
      RETURN
C
C
C
C *** LAST CARD OF (CHEAP) ***
      END
