C*******************************************************************
C     **** DEFAAT ... DEFINE SPARSITY PATTERN OF  A*Atransp ****
C*******************************************************************
C
      SUBROUTINE DEFAAT(AATPAT,AATPNT,ITEMP0,
     X MAXNZL,MAXM,MAXN,MAXNZA,M,TRIANG,
     X MARKER,TEMP,STAVAR,
     X CLPNTS,RWNMBS,RWHEAD,RWLINK,CLNMBS,LENCOL,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,MAXN,MAXNZA,M,TRIANG,IOLOG
      INTEGER*4 AATPNT(MAXM+1),ITEMP0(MAXNZL)
      INTEGER*4 MARKER(MAXM),TEMP(MAXM)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 STAVAR(MAXN),AATPAT(MAXNZL)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROW,IR,K,KBEG,KEND,JCOL,J,IPOS
      INTEGER*4 LENAAT,LENROW
      REAL A1,A2
      CHARACTER*100 BUFFER
C
C
C *** COMMON ARREAS
C     Markers for linking rows.
C     COMMON /ICGRAD/ MSPLIT(100000)
C     INTEGER*2       MSPLIT
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     TRIANG  Indicator of how much of  A*Atransp sparsity pattern
C             is required:
C             0  if square matrix (except its diagonal) is needed;
C             1  if only strictly upper triangle is needed.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicate the position of the original variable.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     AATPAT  Sparsity pattern of  A*Atransp handled as
C             a collection of sparse row vectors (diagonal
C             elements are excluded from the list).
C     AATPNT  Pointers to rows of  A*Atransp.
C
C     WORK ARRAYS:
C     ITEMP0  Array used to handle unordered sparsity pattern
C             of  A*Atransp.
C     MARKER  Array used to mark the rows of  A that are 
C             adjacent to a given row.
C     TEMP    Temporary array. At the beginning it is used
C             to handle sparsity structure of rows. Later
C             it is used to control the reordering within rows.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT,MYSTOP,DTSORT
C
C
C *** PURPOSE:
C     This routine sets up the sparsity pattern of  A*Atransp
C     that is later used by the minimum degree routine  MDO.
C
C
C *** NOTES:
C     1. Two different rows  i and  j are said to be adjacent
C        if there exists a column in which they both have
C        a nonzero entry. The sparsity pattern array  AATPAT
C        contains then an entry  j in row  i (and, by symmetry,
C        an entry  i in row  j, if a square matrix is built).
C     2. Having created the sparsity structure of  A*Atransp
C        by rows in ITEMP0 array, we additionally scan the matrix
C        to obtain column increasing order within each row
C        (see e.g. Duff et al. (1989), section 2.10).
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapters  2 and  10.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  March 19, 1991
C     Last modified: June 9, 1995
C
C
C
C *** BODY OF (DEFAAT) ***
C
C
C
C
C     Initialize for building the sparsity pattern of  A*Atransp.
C     Zero MARKER array (it will indicate rows adjacent
C     to a given one).
      DO 100 IROW=1,M
         MARKER(IROW)=0
  100 CONTINUE
C
C
C     Set the parameters controlling the progress
C     of building  AATPAT array.
C     LENAAT  is the current length of  AATPAT array.
C     LENROW  is the length of a given row of  AATPAT array.
      LENAAT=0
      LENROW=0
C
C
C
C
C
C     Main loop begins here (loop over rows of  A).
C     For every row  IROW, a list of other rows adjacent
C     to a given one is created. Row IROW itself is omitted.
      DO 500 IROW=1,M
         LENROW=0
         AATPNT(IROW)=LENAAT+1
         MARKER(IROW)=1
C
C
C     Scan row IROW. Every column that intersects it, indicates
C     rows adjacent to a given one. Every row that appears
C     for the first time may then be added to the temporary
C     list of rows adjacent to IROW.
         J=RWHEAD(IROW)
  200    IF(J.EQ.0) GO TO 350
         JCOL=CLNMBS(J)
C
C     Omit fixed columns.
C        IF(STAVAR(JCOL).EQ.6) GO TO 300
         KBEG=CLPNTS(JCOL)
         KEND=KBEG+LENCOL(JCOL)-1
         DO 250 K=KBEG,KEND
            IR=RWNMBS(K)
            IF(MARKER(IR).EQ.1) GO TO 250
C
C     Omit adjacencies of linking and structural rows.
C           IF(MSPLIT(IR)+MSPLIT(IROW).GE.1) GO TO 250
C
C     Here if the row appears for the first time.
C     Add it to the adjacency list and mark the adjacent row.
            LENROW=LENROW+1
            TEMP(LENROW)=IR
            MARKER(IR)=1
  250    CONTINUE
  300    J=RWLINK(J)
         GO TO 200
C
C     Copy the adjacency list to ITEMP0 array.
C     Restore zero value of MARKER array.
  350    IF(LENROW.EQ.0) GO TO 450
         IF(LENAAT+LENROW.GT.MAXNZL) GO TO 9000
         DO 400 K=1,LENROW
            LENAAT=LENAAT+1
            IR=TEMP(K)
            ITEMP0(LENAAT)=IR
            MARKER(IR)=0
  400    CONTINUE
C
C     Decide how much of the sparsity pattern of  A*Atransp
C     is required. If only upper triangle of it is to be built,
C     then all rows that have already been scanned should
C     be excluded from further search.
  450    IF(TRIANG.EQ.0) MARKER(IROW)=0
C
C
C
C
C
C     End of main loop.
  500 CONTINUE
      AATPNT(M+1)=LENAAT+1
C
C
C *** DEBUGGING
C     DO 510 IR=1,M
C     KBEG=AATPNT(IR)
C     KEND=AATPNT(IR+1)-1
C     IF(KBEG.GT.KEND) GO TO 510 
C     WRITE(IOLOG,505) IR,(ITEMP0(K),K=KBEG,KEND)
C 505 FORMAT(1X,' List of rows adjacent to row: ',I6/(1X,10I6))
C 510 CONTINUE
C
C
C
C
C     Sort the sparsity pattern of each row of  AATPAT array
C     with increasing order of column numbers.
C     Decide what type of sort have to be done.
      IF(TRIANG.EQ.1) GO TO 700
C
C
C     Here if square matrix is to be built.
C     Set TEMP array to just after where each row
C     ends in a collection of rows.
      DO 550 IR=1,M
         TEMP(IR)=AATPNT(IR+1)
  550 CONTINUE
      DO 650 IROW=M,1,-1
         KBEG=AATPNT(IROW)
         KEND=AATPNT(IROW+1)-1
         DO 600 K=KBEG,KEND
            IR=ITEMP0(K)
            IPOS=TEMP(IR)-1
            TEMP(IR)=IPOS
            AATPAT(IPOS)=IROW
  600    CONTINUE
  650 CONTINUE
      GO TO 1000
C
C
C     Here if triangular matrix is to be built.
C     Go perform a double transpose sort.
  700 DO 800 K=1,LENAAT
         AATPAT(K)=ITEMP0(K)
  800 CONTINUE
C
C     SUBROUTINE DTSORT(ROWNBS,COLPTS,
C    X ICLNBS,IRWPTS,MAXNZ,MAXM,M,IOLOG)
C
      CALL DTSORT(AATPAT,AATPNT,
     X ITEMP0(1),TEMP,MAXNZL,MAXM,M,IOLOG)
C
C
C
C
C     Write problem statistics.
 1000 K=LENAAT/2
      A1=LENAAT*100.0
      IF(TRIANG.EQ.1) THEN
         K=LENAAT
         A1=A1*2.0
      ENDIF
      A2=M*M-M
      IF(M.GT.1) THEN
         A1=A1/A2
      ELSE
         A1=0.0
      ENDIF
      WRITE(BUFFER,1001) K,A1
 1001 FORMAT(1X,'DEFAAT: A*Atransp has       ',I13,
     X  ' subdiagonal elts (density=',F5.1,'%).')
      CALL MYWRT(IOLOG,BUFFER)
C
C
C *** DEBUGGING
C     WRITE(IOLOG,902)
C 902 FORMAT(1x,'DEFAAT: Matrix after reordering within the rows.'/)
C     DO 904 IR=1,M
C     KBEG=AATPNT(IR)
C     KEND=AATPNT(IR+1)-1
C     IF(KBEG.GT.KEND) GO TO 904
C     WRITE(IOLOG,903) IR,(AATPAT(K),K=KBEG,KEND)
C 903 FORMAT(1X,' List of rows adjacent to row: ',I6/(1X,10I6))
C 904 CONTINUE
C
C
C
      RETURN
C
C
C     Here to write error message.
 9000 WRITE(BUFFER,9001) LENAAT+LENROW
 9001 FORMAT(1X,'DEFAAT ERROR: A*Atransp overflow',I10)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9002) MAXNZL
 9002 FORMAT(1X,'    space was provided for only ',I10,' nonzeros.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(7)
C
C
C
C *** LAST CARD OF (DEFAAT) ***
      END
