C*****************************************************************
C     *** DETSPL ... DETERMINE PREFERABLE LENGTH OF DENSE COLS ***
C*****************************************************************
C
      SUBROUTINE DETSPL(IOLOG,
     X MAXM,MAXN,M,N,NSTRCT,
     X MAXCOL,LENCOL,STAVAR)
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MAXM,MAXN,M,N,NSTRCT,MAXCOL
      INTEGER*2 LENCOL(MAXN),STAVAR(MAXN)
C
C *** LOCAL VARIABLES
      INTEGER*4 J,K,NCOL,NONZ,MAXLEN
      CHARACTER*100 BUFFER
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     MAXCOL  A suggested length of dense columns after splitting.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C
C *** ON OUTPUT:
C     MAXCOL  Preferable length of dense columns after splitting.
C
C
C *** PURPOSE
C     This routine determines the "reasonable" length of dense columns
C     after splitting heuristic.
C
C *** SUBROUTINES CALLED
C
C *** NOTES
C
C *** REFERENCES:
C     Gondzio J. (1992). Splitting dense columns of the constraint
C        matrix in interior point methods for large scale linear
C        programming, Optimization 24, pp. 285-297.
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (DETSPL) ***
C
C     Determine the number of columns longer than MAXCOL, 
C     the number of nonzero entries in them 
C     and the column with tha maximum length.
      NCOL=0
      NONZ=0
      MAXLEN=0
      DO 100 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 100
         IF(STAVAR(J).LT.0) THEN
            K=-STAVAR(J)
            IF(J.GE.K) GO TO 100
         ENDIF
         IF(LENCOL(J).GT.MAXLEN) MAXLEN=LENCOL(J)
         IF(LENCOL(J).LT.MAXCOL) GO TO 100
         NCOL=NCOL+1
         NONZ=NONZ+LENCOL(J)
  100 CONTINUE
C
C     Determine the number of entries in split dense columns.
      IF(NCOL.EQ.0) GO TO 200
C
      MAXCOL=80
      IF(MAXLEN.GE.200) MAXCOL=MAXLEN/2+1
      IF(MAXLEN.GE.300) MAXCOL=MAXLEN/3+1
      IF(MAXLEN.GE.400) MAXCOL=MAXLEN/4+1
      IF(MAXLEN.GE.500) MAXCOL=MAXLEN/5+1
      IF(MAXLEN.GE.1000) MAXCOL=200
C
  200 CONTINUE
C     WRITE(BUFFER,201) NCOL,MAXLEN,MAXCOL
C 201 FORMAT(1X,'DETSPL: NCOL=',I6,'  MAXLEN=',I6,'  MAXCOL=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C     CALL MYWRT(0,BUFFER)
C     CALL MYWRT(99,BUFFER)
      RETURN
C
C *** LAST CARD OF (DETSPL) ***
      END
