C*****************************************************
C     ****   DTSORT ... DUOBLE TRANSPOSE SORT   ****
C*****************************************************
C
      SUBROUTINE DTSORT(ROWNBS,COLPTS,
     X ICLNBS,IRWPTS,MAXNZ,MAXM,M,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZ,MAXM,M,IOLOG
      INTEGER*4 COLPTS(MAXM+1),IRWPTS(MAXM+1)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 ROWNBS(MAXNZ),ICLNBS(MAXNZ)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROW,JCOL,K,KBEG,KEND,IPOS,NONZ
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     ROWNBS  Row numbers of nonzeros in columns of the matrix
C             (unordered within columns).
C     COLPTS  Pointers to the beginning of columns of the matrix.
C     MAXNZ   Maximum number of nonzeros of the matrix.
C     MAXM    Maximum dimension of the matrix.
C     M       Dimension of the matrix.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     ROWNBS  Row numbers of nonzeros in columns of the matrix
C             (with an increasing order of row numbers).
C     COLPTS  Pointers to the beginning of columns of the matrix.
C
C     WORK ARRAYS:
C     ICLNBS  Column numbers of nonzeros in rows of the matrix
C     IRWPTS  Pointers to the beginning of rows of the matrix.
C
C
C *** SUBROUTINES CALLED:
C     NONE
C
C
C *** PURPOSE:
C     This routine implements a double transpose sort of a given
C     matrix handled as a collection of sparse columns to obtain
C     the increasing order of row numbers within each column.
C
C
C *** NOTES:
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter 2.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  March 26, 1991
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (DTSORT) ***
C
C
C
C     Determine the number of matrix nonzero entries.
      NONZ=COLPTS(M+1)-1
C
C
C
C
C     *** First transposition.
C     Transpose the matrix to obtain the collection of sparse rows.
C     Count the row lengths.
      DO 700 IROW=1,M
         IRWPTS(IROW)=0
  700 CONTINUE
      DO 740 K=1,NONZ
         IROW=ROWNBS(K)
         IRWPTS(IROW)=IRWPTS(IROW)+1
  740 CONTINUE
C
C     Set IRWPTS array to indicate the positions just after every row.
      IRWPTS(1)=IRWPTS(1)+1
      DO 760 IROW=2,M
         IRWPTS(IROW)=IRWPTS(IROW)+IRWPTS(IROW-1)
  760 CONTINUE
      IRWPTS(M+1)=IRWPTS(M)
C
C     Move the matrix to a form of a collection of sparse rows.
      DO 800 JCOL=1,M
         KBEG=COLPTS(JCOL)
         KEND=COLPTS(JCOL+1)-1
         DO 780 K=KBEG,KEND
            IROW=ROWNBS(K)
            IPOS=IRWPTS(IROW)-1
            IRWPTS(IROW)=IPOS
            ICLNBS(IPOS)=JCOL
  780    CONTINUE
  800 CONTINUE
C
C
C
C
C     *** Second transposition.
C     Transpose the matrix back to the collection of sparse columns.
C     Set COLPTS array to indicate the positions just after every column.
      DO 820 JCOL=1,M
         COLPTS(JCOL)=COLPTS(JCOL+1)
  820 CONTINUE
C
C     Move the matrix back to a form of a collection of sparse columns.
      DO 860 IROW=M,1,-1
         KBEG=IRWPTS(IROW)
         KEND=IRWPTS(IROW+1)-1
         DO 840 K=KBEG,KEND
            JCOL=ICLNBS(K)
            IPOS=COLPTS(JCOL)-1
            COLPTS(JCOL)=IPOS
            ROWNBS(IPOS)=IROW
  840    CONTINUE
  860 CONTINUE
C
C
C
C
      RETURN
C
C
C
C *** LAST CARD OF (DTSORT) ***
      END
