C************************************************
C     *** FDAGGR ... FIND AGGREGATE VARIABLES ***
C************************************************
C
      SUBROUTINE FDAGGR(IOLOG,MSGLEV,ITIGHT,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,RMTMP1,INTMP1,RNTMP1,RNTMP2,RNTMP3,
     X B,RANGES,C,LOBND,UPBND,BNDBIG,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X P,Q,PRLVAR,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X IAGGR,MARKER,LENROW)
C
C
C
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array that contains real  LP problem data.
C     IWORK   Integer work array that contains integer  LP problem data.
C     RMAP    Map of RWORK array.
C     IMAP    Map of IWORK array.
C
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,ITIGHT
      INTEGER*4 MAXM,MAXN,MAXNZA,M,N,NSTRCT
      INTEGER*4 LNHIST,MXHIST,IROW(MAXN),IMTMP1(MAXM),INTMP1(MAXN)
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
      DOUBLE PRECISION RELT(MAXN),RMTMP1(MAXM)
      DOUBLE PRECISION RNTMP1(MAXN),RNTMP2(MAXN),RNTMP3(MAXN)
      DOUBLE PRECISION ACOEFF(MAXNZA),B(MAXM),RANGES(MAXM)
      DOUBLE PRECISION C(MAXN),LOBND(MAXN),UPBND(MAXN),BNDBIG
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      DOUBLE PRECISION P(MAXM),Q(MAXM),PRLVAR(MAXN)
      CHARACTER*8 RWNAME(MAXM),CLNAME(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      INTEGER*2 IAGGR(MAXN),LENROW(MAXM),MARKER(MAXM)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 NAGGR,NELIM,NEL0,NFIXED,NFREE,NIDEN,NTIGHT,NT0,N1
      INTEGER*4 I,IHEAD,IKX,IPOS,IR,IRUN,J,J1,JCOL,KSTAT,KROW
      INTEGER*4 K,K1,KBEG,KEND,K2,K2BEG,K2END,KOK,KOUT,KRWBEG
      INTEGER*4 LROW,LIMIT,MNEW,NNEG,NPOS,NNEGBG,NPOSBG,KNEGBG,KPOSBG
      DOUBLE PRECISION ALPHA,BIG,BIGNEW,BNDNEW,BNDJLO,BNDJUP,DP,X0
      DOUBLE PRECISION RNRM,RHS0,BLOWER,BUPPER,FSBTOL,BNDTOL,SMALLA
      CHARACTER*100 BUFFER
      CHARACTER*2   RTYPE
C
C *** LOCAL VARIABLES USED BY THE ONE-ROW SIMPLEX
      INTEGER*4 NMAX,NX,NXSTR,ROWST
      DOUBLE PRECISION COEFF(100),X(100),COBJ(100),UPPER(100)
      DOUBLE PRECISION RDCOST(100),RHS,PI,QI,DUAL
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     ITIGHT  A control of tightening UPPER bounds.
C             1  incomplete analysis (no tightening UPPER bounds);
C             2  maximum analysis possible.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     LNHIST  Length of the PRE_SOLVE history list.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     RANGES  Array of constraint ranges.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     LOBND   Array of lower bounds.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     BNDBIG  Value of an unacceptably large implicit bound.
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     PRLVAR  Primal variables of the linear program.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWNAME  Array of row names (increasing order sort).
C     CLNAME  Array of column names (unordered).
C     LENROW  Lengths of (sparse) rows of matrix A.
C
C *** ON OUTPUT:
C
C
C
C
C *** WORK ARRAYS:
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     IMTMP1  Integer work array of size MAXM.
C     INTMP1  Positions of pivot elements in aggregate columns.
C     RMTMP1  Nonzero elements of the analysed column.
C     RNTMP1  LOWER bounds for aggregate variables.
C     RNTMP2  UPPER bounds for aggregate variables.
C     RNTMP3  Linear combination of an aggregate variable.
C     IAGGR   Linked lists of aggregate variables.
C     MARKER  Marker for rows that define aggregate variables
C             (such rows cannot be removed).
C
C
C
C
C *** PURPOSE
C     This routine finds variables that have identical structure.
C     It then builds an aggregate column and checks if such a column
C     can be eliminated.
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,MYSTOP,DABS,EMPTYR,REORDA,REORDI,REORDV,SMPLX
C
C
C *** NOTES:
C
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  January 29, 1993
C     Last modified: June 9, 1995
C
C
C
C
C *** BODY OF (FDAGGR) ***
C
C
C
C     Initialize.
      BIG=1.0D+30
      BIGNEW=1.0D+20
      FSBTOL=5.0D-8
      BNDTOL=1.0D-5
      SMALLA=1.0D-8
      NELIM=0
      NFIXED=0
      NTIGHT=0
C
C
C
C
C     Zero  MARKER, LENROW and RMTMP1 arrays.
      DO 100 I=1,M
         MARKER(I)=0
         LENROW(I)=0
         RMTMP1(I)=0.0D0
  100 CONTINUE
C
C     Zero  IAGGR and  INTMP1 arrays.
C     Count nonzero elements in all rows of  A.
C     Count FREE and still active structural variables.
C     Define LOWER and UPPER bounds for aggregate columns.
      NFREE=0
      N1=0
      DO 160 J=1,NSTRCT
         IAGGR(J)=0
         INTMP1(J)=0
         KSTAT=STAVAR(J)
         IF(KSTAT.GE.6) GO TO 160
         N1=N1+1
         IF(KSTAT.LT.0) NFREE=NFREE+1
         RNTMP1(J)=0.0D0
         RNTMP2(J)=BIG
         IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) RNTMP2(J)=UPBND(J)
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 140 K=KBEG,KEND
            IR=RWNMBS(K)
            LENROW(IR)=LENROW(IR)+1
  140    CONTINUE
  160 CONTINUE
      NFREE=NFREE/2
C
C
C
C
C
C
C     First main loop begins here.
C     Loop over all (active) structural columns of  A.
C
C     Build linked lists of aggregate columns.
C     Define LOWER and UPPER bounds for aggregate columns.
C     Count AGGREGATE variables.
      NAGGR=0
      DO 1000 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 1000
         IF(LENCOL(J).EQ.0) GO TO 1000
         IF(IAGGR(J).GT.0) GO TO 1000
C
C
C     Save nonzero elements of column J in RMTMP1 array.
C     Determine the shortest row with an entry in column J.
C     Equality-type rows are prefered if ties are to be broken.
         KROW=0
         LROW=NSTRCT+1
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 380 K=KBEG,KEND
            IR=RWNMBS(K)
            RMTMP1(IR)=ACOEFF(K)
            IF(LENROW(IR)-LROW) 360,340,380
  340       IF(RWSTAT(IR).NE.1) GO TO 380
  360       LROW=LENROW(IR)
            KROW=IR
            KOK=K
  380    CONTINUE
C
C     Save position of the pivot element in INTMP1 array.
         INTMP1(J)=KOK
         IF(KROW.EQ.0) GO TO 940
C
C     Analyse all columns that have entries in row KROW.
C     Look for a column with identical sparsity structure as column J.
         NIDEN=1
         IHEAD=J
         BLOWER=RNTMP1(J)
         BUPPER=RNTMP2(J)
         IPOS=RWHEAD(KROW)
         IF(RWSTAT(KROW).GE.2) IPOS=RWLINK(IPOS)
  400    IF(IPOS.EQ.0) GO TO 810
            JCOL=CLNMBS(IPOS)
            IF(LENCOL(JCOL).NE.LENCOL(J)) GO TO 800
            IF(STAVAR(JCOL).GE.6) GO TO 800
            IF(IAGGR(J).GT.0) GO TO 800
            IF(JCOL.LE.J) GO TO 800
C
C
C     Here if two columns J and JCOL have the same length.
            K2BEG=CLPNTS(JCOL)
            K2END=K2BEG+LENCOL(JCOL)-1
C
C
C
C     Check if columns J and JCOL are linearly dependent.
            IR=RWNMBS(K2BEG)
            ALPHA=RMTMP1(IR)/ACOEFF(K2BEG)
            DO 500 K2=K2BEG+1,K2END
               IR=RWNMBS(K2)
               DP=DABS(RMTMP1(IR)/ACOEFF(K2)-ALPHA)
               IF(DP.GE.SMALLA) GO TO 800
  500       CONTINUE
C
C
C     Here if two columns J and JCOL are linearly dependent.
            IF(MSGLEV.LE.2) GO TO 510
            WRITE(BUFFER,501) CLNAME(J),CLNAME(JCOL)
  501       FORMAT(1X,'FDAGGR: LP variables:   ',
     X       A8,' and ',A8,' are linearly dependent.')
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,502) J,C(J),STAVAR(J),LENCOL(J)
  502       FORMAT(1X,'    var=',I6,' Cj=',D14.8,' st=',I6,' ln=',I6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,503) JCOL,C(JCOL),STAVAR(JCOL),LENCOL(JCOL)
  503       FORMAT(1X,'and var=',I6,' Cj=',D14.8,' st=',I6,' ln=',I6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,504) J,JCOL,ALPHA
  504       FORMAT(1X,'FDAGGR: J=',I6,' JCOL=',I6,' ALPHA=',D14.8)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,505) KROW,RWSTAT(KROW),LROW
  505       FORMAT(1X,'        row=',I5,'  rwstat=',I5,'  len=',I5)
            CALL MYWRT(IOLOG,BUFFER)
  510       CONTINUE
C
C
C     Save information on the aggregate column.
C     Mark row that defines aggregate variable.
C     Add column JCOL to the linked list.
C     Update LOWER and UPPER bounds of the aggregate variable.
C     Save the linear dependency of columns in RNTMP3 array.
            NIDEN=NIDEN+1
            RNTMP3(JCOL)=1.0D0/ALPHA
            IAGGR(JCOL)=IHEAD
            IHEAD=JCOL
            KSTAT=STAVAR(JCOL)
            BNDJUP=BIG
            IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) BNDJUP=UPBND(JCOL)
            DP=ACOEFF(IPOS)/ACOEFF(KOK)
            IF(DP.LT.0.0D0) THEN
               BLOWER=BLOWER+BNDJUP*DP
            ELSE
               BUPPER=BUPPER+BNDJUP*DP
            ENDIF
C
C
C
  800    IPOS=RWLINK(IPOS)
         GO TO 400
C
C
C     Save LOWER and UPPER bounds of the aggregate variable.
C     Close the linked list (to allow wrap around).
  810    IF(NIDEN.GE.2) THEN 
            NAGGR=NAGGR+1
            IAGGR(J)=-IHEAD
            RNTMP3(J)=1.0D0
            MARKER(KROW)=MARKER(KROW)+1
         ENDIF
         RNTMP1(J)=BLOWER
         RNTMP2(J)=BUPPER
C
C
C     Restore zero value of  RMTMP1 array.
  940    DO 960 K=KBEG,KEND
            IR=RWNMBS(K)
            RMTMP1(IR)=0.0D0
  960    CONTINUE
C
C
C *** Debugging.
         IF(NIDEN.LE.1) GO TO 970
         IF(MSGLEV.LE.2) GO TO 970
         WRITE(BUFFER,961) J,LENCOL(J),NIDEN
  961    FORMAT(1X,'FDAGGR: Aggr. var=',I6,' ln=',I6,' NIDEN=',I6)
         CALL MYWRT(IOLOG,BUFFER)
C        WRITE(BUFFER,962) P(KROW),Q(KROW)
C 962    FORMAT(1X,'        P=',D14.8,'  Q=',D14.8)
C        CALL MYWRT(IOLOG,BUFFER)
C        WRITE(BUFFER,963) BLOWER,BUPPER
C 963    FORMAT(1X,'        LO bnd=',D14.8,'  UP bnd=',D14.8)
C        CALL MYWRT(IOLOG,BUFFER)
  970    CONTINUE
C
C
C
C
C
C
C     End of the first main loop.
 1000 CONTINUE
C
C     Check if RMTMP1 array is zero.
      DO 1010 I=1,M
         IF(DABS(RMTMP1(I)).GE.SMALLA) THEN
            WRITE(BUFFER,1011) I,RMTMP1(I)
 1011       FORMAT(1X,'FDAGGR ERROR: RMTMP1(',I6,')=',D14.8)
            CALL ERRWRT(IOLOG,BUFFER)
            CALL MYSTOP(95)
         ENDIF
 1010 CONTINUE
C
C
C     Check if there exist aggregate variables.
      IF(NAGGR.EQ.0) GO TO 6000
C
C
C
C
C
C
C     Compute norms of the  LP constraints.
      DO 1050 I=1,M
         RMTMP1(I)=1.0D-4+DABS(B(I))
 1050 CONTINUE
      DO 1080 J=1,NSTRCT
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 1060 K=KBEG,KEND
            I=RWNMBS(K)
            IF(DABS(ACOEFF(K)).GT.RMTMP1(I)) RMTMP1(I)=DABS(ACOEFF(K))
 1060    CONTINUE
 1080 CONTINUE
      DO 1090 I=1,M
         RNRM=SMALLA*RMTMP1(I)
         IF(RNRM.LE.FSBTOL) RNRM=FSBTOL
         IF(RNRM.GE.1.0D-6) RNRM=1.0D-6
         IF(DABS(B(I)).LE.RNRM) B(I)=0.0D0
C        IF(RMTMP1(I).LE.1.0E-15) GO TO 1090
C        IF(RMTMP1(I).LE.1.0E-2) THEN
C           WRITE(BUFFER,1091) I,RMTMP1(I)
C1091       FORMAT(1X,'   FDAGGR: row=',I6,' has norm=',D10.3)
C           CALL MYWRT(IOLOG,BUFFER)
C        ENDIF
 1090 CONTINUE
C
C
C
C
C
C
C     Second main loop begins here.
C     Loop over all (yet uneliminated) LP constraints.
C     Eliminate redundant constraints.
      IRUN=1
 1200 NT0=NTIGHT
      NEL0=NELIM
      DO 2000 I=1,M
         IF(RWHEAD(I).LE.0) GO TO 2000
         KRWBEG=RWHEAD(I)
         IPOS=KRWBEG
         IF(RWSTAT(I).GE.2) IPOS=RWLINK(KRWBEG)
C
C     Compute LOWER and UPPER limits of the LP constraint.
C     Loop over nonzero entries of row I.
         BLOWER=0.0D0
         BUPPER=0.0D0
 1300    IF(IPOS.EQ.0) GO TO 1400
            J=CLNMBS(IPOS)
            K=STAVAR(J)
            IF(K.GE.6) GO TO 1360
            IF(IAGGR(J)) 1340,1340,1360
 1340       BNDJLO=RNTMP1(J)
            BNDJUP=RNTMP2(J)
            IF(ACOEFF(IPOS).LT.0.0D0) THEN
               BLOWER=BLOWER+BNDJUP*ACOEFF(IPOS)
               BUPPER=BUPPER+BNDJLO*ACOEFF(IPOS)
            ELSE
               BLOWER=BLOWER+BNDJLO*ACOEFF(IPOS)
               BUPPER=BUPPER+BNDJUP*ACOEFF(IPOS)
            ENDIF
 1360    IPOS=RWLINK(IPOS)
         GO TO 1300
C
 1400    CONTINUE
         RNRM=SMALLA*(RMTMP1(I)+DABS(B(I)))
         IF(RNRM.LE.FSBTOL) RNRM=FSBTOL
         IF(RNRM.GE.1.0D-6) RNRM=1.0D-6
         IF(DABS(BLOWER-B(I)).LE.RNRM) BLOWER=B(I)
         IF(DABS(BUPPER-B(I)).LE.RNRM) BUPPER=B(I)
C        RTYPE='EQ'
C        IF(RWSTAT(I).EQ.2) RTYPE='GE'
C        IF(RWSTAT(I).EQ.3) RTYPE='LE'
C        WRITE(BUFFER,1401) I,RTYPE,BLOWER,BUPPER,B(I)
C1401    FORMAT(1X,'row=',I6,' type=',A2,
C    X    ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
C        CALL MYWRT(IOLOG,BUFFER)
C
C
C
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY type constraint.
            RTYPE='EQ'
            IF(BLOWER-B(I).GT.-FSBTOL) THEN
               IF(BLOWER-B(I).GT.FSBTOL) GO TO 9010
               LIMIT=0
               GO TO 1500
            ENDIF
            IF(BUPPER-B(I).LT.FSBTOL) THEN
               IF(BUPPER-B(I).LT.-FSBTOL) GO TO 9010
               LIMIT=1
               GO TO 1500
            ENDIF
            GO TO 2000
         ENDIF
C
C
C
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
            RTYPE='GE'
            IF(BUPPER-B(I).LT.FSBTOL) THEN
               IF(BUPPER-B(I).LT.-FSBTOL) GO TO 9010
               LIMIT=1
               GO TO 1500
            ENDIF
            IF(BLOWER-B(I).GT.-FSBTOL) THEN
               LIMIT=-1
               GO TO 1500
            ENDIF
         ENDIF
C
C
C
         IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
            RTYPE='LE'
            IF(BLOWER-B(I).GT.-FSBTOL) THEN
               IF(BLOWER-B(I).GT.FSBTOL) GO TO 9010
               LIMIT=0
               GO TO 1500
            ENDIF
            IF(BUPPER-B(I).LT.FSBTOL) THEN
               LIMIT=-1
               GO TO 1500
            ENDIF
         ENDIF
C
C
C
         GO TO 2000
C
C
C
C     Here to eliminate the LP constraint.
C     Do not eliminate the row if it defines an aggregate.
 1500    IF(MARKER(I).GT.0) THEN
C           WRITE(BUFFER,1501) I,MARKER(I)
C1501       FORMAT(1X,'FDAGGR: Row      ',I6,
C    X       ' cannot be eliminated, MARKER=',I3)
C           CALL MYWRT(IOLOG,BUFFER)
C           WRITE(BUFFER,1502) I,RTYPE,BLOWER,BUPPER,B(I)
C1502       FORMAT(1X,'row=',I6,' type=',A2,
C    X       ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
C           CALL MYWRT(IOLOG,BUFFER)
            IF(LIMIT.EQ.-1) GO TO 2000
            GO TO 1580
         ENDIF
         NELIM=NELIM+1
         RWHEAD(I)=-RWHEAD(I)
C
C *** DEBUGGING
         IF(MSGLEV.LE.1) GO TO 1504
         WRITE(BUFFER,1503) I,RWNAME(I),RTYPE
 1503    FORMAT(1X,'FDAGGR: Row      ',I6,' (name=',A8,
     X    ' type=',A2,') is eliminated.')
         CALL MYWRT(IOLOG,BUFFER)
 1504    CONTINUE
         IF(MSGLEV.LE.2) GO TO 1510
         WRITE(BUFFER,1505) I,RTYPE,BLOWER,BUPPER,B(I)
 1505    FORMAT(1X,'row=',I6,' type=',A2,
     X    ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
 1510    CONTINUE
C
         IF(LIMIT.EQ.-1) THEN
            NFIXED=NFIXED+1
            J=CLNMBS(KRWBEG)
            PRLVAR(J)=0.0D0
            STAVAR(J)=14
            GO TO 2000
         ENDIF
C
C     Here to eliminate the constraint and fix variables.
C     Loop over nonzero entries of row I.
 1580    IPOS=KRWBEG
         IF(RWSTAT(I).GE.2) IPOS=RWLINK(IPOS)
 1600    IF(IPOS.EQ.0) GO TO 1800
            J=CLNMBS(IPOS)
            IF(STAVAR(J).GE.6) GO TO 1750
            IF(IAGGR(J)) 1640,1680,1750
C
C     Here for an aggregate variable.
 1640       IF(ACOEFF(IPOS).LT.0.0D0) THEN
               IF(LIMIT.EQ.0) RNTMP1(J)=RNTMP2(J)
               IF(LIMIT.EQ.1) RNTMP2(J)=RNTMP1(J)
            ELSE
               IF(LIMIT.EQ.0) RNTMP2(J)=RNTMP1(J)
               IF(LIMIT.EQ.1) RNTMP1(J)=RNTMP2(J)
            ENDIF
            IF(MSGLEV.LE.1) GO TO 1642
            WRITE(BUFFER,1641) J,RNTMP1(J)
 1641       FORMAT(1X,'FDAGGR: Variable ',I6,
     X       ' (aggregate)     is being FIXED on X=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
 1642       CONTINUE
            GO TO 1750
C
C     Here for a standard LP variable.
 1680       IF(ACOEFF(IPOS).LT.0.0D0) THEN
               IF(LIMIT.EQ.0) X0=UPBND(J)
               IF(LIMIT.EQ.1) X0=0.0D0
            ELSE
               IF(LIMIT.EQ.0) X0=0.0D0
               IF(LIMIT.EQ.1) X0=UPBND(J)
            ENDIF
C
C     Fix and eliminate column J. Omit already FIXED variables.
C     Update RHS array.
            NFIXED=NFIXED+1
            IF(DABS(X0).LE.FSBTOL) X0=0.0D0
            PRLVAR(J)=X0
            STAVAR(J)=6
            IF(DABS(X0).LE.FSBTOL) GO TO 1720
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            DO 1700 K=KBEG,KEND
               IR=RWNMBS(K)
               B(IR)=B(IR)-X0*ACOEFF(K)
               IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 1700       CONTINUE
 1720       CONTINUE
            IF(MSGLEV.LE.1) GO TO 1722
            WRITE(BUFFER,1721) J,CLNAME(J),X0
 1721       FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X       ') is being FIXED on X=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
 1722       CONTINUE
C
 1750       IPOS=RWLINK(IPOS)
            GO TO 1600
C
C
C     Remove slack variable of an inequality-type constraint.
 1800    IF(MARKER(I).GT.0) GO TO 2000
         IF(RWSTAT(I).EQ.1) GO TO 2000
         NFIXED=NFIXED+1
         J=CLNMBS(KRWBEG)
         PRLVAR(J)=0.0D0
         STAVAR(J)=14
C
C
C
C
C
C     End of the second main loop.
 2000 CONTINUE
C
C
C
C
C
C
C     Third main loop begins here.
C     Loop over all (yet uneliminated) LP constraints.
C     Tighten bounds on variables.
C     NNEG    Number of negative entries in a given row.
C     NPOS    Number of positive entries in a given row.
C     NNEGBG  Number of negative entries with an infinite UP bound
C             and positive entries with an infinite LO bound.
C     NPOSBG  Number of positive entries with an infinite UP bound.
C             and negative entries with an infinite LO bound.
      DO 3000 I=1,M
         IF(RWHEAD(I).LE.0) GO TO 3000
         RHS0=B(I)
         RTYPE='EQ'
         IF(RWSTAT(I).EQ.2) RTYPE='GE'
         IF(RWSTAT(I).EQ.3) RTYPE='LE'
C
C
C     Compute LOWER and UPPER limits of the LP constraint.
         KRWBEG=RWHEAD(I)
         IF(RWSTAT(I).GE.2) KRWBEG=RWLINK(KRWBEG)
         BLOWER=0.0D0
         BUPPER=0.0D0
         NPOS=0
         NNEG=0
         NPOSBG=0
         NNEGBG=0
C
C
C
C     Loop over nonzero entries of row I.
C     After this loop:
C     if  NNEGBG > 0,  then  BLOWER = - Inf,
C     if  NPOSBG > 0,  then  BUPPER = + Inf.
         IPOS=KRWBEG
 2100    IF(IPOS.EQ.0) GO TO 2200
            J=CLNMBS(IPOS)
            K=STAVAR(J)
            IF(K.GE.6) GO TO 2160
            IF(IAGGR(J)) 2120,2140,2160
C
C      Here for an aggregate variable:  RNTMP1(j) <= Xj <= RNTMP2(j).
 2120       BNDJLO=RNTMP1(J)
            BNDJUP=RNTMP2(J)
            IF(ACOEFF(IPOS).LT.0.0D0) THEN
               NNEG=NNEG+1
               IF(BNDJUP.GT.BIGNEW) THEN
                  NNEGBG=NNEGBG+1
                  KNEGBG=IPOS
               ELSE
                  BLOWER=BLOWER+BNDJUP*ACOEFF(IPOS)
               ENDIF
               IF(BNDJLO.LT.-BIGNEW) THEN
                  NPOSBG=NPOSBG+1
                  KPOSBG=IPOS
               ELSE
                  BUPPER=BUPPER+BNDJLO*ACOEFF(IPOS)
               ENDIF
            ELSE
               NPOS=NPOS+1
               IF(BNDJUP.GT.BIGNEW) THEN 
                  NPOSBG=NPOSBG+1
                  KPOSBG=IPOS
               ELSE
                  BUPPER=BUPPER+BNDJUP*ACOEFF(IPOS)
               ENDIF
               IF(BNDJLO.LT.-BIGNEW) THEN 
                  NNEGBG=NNEGBG+1
                  KNEGBG=IPOS
               ELSE
                  BLOWER=BLOWER+BNDJLO*ACOEFF(IPOS)
               ENDIF
            ENDIF
            GO TO 2160
C
C      Here for a standard LP variable:  0 <= Xj <= RNTMP2(j).
 2140       BNDJUP=RNTMP2(J)
            IF(ACOEFF(IPOS).LT.0.0D0) THEN
               NNEG=NNEG+1
               IF(BNDJUP.GT.BIGNEW) THEN
                  NNEGBG=NNEGBG+1
                  KNEGBG=IPOS
               ELSE
                  BLOWER=BLOWER+BNDJUP*ACOEFF(IPOS)
               ENDIF
            ELSE
               NPOS=NPOS+1
               IF(BNDJUP.GT.BIGNEW) THEN 
                  NPOSBG=NPOSBG+1
                  KPOSBG=IPOS
               ELSE
                  BUPPER=BUPPER+BNDJUP*ACOEFF(IPOS)
               ENDIF
            ENDIF
 2160       IPOS=RWLINK(IPOS)
            GO TO 2100
C
C
C
 2200    CONTINUE
         RNRM=SMALLA*(RMTMP1(I)+DABS(B(I)))
         IF(RNRM.LE.FSBTOL) RNRM=FSBTOL
         IF(RNRM.GE.1.0D-6) RNRM=1.0D-6
         IF(DABS(BLOWER-B(I)).LE.RNRM) BLOWER=B(I)
         IF(DABS(BUPPER-B(I)).LE.RNRM) BUPPER=B(I)
C
C *** DEBUGGING
         IF(MSGLEV.LE.2) GO TO 2210
         WRITE(BUFFER,2201) I,RTYPE,BLOWER,BUPPER,B(I)
 2201    FORMAT(1X,'Row=',I6,' type=',A2,
     X    ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,2202) NPOS,NNEG,NPOSBG,NNEGBG
 2202    FORMAT(1X,' Npos=',I5,' Nneg=',I5,
     X    '   Nposbg=',I5,' Nnegbg=',I5)
         CALL MYWRT(IOLOG,BUFFER)
 2210    CONTINUE
C
C
C
         IF(RWSTAT(I).EQ.1.OR.RWSTAT(I).EQ.3) THEN
C
C     Here for EQUALITY type or LESS OR EQUAL type constraint.
            IF(BLOWER-RHS0.GT.FSBTOL.AND.NNEGBG.EQ.0) GO TO 9020
            IF(NNEGBG.GE.1) GO TO 2400
C
C
C     Here if there are no negative entries with BIG Uj
C         and there are no positive entries with BIG Lj,
C         i.e. BLOWER is finite.
C
C     Loop over nonzero entries of row I.
            BLOWER=RHS0-BLOWER
            IPOS=KRWBEG
 2300       IF(IPOS.EQ.0) GO TO 2380
               J=CLNMBS(IPOS)
               K=STAVAR(J)
               IF(K.GE.6) GO TO 2360
               IF(IAGGR(J).GT.0) GO TO 2360
               IF(ACOEFF(IPOS).GT.0.0D0) THEN
C
C     Implicit UPPER bound can be defined for each variable
C     refering to POSITIVE entry of row I.
                  BNDJUP=RNTMP2(J)
                  BNDNEW=RNTMP1(J)+BLOWER/ACOEFF(IPOS)
                  IF(BNDNEW.GE.BNDJUP-BNDTOL) GO TO 2360
C                 IF(BNDNEW.GE.BNDBIG) GO TO 2360
C                 IF(BNDNEW.GE.BNDBIG) THEN 
C                    IF(IAGGR(J).EQ.0) GO TO 2360
C                 ENDIF
                  IF(BNDNEW.GE.BNDBIG) THEN 
                     IF((K.EQ.0.OR.K.EQ.2).AND.
     X                IAGGR(J).EQ.0) GO TO 2360
                  ENDIF
                  IF(ITIGHT.LE.1) GO TO 2360
                  NTIGHT=NTIGHT+1
                  RNTMP2(J)=BNDNEW
                  IF(IAGGR(J).EQ.0) THEN
                     UPBND(J)=BNDNEW
                     IF(K.NE.1.AND.K.NE.3) STAVAR(J)=STAVAR(J)+1
                  ENDIF
C
                  IF(MSGLEV.LE.1) GO TO 2304
C                 WRITE(BUFFER,2301) J,STAVAR(J),BNDJUP,BNDNEW
C2301             FORMAT(1X,'cl=',I6,' st=',I6,' oldUPj=',D16.8,
C    X             ' newUPj=',D16.8)
C                 CALL MYWRT(IOLOG,BUFFER)
                  IF(IAGGR(J).EQ.0) THEN
                     WRITE(BUFFER,2302) J,CLNAME(J),BNDNEW
 2302                FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X                ') has new UPPER bound=',D14.6)
                     CALL MYWRT(IOLOG,BUFFER)
                  ELSE
                     WRITE(BUFFER,2303) J,BNDNEW
 2303                FORMAT(1X,'FDAGGR: Variable ',I6,
     X                ' (aggregate)     has new UPPER bound=',D14.6)
                     CALL MYWRT(IOLOG,BUFFER)
                  ENDIF
 2304             CONTINUE
C
C     Reinitialize bounds on shadow prices.
                  KBEG=CLPNTS(J)
                  KEND=KBEG+LENCOL(J)-1
                  DO 2305 IKX=KBEG,KEND
                     IR=RWNMBS(IKX)
                     P(IR)=-BIG
                     Q(IR)=BIG
                     IF(RANGES(IR).LE.BIGNEW) GO TO 2305
                     IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                     IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
 2305             CONTINUE
C
                  IF(BNDNEW.LE.RNTMP1(J)+FSBTOL) THEN
C
C     Fix variable J on its LOWER bound.
                     IF(IAGGR(J)) 2306,2310,2360
C
C     Here for an aggregate variable.
 2306                RNTMP2(J)=RNTMP1(J)
                     IF(MSGLEV.LE.1) GO TO 2309
                     WRITE(BUFFER,2308) J,RNTMP1(J)
 2308                FORMAT(1X,'FDAGGR: Variable ',I6,
     X                ' (aggregate)     is being FIXED on X=',D14.6)
                     CALL MYWRT(IOLOG,BUFFER)
 2309                CONTINUE
                     GO TO 2360
C
C     Here for a standard LP variable.
 2310                NFIXED=NFIXED+1
                     X0=0.0D0
                     PRLVAR(J)=X0
                     STAVAR(J)=6
                     IF(MSGLEV.LE.1) GO TO 2312
                     WRITE(BUFFER,2311) J,CLNAME(J),X0
 2311                FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X                ') is being FIXED on X=',D14.6)
                     CALL MYWRT(IOLOG,BUFFER)
 2312                CONTINUE
                     GO TO 2360
                  ENDIF
C
               ELSE
C
C     Implicit LOWER bound can be defined for each variable
C     refering to NEGATIVE entry of row I.
C     Line added to break ties for optimization on SUN.
C     The following line has to be uncommented to help the 'F77' 
C     compiler of SUN SPARC computer to produce correct code 
C     for a -O2 -xcg92 options.
                  J=J+1-1
                  BNDJUP=RNTMP2(J)
                  BNDNEW=BNDJUP+BLOWER/ACOEFF(IPOS)
                  IF(BNDNEW.LE.RNTMP1(J)+BNDTOL) GO TO 2360
                  IF(IAGGR(J)) 2320,2330,2360
C
C     Here for an aggregate variable.
 2320             RNTMP1(J)=BNDNEW
                  IF(MSGLEV.LE.1) GO TO 2329
                  WRITE(BUFFER,2328) J,RNTMP1(J)
 2328             FORMAT(1X,'FDAGGR: Variable ',I6,
     X             ' (aggregate)     has new LOWER bound=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 2329             CONTINUE
                  GO TO 2360
C
C     Here for a standard LP variable.
 2330             NTIGHT=NTIGHT+1
                  LOBND(J)=LOBND(J)+BNDNEW
                  UPBND(J)=UPBND(J)-BNDNEW
                  RNTMP2(J)=UPBND(J)
                  STAVAR(J)=3
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
                  IF(LNHIST.GE.MXHIST) GO TO 9200
                  LNHIST=LNHIST+1
                  INHIST(LNHIST)=-J
                  DPHIST(LNHIST)=BNDNEW
C
C     Modify RHS (take account of the new LOWER bound on Xj).
                  KBEG=CLPNTS(J)
                  KEND=KBEG+LENCOL(J)-1
                  DO 2340 IKX=KBEG,KEND
                     IR=RWNMBS(IKX)
                     B(IR)=B(IR)-BNDNEW*ACOEFF(IKX)
                     IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 2340             CONTINUE
C
C *** DEBUGGING
                  IF(MSGLEV.LE.2) GO TO 2342
                  WRITE(BUFFER,2341) J,STAVAR(J),BNDNEW
 2341             FORMAT(1X,'cl=',I6,' st=',I6,' oldLOj=0.0D0',
     X             ' newLOj=',D16.8)
                  CALL MYWRT(IOLOG,BUFFER)
 2342             CONTINUE
                  IF(MSGLEV.LE.1) GO TO 2344
                  WRITE(BUFFER,2343) J,CLNAME(J),BNDNEW
 2343             FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X             ') has new LOWER bound=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 2344             CONTINUE
C
               ENDIF
 2360          IPOS=RWLINK(IPOS)
               GO TO 2300
C
 2380       CONTINUE
            BLOWER=RHS0-BLOWER
            GO TO 2500
C
C
C     Here if there exists at least one negative entry with BIG Uj
C          or there exists at least one positive entry with BIG Lj.
C     If only one such an entry exists, then its -big LOWER bound or
C     big UPPER bound can be improved. KNEGBG indicates its position.
 2400       IF(NNEGBG.GE.2) GO TO 2500
C
            J=CLNMBS(KNEGBG)
            K=STAVAR(J)
            IF(K.GE.6) GO TO 2500
            BNDNEW=(RHS0-BLOWER)/ACOEFF(KNEGBG)
            IF(IAGGR(J)) 2440,2410,2500
C
C     Here for a standard LP variable (it must have been negative
C     entry with big Uj). Variable's LOWER bound can be improved.
 2410       IF(BNDNEW.LE.BNDTOL) GO TO 2500
            NTIGHT=NTIGHT+1
            LOBND(J)=LOBND(J)+BNDNEW
            STAVAR(J)=2
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
            IF(LNHIST.GE.MXHIST) GO TO 9200
            LNHIST=LNHIST+1
            INHIST(LNHIST)=-J
            DPHIST(LNHIST)=BNDNEW
C
C     Modify RHS (take account of the new LOWER bound on Xj).
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            DO 2420 IKX=KBEG,KEND
               IR=RWNMBS(IKX)
               B(IR)=B(IR)-BNDNEW*ACOEFF(IKX)
               IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 2420       CONTINUE
C
C *** DEBUGGING
            IF(MSGLEV.LE.2) GO TO 2422
            WRITE(BUFFER,2421) J,STAVAR(J),BNDNEW
 2421       FORMAT(1X,'BIG Uj, cl=',I6,' st=',I6,' oldLOj=0.0D0',
     X       ' newLOj=',D16.8)
            CALL MYWRT(IOLOG,BUFFER)
 2422       CONTINUE
            IF(MSGLEV.LE.1) GO TO 2424
            WRITE(BUFFER,2423) J,CLNAME(J),BNDNEW
 2423       FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X       ') has new LOWER bound=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
 2424       CONTINUE
            GO TO 2500
C
C     Here for an aggregate variable. Depending on its entry sign
C     (+ or -) its UPPER or LOWER bound can be improved, respectively.
 2440       IF(ACOEFF(KNEGBG).LT.0.0D0) THEN
               IF(BNDNEW.LE.RNTMP1(J)+BNDTOL) GO TO 2500
               RNTMP1(J)=BNDNEW
               IF(MSGLEV.LE.1) GO TO 2442
               WRITE(BUFFER,2441) J,BNDNEW
 2441          FORMAT(1X,'FDAGGR: Variable ',I6,
     X          ' (aggregate)     has new LOWER bound=',D14.6)
               CALL MYWRT(IOLOG,BUFFER)
 2442          CONTINUE
            ELSE
               IF(BNDNEW.GE.RNTMP2(J)-BNDTOL) GO TO 2500
               RNTMP2(J)=BNDNEW
               IF(MSGLEV.LE.1) GO TO 2444
               WRITE(BUFFER,2443) J,BNDNEW
 2443          FORMAT(1X,'FDAGGR: Variable ',I6,
     X          ' (aggregate)     has new UPPER bound=',D14.6)
               CALL MYWRT(IOLOG,BUFFER)
 2444          CONTINUE
            ENDIF
C
         ENDIF
C
C
C
 2500    CONTINUE
         IF(RWSTAT(I).EQ.1.OR.RWSTAT(I).EQ.2) THEN
C
C     Here for EQUALITY type or GREATER OR EQUAL type constraint.
            IF(BUPPER-RHS0.LT.-FSBTOL.AND.NPOSBG.EQ.0) GO TO 9020
            IF(NPOSBG.GE.1) GO TO 2700
C
C
C     Here if there are no positive entries with BIG Uj
C         and there are no negative entries with BIG Lj,
C         i.e. BUPPER is finite.
C
C     Loop over nonzero entries of row I.
            BUPPER=RHS0-BUPPER
            IPOS=KRWBEG
 2600       IF(IPOS.EQ.0) GO TO 2680
               J=CLNMBS(IPOS)
               K=STAVAR(J)
               IF(K.GE.6) GO TO 2660
               IF(IAGGR(J).GT.0) GO TO 2660
               IF(ACOEFF(IPOS).LT.0.0D0) THEN
C
C     Implicit UPPER bound can be defined for each variable
C     refering to NEGATIVE entry of row I.
                  BNDJUP=RNTMP2(J)
                  BNDNEW=RNTMP1(J)+BUPPER/ACOEFF(IPOS)
                  IF(BNDNEW.GE.BNDJUP-BNDTOL) GO TO 2660
C                 IF(BNDNEW.GE.BNDBIG) GO TO 2660
C                 IF(BNDNEW.GE.BNDBIG) THEN 
C                    IF(IAGGR(J).EQ.0) GO TO 2660
C                 ENDIF
                  IF(BNDNEW.GE.BNDBIG) THEN 
                     IF((K.EQ.0.OR.K.EQ.2).AND.
     X                IAGGR(J).EQ.0) GO TO 2660
                  ENDIF
                  IF(ITIGHT.LE.1) GO TO 2660
                  NTIGHT=NTIGHT+1
                  RNTMP2(J)=BNDNEW
                  IF(IAGGR(J).EQ.0) THEN
                     UPBND(J)=BNDNEW
                     IF(K.NE.1.AND.K.NE.3) STAVAR(J)=STAVAR(J)+1
                  ENDIF
C
                  IF(MSGLEV.LE.1) GO TO 2604
C                 WRITE(BUFFER,2601) J,STAVAR(J),BNDJUP,BNDNEW
C2601             FORMAT(1X,'cl=',I6,' st=',I6,' oldUPj=',D16.8,
C    X             ' newUPj=',D16.8)
C                 CALL MYWRT(IOLOG,BUFFER)
                  IF(IAGGR(J).EQ.0) THEN
                     WRITE(BUFFER,2602) J,CLNAME(J),BNDNEW
 2602                FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X                ') has new UPPER bound=',D14.6)
                     CALL MYWRT(IOLOG,BUFFER)
                  ELSE
                     WRITE(BUFFER,2603) J,BNDNEW
 2603                FORMAT(1X,'FDAGGR: Variable ',I6,
     X                ' (aggregate)     has new UPPER bound=',D14.6)
                     CALL MYWRT(IOLOG,BUFFER)
                  ENDIF
 2604             CONTINUE
C
C     Reinitialize bounds on shadow prices.
                  KBEG=CLPNTS(J)
                  KEND=KBEG+LENCOL(J)-1
                  DO 2605 IKX=KBEG,KEND
                     IR=RWNMBS(IKX)
                     P(IR)=-BIG
                     Q(IR)=BIG
                     IF(RANGES(IR).LE.BIGNEW) GO TO 2605
                     IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                     IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
 2605             CONTINUE
C
                  IF(BNDNEW.LE.RNTMP1(J)+FSBTOL) THEN
C
C     Fix variable J on its LOWER bound.
                     IF(IAGGR(J)) 2606,2610,2660
C
C     Here for an aggregate variable.
 2606                RNTMP2(J)=RNTMP1(J)
                     IF(MSGLEV.LE.1) GO TO 2609
                     WRITE(BUFFER,2608) J,RNTMP1(J)
 2608                FORMAT(1X,'FDAGGR: Variable ',I6,
     X                ' (aggregate)     is being FIXED on X=',D14.6)
                     CALL MYWRT(IOLOG,BUFFER)
 2609                CONTINUE
                     GO TO 2660
C
C     Here for a standard LP variable.
 2610                NFIXED=NFIXED+1
                     X0=0.0D0
                     PRLVAR(J)=X0
                     STAVAR(J)=6
                     IF(MSGLEV.LE.1) GO TO 2612
                     WRITE(BUFFER,2611) J,CLNAME(J),X0
 2611                FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X                ') is being FIXED on X=',D14.6)
                     CALL MYWRT(IOLOG,BUFFER)
 2612                CONTINUE
                     GO TO 2660
                  ENDIF
C
               ELSE
C
C     Implicit LOWER bound can be defined for each variable
C     refering to POSITIVE entry of row I.
                  BNDJUP=RNTMP2(J)
                  BNDNEW=BNDJUP+BUPPER/ACOEFF(IPOS)
                  IF(BNDNEW.LE.RNTMP1(J)+BNDTOL) GO TO 2660
                  IF(IAGGR(J)) 2620,2630,2660
C
C     Here for an aggregate variable.
 2620             RNTMP1(J)=BNDNEW
                  IF(MSGLEV.LE.1) GO TO 2629
                  WRITE(BUFFER,2628) J,RNTMP1(J)
 2628             FORMAT(1X,'FDAGGR: Variable ',I6,
     X             ' (aggregate)     has new LOWER bound=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 2629             CONTINUE
                  GO TO 2660
C
C     Here for a standard LP variable.
 2630             NTIGHT=NTIGHT+1
                  LOBND(J)=LOBND(J)+BNDNEW
                  UPBND(J)=UPBND(J)-BNDNEW
                  RNTMP2(J)=UPBND(J)
                  STAVAR(J)=3
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
                  IF(LNHIST.GE.MXHIST) GO TO 9200
                  LNHIST=LNHIST+1
                  INHIST(LNHIST)=-J
                  DPHIST(LNHIST)=BNDNEW
C
C     Modify RHS (take account of the new LOWER bound on Xj).
                  KBEG=CLPNTS(J)
                  KEND=KBEG+LENCOL(J)-1
                  DO 2640 IKX=KBEG,KEND
                     IR=RWNMBS(IKX)
                     B(IR)=B(IR)-BNDNEW*ACOEFF(IKX)
                     IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 2640             CONTINUE
C
C *** DEBUGGING
                  IF(MSGLEV.LE.2) GO TO 2642
                  WRITE(BUFFER,2641) J,STAVAR(J),BNDNEW
 2641             FORMAT(1X,'cl=',I6,' st=',I6,' oldLOj=0.0D0',
     X             ' newLOj=',D16.8)
                  CALL MYWRT(IOLOG,BUFFER)
 2642             CONTINUE
                  IF(MSGLEV.LE.1) GO TO 2644
                  WRITE(BUFFER,2643) J,CLNAME(J),BNDNEW
 2643             FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X             ') has new LOWER bound=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 2644             CONTINUE
C
               ENDIF
 2660          IPOS=RWLINK(IPOS)
               GO TO 2600
C
 2680       CONTINUE
            BUPPER=RHS0-BUPPER
            GO TO 3000
C
C
C     Here if there exists at least one positive entry with BIG Uj
C          or there exists at least one negative entry with BIG Lj.
C     If only one such an entry exists, then its -big LOWER bound or
C     big UPPER bound can be improved. KPOSBG indicates its position.
 2700       IF(NPOSBG.GE.2) GO TO 3000
C
            J=CLNMBS(KPOSBG)
            K=STAVAR(J)
            IF(K.GE.6) GO TO 3000
            BNDNEW=(RHS0-BUPPER)/ACOEFF(KPOSBG)
            IF(IAGGR(J)) 2740,2710,3000
C
C     Here for a standard LP variable (it must have been positive
C     entry with big Uj). Variable's LOWER bound can be improved.
 2710       IF(BNDNEW.LE.BNDTOL) GO TO 3000
            NTIGHT=NTIGHT+1
            LOBND(J)=LOBND(J)+BNDNEW
            STAVAR(J)=2
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
            IF(LNHIST.GE.MXHIST) GO TO 9200
            LNHIST=LNHIST+1
            INHIST(LNHIST)=-J
            DPHIST(LNHIST)=BNDNEW
C
C     Modify RHS (take account of the new LOWER bound on Xj).
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            DO 2720 IKX=KBEG,KEND
               IR=RWNMBS(IKX)
               B(IR)=B(IR)-BNDNEW*ACOEFF(IKX)
               IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 2720       CONTINUE
C
C *** DEBUGGING
            IF(MSGLEV.LE.2) GO TO 2722
            WRITE(BUFFER,2721) J,STAVAR(J),BNDNEW
 2721       FORMAT(1X,'BIG Uj, cl=',I6,' st=',I6,' oldLOj=0.0D0',
     X       ' newLOj=',D16.8)
            CALL MYWRT(IOLOG,BUFFER)
 2722       CONTINUE
            IF(MSGLEV.LE.1) GO TO 2724
            WRITE(BUFFER,2723) J,CLNAME(J),BNDNEW
 2723       FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X       ') has new LOWER bound=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
 2724       CONTINUE
            GO TO 3000
C
C     Here for an aggregate variable. Depending on its entry sign
C     (+ or -) its LOWER or UPPER bound can be improved, respectively.
 2740       IF(ACOEFF(KPOSBG).GT.0.0D0) THEN
               IF(BNDNEW.LE.RNTMP1(J)+BNDTOL) GO TO 3000
               RNTMP1(J)=BNDNEW
               IF(MSGLEV.LE.1) GO TO 2742
               WRITE(BUFFER,2741) J,BNDNEW
 2741          FORMAT(1X,'FDAGGR: Variable ',I6,
     X          ' (aggregate)     has new LOWER bound=',D14.6)
               CALL MYWRT(IOLOG,BUFFER)
 2742          CONTINUE
            ELSE
               IF(BNDNEW.GE.RNTMP2(J)-BNDTOL) GO TO 3000
               RNTMP2(J)=BNDNEW
               IF(MSGLEV.LE.1) GO TO 2744
               WRITE(BUFFER,2743) J,BNDNEW
 2743          FORMAT(1X,'FDAGGR: Variable ',I6,
     X          ' (aggregate)     has new UPPER bound=',D14.6)
               CALL MYWRT(IOLOG,BUFFER)
 2744          CONTINUE
            ENDIF
C
         ENDIF
C
C
C
C
C
C     End of the third main loop.
 3000 CONTINUE
C
C
C
C     Check if 2000 and 3000 loops should be repeated.
      IF(NELIM.GT.NEL0) THEN
         IRUN=IRUN+1
         GO TO 1200
      ENDIF
      IF(100*(NTIGHT-NT0).GE.N1) THEN
         IRUN=IRUN+1
         IF(IRUN.GE.6) GO TO 3100
         GO TO 1200
      ELSE
         GO TO 3100
      ENDIF
C
C
C
C
C
C
C     Analyse all aggregate columns.
 3100 CONTINUE
C     WRITE(BUFFER,3101)
C3101 FORMAT(1X,'FDAGGR: Summary on aggregation and elimination.')
C     CALL MYWRT(IOLOG,BUFFER)
C
C
      NFREE=0
      DO 4000 J=1,N
         KSTAT=STAVAR(J)
         IF(KSTAT.GE.6) GO TO 4000
         IF(J.GT.NSTRCT) GO TO 4000
C
C
C
C     Check if new variables' bounds are OK.
         IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) THEN
            IF(UPBND(J).GE.1.0D+10) THEN
               WRITE(BUFFER,3111) J,STAVAR(J),LOBND(J),UPBND(J)
 3111          FORMAT(1X,'cl=',I6,' st=',I2,' L=',D10.3,' U=',D10.3)
               CALL MYWRT(IOLOG,BUFFER)
            ENDIF
         ENDIF
         IF(KSTAT.EQ.2.OR.KSTAT.EQ.3) THEN
            IF(DABS(LOBND(J)).GE.1.0D+10) THEN
               WRITE(BUFFER,3112) J,STAVAR(J),LOBND(J),UPBND(J)
 3112          FORMAT(1X,'cl=',I6,' st=',I2,' L=',D10.3,' U=',D10.3)
               CALL MYWRT(IOLOG,BUFFER)
            ENDIF
         ENDIF
C
         IF(IAGGR(J).GE.0) GO TO 4000
C
C
C
C     Check if an aggregate can be eliminated.
C     It can in two cases:
C     1. when it is FIXED (i.e. RNTMP1(J) = RNTMP2(J));
C     2. when it is an isolated row of the LP problem;
C     3. when it is a bounded FREE variable.
C     Count the number of entries in an aggregate.
C     Count the number of positive entries in an aggregate.
         NIDEN=0
         NPOS=0
         JCOL=-IAGGR(J)
 3210    IF(JCOL.LE.0) GO TO 3220
            NIDEN=NIDEN+1
            IF(RNTMP3(JCOL).GE.SMALLA) NPOS=NPOS+1
         JCOL=IAGGR(JCOL)
         GO TO 3210
C
 3220    KOK=INTMP1(J)
         KROW=RWNMBS(KOK)
         IF(RNTMP2(J).GE.RNTMP1(J)+FSBTOL.AND.
     X    (LENROW(KROW).GT.NIDEN.OR.LENCOL(J).GT.1)) GO TO 3300
C
C
C     Here if an aggregate can be eliminated.
C     Define a one-row linear program.
         NXSTR=0
         JCOL=-IAGGR(J)
 3230    IF(JCOL.LE.0) GO TO 3240
            NXSTR=NXSTR+1
            COEFF(NXSTR)=RNTMP3(JCOL)
            COBJ(NXSTR)=C(JCOL)
            KSTAT=STAVAR(JCOL)
            BNDJUP=BIG
            IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) BNDJUP=UPBND(JCOL)
            UPPER(NXSTR)=BNDJUP
         JCOL=IAGGR(JCOL)
         GO TO 3230
C
 3240    NMAX=100
         IF(NXSTR.GE.NMAX-1) GO TO 3400
         PI=P(KROW)
         QI=Q(KROW)
         ROWST=RWSTAT(KROW)
         RHS=B(KROW)
         IF(RNTMP2(J).LE.RNTMP1(J)+FSBTOL) THEN
            ROWST=1
            RHS=RNTMP1(J)
         ELSE
            DO 3245 I=1,NXSTR
               COEFF(I)=COEFF(I)*ACOEFF(KOK)
 3245       CONTINUE
         ENDIF
C
C     Solve the one-row LP problem.
         CALL SMPLX(IOLOG,MSGLEV,NMAX,NX,NXSTR,ROWST,
     X    COEFF,X,COBJ,UPPER,RDCOST,RHS,P,Q,DUAL)
C
C     Desaggregate optimal solution of the one-row LP problem.
C     Fix all columns that belong to the aggregate.
         NXSTR=0
         JCOL=-IAGGR(J)
         X0=0.0D0
 3250    IF(JCOL.LE.0) GO TO 3260
            NXSTR=NXSTR+1
            NFIXED=NFIXED+1
            PRLVAR(JCOL)=X(NXSTR)
            STAVAR(JCOL)=6
            IF(MSGLEV.LE.1) GO TO 3252
            WRITE(BUFFER,3251) JCOL,CLNAME(JCOL),X(NXSTR)
 3251       FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X       ') is being FIXED on X=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
 3252       CONTINUE
            X0=X0+X(NXSTR)*COEFF(NXSTR)
            K=JCOL
         JCOL=IAGGR(JCOL)
         IAGGR(K)=0
         GO TO 3250
C
C     Update RHS after aggregate elimination.
 3260    IF(RNTMP2(J).GT.RNTMP1(J)+FSBTOL) X0=X0/ACOEFF(KOK)
         IF(DABS(X0).LE.SMALLA) GO TO 3280
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 3270 K=KBEG,KEND
            IR=RWNMBS(K)
            B(IR)=B(IR)-X0*ACOEFF(K)
            IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 3270    CONTINUE
 3280    CONTINUE
C
C     Check if a row defining the aggregate can be eliminated.
         IF(LENROW(KROW).NE.NXSTR) GO TO 4000
         NELIM=NELIM+1
         RWHEAD(KROW)=-RWHEAD(KROW)
         IF(MSGLEV.LE.1) GO TO 3282
         RTYPE='EQ'
         IF(RWSTAT(KROW).EQ.2) RTYPE='GE'
         IF(RWSTAT(KROW).EQ.3) RTYPE='LE'
         WRITE(BUFFER,3281) KROW,RWNAME(KROW),RTYPE
 3281    FORMAT(1X,'FDAGGR: Row      ',I6,' (name=',A8,
     X    ' type=',A2,') is eliminated.')
         CALL MYWRT(IOLOG,BUFFER)
 3282    CONTINUE
C
         GO TO 4000
C
C
C
C     Check if a split FREE variable has been bounded.
 3300    IHEAD=-IAGGR(J)
         IF(IAGGR(IHEAD).NE.J) GO TO 3400
         IF(-STAVAR(IHEAD).NE.J) GO TO 3400
         IF(RNTMP1(J).LE.-BIGNEW.AND.
     X    RNTMP2(J).GE.BIGNEW) GO TO 4000
         IF(MSGLEV.LE.2) GO TO 3302
         WRITE(BUFFER,3301) J,RNTMP1(J),RNTMP2(J)
 3301    FORMAT(1X,'FREE var=',I6,' Lbnd=',D10.3,' Ubnd=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
 3302    CONTINUE
C
C     Check if a variable has a finite LOWER bound.
         IF(RNTMP1(J).LE.-BIGNEW) GO TO 3340
         IF(MSGLEV.LE.2) GO TO 3312
         WRITE(BUFFER,3311) J
 3311    FORMAT(1X,'FDAGGR: LO bnd on a FREE variable ',I6)
         CALL MYWRT(IOLOG,BUFFER)
 3312    CONTINUE
         JCOL=J
 3310    KSTAT=STAVAR(JCOL)
         KBEG=CLPNTS(JCOL)
         KEND=KBEG+LENCOL(JCOL)-1
         DO 3315 K=KBEG,KEND
            IR=RWNMBS(K)
            B(IR)=B(IR)-RNTMP1(JCOL)*ACOEFF(K)
            IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 3315    CONTINUE
         STAVAR(JCOL)=2
         LOBND(JCOL)=RNTMP1(JCOL)
         IAGGR(JCOL)=0
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
         IF(LNHIST.GE.MXHIST) GO TO 9200
         LNHIST=LNHIST+1
         INHIST(LNHIST)=-JCOL
         DPHIST(LNHIST)=RNTMP1(JCOL)
C
C     Check if a variable has also an UPPER bound.
         RNTMP2(JCOL)=RNTMP2(JCOL)-RNTMP1(JCOL)
         IF(RNTMP2(JCOL).LE.BIGNEW) THEN
            STAVAR(JCOL)=3
            UPBND(JCOL)=RNTMP2(JCOL)
C
C     Reinitialize bounds on shadow prices.
            KBEG=CLPNTS(JCOL)
            KEND=KBEG+LENCOL(JCOL)-1
            DO 3320 IKX=KBEG,KEND
               IR=RWNMBS(IKX)
               P(IR)=-BIG
               Q(IR)=BIG
               IF(RANGES(IR).LE.BIGNEW) GO TO 3320
               IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
               IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
 3320       CONTINUE
         ENDIF
C
C     Fix the brother of a split variable.
         JCOL=-KSTAT
         X0=0.0D0
         PRLVAR(JCOL)=X0
         STAVAR(JCOL)=6
         IAGGR(JCOL)=0
         NFIXED=NFIXED+1
         IF(MSGLEV.LE.1) GO TO 3322
         WRITE(BUFFER,3321) JCOL,CLNAME(JCOL),X0
 3321    FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X    ') is being FIXED on X=',D14.6)
         CALL MYWRT(IOLOG,BUFFER)
 3322    CONTINUE
         GO TO 4000
C
C     Here if a split variable has infinite LOWER bound
C     and a finite UPPER bound.
 3340    CONTINUE
         IF(MSGLEV.LE.2) GO TO 3342
         WRITE(BUFFER,3341) J
 3341    FORMAT(1X,'FDAGGR: UP bnd on a FREE variable ',I6)
         CALL MYWRT(IOLOG,BUFFER)
 3342    CONTINUE
C
C     UP bnd on x1 can be handled as LO bnd on x2.
         JCOL=-KSTAT
         RNTMP1(JCOL)=-RNTMP2(J)
         RNTMP2(JCOL)=-RNTMP1(J)
         GO TO 3310
C
C
C
C     Here for a general analysis of an aggregate.
C     Fix columns with strictly positive (or strictly negative)
C     reduced costs. Analyse every pair of columns from an aggregate.
C     J1 and JCOL are the column numbers.
 3400    J1=-IAGGR(J)
 3410    IF(J1.LE.0) GO TO 3900
            K1=STAVAR(J1)
            IF(K1.GE.6) GO TO 3880
            IF(K1.LT.0) GO TO 3880
            JCOL=IAGGR(J1)
 3420       IF(JCOL.LE.0) GO TO 3880
               KSTAT=STAVAR(JCOL)
               IF(KSTAT.GE.6) GO TO 3870
               IF(KSTAT.LT.0) GO TO 3870
C
C     ALPHA here is the inverse of what we used in 500 loop.
C     A(jcol) = ALPHA * A(j1).
               ALPHA=RNTMP3(JCOL)/RNTMP3(J1)
               DP=C(JCOL)-ALPHA*C(J1)
C              WRITE(BUFFER,3421) J1,JCOL,ALPHA
C3421          FORMAT(1X,'Pair of cl: J1=',I6,' J2=',I6,' A=',D14.8)
C              CALL MYWRT(IOLOG,BUFFER)
C
               IF(DABS(DP).LE.SMALLA) THEN
C
C     Two cols J1 and JCOL are identical subject to a scaling factor.
C     Two actions are possible:
C     if ALPHA > 0,  then aggregation;
C     if ALPHA < 0,  then a new split FREE variable is found;
                  IF(ALPHA.LE.0.0D0) GO TO 3460
C
C     Linearly dependent variables are found.
C
C                 IF(MSGLEV.LE.2) GO TO 3440
C                 WRITE(BUFFER,3431) CLNAME(J1),CLNAME(JCOL)
C3431             FORMAT(1X,'FDAGGR: LP variables:   ',
C    X             A8,' and ',A8,' are lin dep. 3431')
C                 CALL MYWRT(IOLOG,BUFFER)
C                 WRITE(BUFFER,3432) J1,C(J1),STAVAR(J1),
C    X             LENCOL(J1),UPBND(J1)
C3432             FORMAT(1X,'    var=',I6,' Cj=',D14.8,' st=',I6,
C    X             ' ln=',I6,' Uj=',D14.8)
C                 CALL MYWRT(IOLOG,BUFFER)
C                 WRITE(BUFFER,3433) JCOL,C(JCOL),STAVAR(JCOL),
C    X             LENCOL(JCOL),UPBND(JCOL)
C3433             FORMAT(1X,'and var=',I6,' Cj=',D14.8,' st=',I6,
C    X             ' ln=',I6,' Uj=',D14.8)
C                 CALL MYWRT(IOLOG,BUFFER)
C                 WRITE(BUFFER,3434) J1,JCOL,ALPHA,DP
C3434             FORMAT(1X,'FDAGGR: J1=',I6,' JCOL=',I6,' ALPHA=',
C    X             D14.8,' DP=',D14.8)
C                 CALL MYWRT(IOLOG,BUFFER)
C3440             CONTINUE
C
                  IF(MSGLEV.LE.1) GO TO 3442
                  WRITE(BUFFER,3441) CLNAME(J1),CLNAME(JCOL)
 3441             FORMAT(1X,'FDAGGR: Variable=',A8,
     X             ' becomes a weighted sum of ',A8,' and itself.')
                  CALL MYWRT(IOLOG,BUFFER)
 3442             CONTINUE
                  NFIXED=NFIXED+1
                  X0=0.0D0
                  PRLVAR(JCOL)=X0
                  STAVAR(JCOL)=6
                  IF(MSGLEV.LE.1) GO TO 3452
                  WRITE(BUFFER,3451) JCOL,CLNAME(JCOL),X0
 3451             FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X             ') is being FIXED on X=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 3452             CONTINUE
                  BNDJUP=UPBND(J1)+ALPHA*UPBND(JCOL)
                  STAVAR(J1)=2
                  UPBND(J1)=BNDJUP
                  IF(BNDJUP.LE.BIGNEW) THEN
                     STAVAR(J1)=3
C
C     Reinitialize bounds on shadow prices.
                     KBEG=CLPNTS(J1)
                     KEND=KBEG+LENCOL(J1)-1
                     DO 3450 IKX=KBEG,KEND
                        IR=RWNMBS(IKX)
                        P(IR)=-BIG
                        Q(IR)=BIG
                        IF(RANGES(IR).LE.BIGNEW) GO TO 3450
                        IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                        IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
 3450                CONTINUE
                  ENDIF
                  GO TO 3870
C
C     Split FREE variable is found.
C     We should check if none of them is bounded.
 3460             IF(K1.EQ.1.OR.K1.EQ.3.OR.
     X             KSTAT.EQ.1.OR.KSTAT.EQ.3) THEN
C                    WRITE(BUFFER,3461) J1,C(J1),STAVAR(J1),
C    X                LENCOL(J1),UPBND(J1)
C3461                FORMAT(1X,'3461,var=',I6,' Cj=',D14.8,' st=',I6,
C    X                ' ln=',I6,' Uj=',D14.8)
C                    CALL MYWRT(IOLOG,BUFFER)
C                    WRITE(BUFFER,3462) JCOL,C(JCOL),STAVAR(JCOL),
C    X                LENCOL(JCOL),UPBND(JCOL)
C3462                FORMAT(1X,' and var=',I6,' Cj=',D14.8,' st=',I6,
C    X                ' ln=',I6,' Uj=',D14.8)
C                    CALL MYWRT(IOLOG,BUFFER)
                     GO TO 3480
                  ENDIF
                  NFREE=NFREE+1
                  IF(MSGLEV.LE.1) GO TO 3464
                  WRITE(BUFFER,3463) CLNAME(J1),CLNAME(JCOL)
 3463             FORMAT(1X,'FDAGGR: LP variables:   ',
     X             A8,' and ',A8,' are split FREE variable.')
                  CALL MYWRT(IOLOG,BUFFER)
 3464             CONTINUE
                  STAVAR(J1)=-JCOL
                  STAVAR(JCOL)=-J1
                  C(JCOL)=-C(JCOL)/ALPHA
                  KBEG=CLPNTS(JCOL)
                  KEND=KBEG+LENCOL(JCOL)-1
                  DO 3470 K=KBEG,KEND
                     ACOEFF(K)=-ACOEFF(K)/ALPHA
 3470             CONTINUE
C
C     Save column scaling factor in a PRE_SOLVE history list.
C     At the moment we ignore this.
                  GO TO 3880
C
C     Linearly dependent variables are found.
C     Even for negative ALPHA, aggregation can be done if at least
C     one variable has a finite UPPER bound.
 3480             CONTINUE
                  WRITE(BUFFER,3481) J1,JCOL
 3481             FORMAT(1X,'3481F,  aggr J1=',I6,' JCOL=',I6)
                  CALL MYWRT(IOLOG,BUFFER)
                  GO TO 3870
C
C
C
               ENDIF
C
               IF(K1.EQ.1.OR.K1.EQ.3) GO TO 3540
C
C     Here if variable J1 is unbounded.
C     This means that its reduced cost must be nonnegative. We thus
C     might be able to determine the sign of the reduced cost for
C     variable JCOL and, consequently, eliminate it.
C
               IF(DP.GE.SMALLA.AND.ALPHA.GE.0.0D0) THEN
C
C     Fix variable JCOL on its LOWER bound.
                  NFIXED=NFIXED+1
                  X0=0.0D0
                  PRLVAR(JCOL)=X0
                  STAVAR(JCOL)=6
                  IF(MSGLEV.LE.1) GO TO 3502
                  WRITE(BUFFER,3501) JCOL,CLNAME(JCOL),X0
 3501             FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X             ') is being FIXED on X=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 3502             CONTINUE
                  GO TO 3870
               ENDIF
C
               IF(DP.LE.-SMALLA.AND.ALPHA.LE.0.0D0) THEN
C
C     Fix variable JCOL on its UPPER bound (if it has one).
                  IF(KSTAT.NE.1.AND.KSTAT.NE.3) GO TO 3540
                  NFIXED=NFIXED+1
                  X0=UPBND(JCOL)
                  PRLVAR(JCOL)=X0
                  STAVAR(JCOL)=6
                  KBEG=CLPNTS(JCOL)
                  KEND=KBEG+LENCOL(JCOL)-1
                  DO 3510 K=KBEG,KEND
                     IR=RWNMBS(K)
                     B(IR)=B(IR)-X0*ACOEFF(K)
                     IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 3510             CONTINUE
                  IF(MSGLEV.LE.1) GO TO 3512
                  WRITE(BUFFER,3511) JCOL,CLNAME(JCOL),X0
 3511             FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X             ') is being FIXED on X=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 3512             CONTINUE
                  GO TO 3870
               ENDIF
C
 3540          IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) GO TO 3870
C
C     Here if variable JCOL is unbounded.
C     This means that its reduced cost must be nonnegative. We thus
C     might be able to determine the sign of the reduced cost for
C     variable J1 and, consequently, eliminate it.
C
               IF(DP.LE.-SMALLA.AND.ALPHA.GE.0.0D0) THEN
C
C     Fix variable J1 on its LOWER bound.
                  NFIXED=NFIXED+1
                  X0=0.0D0
                  PRLVAR(J1)=X0
                  STAVAR(J1)=6
                  IF(MSGLEV.LE.1) GO TO 3562
                  WRITE(BUFFER,3561) J1,CLNAME(J1),X0
 3561             FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X             ') is being FIXED on X=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 3562             CONTINUE
                  GO TO 3880
               ENDIF
C
               IF(DP.LE.-SMALLA.AND.ALPHA.LE.0.0D0) THEN
C
C     Fix variable J1 on its UPPER bound (if it has one).
                  IF(K1.NE.1.AND.K1.NE.3) GO TO 3870
                  NFIXED=NFIXED+1
                  X0=UPBND(J1)
                  PRLVAR(J1)=X0
                  STAVAR(J1)=6
                  KBEG=CLPNTS(J1)
                  KEND=KBEG+LENCOL(J1)-1
                  DO 3570 K=KBEG,KEND
                     IR=RWNMBS(K)
                     B(IR)=B(IR)-X0*ACOEFF(K)
                     IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 3570             CONTINUE
                  IF(MSGLEV.LE.1) GO TO 3572
                  WRITE(BUFFER,3571) J1,CLNAME(J1),X0
 3571             FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X             ') is being FIXED on X=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 3572             CONTINUE
                  GO TO 3880
               ENDIF
C
C
C
 3870       JCOL=IAGGR(JCOL)
            GO TO 3420
C
 3880    J1=IAGGR(J1)
         GO TO 3410
C
C
C
C     Here to tighten bounds on variables entering an aggregate.
 3900    IF(NIDEN.GT.NPOS) GO TO 4000
C        IF(RNTMP1(J).LE.-BIGNEW.AND.
C    X    RNTMP2(J).GE.BIGNEW) GO TO 4000
         IF(RNTMP2(J).GE.BIGNEW) GO TO 4000
         IF(MSGLEV.LE.2) GO TO 3903
         WRITE(BUFFER,3901) J,RNTMP1(J),RNTMP2(J)
 3901    FORMAT(1X,'Bnd aggr J=',I6,' Lbnd=',D10.3,' Ubnd=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,3902) J,NIDEN,NPOS
 3902    FORMAT(1X,'Bnd aggr J=',I6,' NIDEN=',I6,' NPOS=',I6)
         CALL MYWRT(IOLOG,BUFFER)
 3903    CONTINUE
C
         JCOL=-IAGGR(J)
 3910    IF(JCOL.LE.0) GO TO 4000
            KSTAT=STAVAR(JCOL)
            IF(KSTAT.GE.6) GO TO 3990
            IF(KSTAT.LT.0) GO TO 3990
C
C     Implicit UPPER bound can be defined for each variable
C     refering to POSITIVE entry of an aggredate.
            BNDJUP=BIGNEW
            IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) BNDJUP=UPBND(JCOL)
            BNDNEW=RNTMP2(J)/RNTMP3(JCOL)
            IF(BNDNEW.GE.BNDJUP-BNDTOL) GO TO 3990
C           IF(BNDNEW.GE.BNDBIG) GO TO 3990
            IF(BNDNEW.GE.BNDBIG) THEN 
               IF(KSTAT.EQ.0.OR.KSTAT.EQ.2) GO TO 3990
            ENDIF
            IF(ITIGHT.LE.1) GO TO 3990
            NTIGHT=NTIGHT+1
            UPBND(JCOL)=BNDNEW
            IF(KSTAT.NE.1.AND.KSTAT.NE.3) STAVAR(JCOL)=STAVAR(JCOL)+1
C
C     Reinitialize bounds on shadow prices.
            KBEG=CLPNTS(JCOL)
            KEND=KBEG+LENCOL(JCOL)-1
            DO 3920 IKX=KBEG,KEND
               IR=RWNMBS(IKX)
               P(IR)=-BIG
               Q(IR)=BIG
               IF(RANGES(IR).LE.BIGNEW) GO TO 3920
               IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
               IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
 3920       CONTINUE
C
            IF(MSGLEV.LE.1) GO TO 3924
C           WRITE(BUFFER,3921) JCOL,STAVAR(JCOL),BNDJUP,BNDNEW
C3921       FORMAT(1X,'cl=',I6,' st=',I6,' oldUPj=',D16.8,
C    X       ' newUPj=',D16.8)
C           CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,3922) JCOL,CLNAME(JCOL),BNDNEW
 3922       FORMAT(1X,'FDAGGR: Variable ',I6,' (name=',A8,
     X       ') has new UPPER bound=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
 3924       CONTINUE
C
C
 3990    JCOL=IAGGR(JCOL)
         GO TO 3910
C
C
C
C
 4000 CONTINUE
C
C
C
C
C
C     Check if there are inequality type rows to be eliminated.
C     Check if the eliminated rows were not violated.
      FSBTOL=1.0D-7
      DO 4100 I=1,M
         IF(DABS(B(I)).LE.FSBTOL) B(I)=0.0D0
         K=RWHEAD(I)
C        WRITE(BUFFER,4101) I,RWSTAT(I),LENROW(I),K
C4101    FORMAT(1X,'row=',I6,'  st=',I2,'  ln=',I6,'  K=',I8)
C        CALL MYWRT(IOLOG,BUFFER)
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY constraint.
            IF(K.GT.0) GO TO 4100
            IF(DABS(B(I)).GT.FSBTOL) GO TO 9030
            GO TO 4100
         ENDIF
         IF(LENROW(I).GE.1) GO TO 4100
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
            IF(B(I).GT.FSBTOL) GO TO 9030
            IF(K.LE.0) GO TO 4100
            NELIM=NELIM+1
            RWHEAD(I)=-RWHEAD(I)
            J=CLNMBS(K)
            PRLVAR(J)=0.0D0
            STAVAR(J)=14
            NFIXED=NFIXED+1
            GO TO 4100
         ENDIF
         IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
            IF(B(I).LT.-FSBTOL) GO TO 9030
            IF(K.LE.0) GO TO 4100
            NELIM=NELIM+1
            RWHEAD(I)=-RWHEAD(I)
            J=CLNMBS(K)
            PRLVAR(J)=0.0D0
            STAVAR(J)=14
            NFIXED=NFIXED+1
            GO TO 4100
         ENDIF
 4100 CONTINUE
C
C
C
C
C
C
C     Determine the permutation that puts all empty and inactive
C     rows at the end of the list.
C
      I=3
      IF(MSGLEV.LE.1) I=4
      CALL EMPTYR(MAXM,M,MNEW,I,
     X RWHEAD,STAROW,LENROW,MARKER,IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the analysis of EMPTYR.
      IF(MNEW.LT.M.OR.NFIXED.GT.0) THEN
C
         CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X    CLPNTS,RWNMBS,
     X    RWHEAD,RWLINK,CLNMBS,LENCOL,
     X    LENROW,MARKER,IMTMP1,IROW,RELT,
     X    RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C     Reorder bounds on shadow prices P and Q.
         CALL REORDV(MAXM,M,
     X    LENROW,MARKER,P,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    LENROW,MARKER,Q,RELT,IOLOG)
C
C     Reorder elements within each column of the  LP constraint
C     matrix in such a way that those of the active part of  A
C     are at the beginning of the lists. The column lengths will
C     later be decreased to forget inactive part of matrix  A.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD and LENROW arrays.
         DO 4220 I=1,M
            RWHEAD(I)=0
            LENROW(I)=0
 4220    CONTINUE
C
C     Reorder nonzero elements within each column.
         DO 4280 J=1,N
            IF(STAVAR(J).GE.6) GO TO 4280
            KBEG=CLPNTS(J)-1
            KOK=0
            KOUT=0
            DO 4240 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=RWNMBS(K)
               IF(I.LE.MNEW) THEN
                  KOK=KOK+1
                  IROW(KOK)=RWNMBS(K)
                  RELT(KOK)=ACOEFF(K)
               ELSE
                  IPOS=LENCOL(J)-KOUT
                  KOUT=KOUT+1
                  IROW(IPOS)=RWNMBS(K)
                  RELT(IPOS)=ACOEFF(K)
               ENDIF
 4240       CONTINUE
            LENCOL(J)=KOK
C
C     Set the row linked lists.
C     Count nonzero elements in all rows of  A.
            DO 4260 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=IROW(IKX)
               RWNMBS(K)=I
               ACOEFF(K)=RELT(IKX)
               RWLINK(K)=RWHEAD(I)
               RWHEAD(I)=K
               LENROW(I)=LENROW(I)+1
 4260       CONTINUE
 4280    CONTINUE
C
C     Set the new number of rows of the constraint matrix.
C     Observe that row linked lists are OK.
         M=MNEW
C
      ENDIF
C
C
C
C
C
C
C     Here if a successful run of the loop has been completed.
      IF(MSGLEV.LE.0) GO TO 5010
      WRITE(BUFFER,5001) NELIM
 5001 FORMAT(1X,'FDAGGR: Constraints eliminated: ',I9)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,5002) NFIXED
 5002 FORMAT(1X,'        Variables eliminated:   ',I9)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,5003) NTIGHT
 5003 FORMAT(1X,'        Variable bounds improved:',I8)
      CALL MYWRT(IOLOG,BUFFER)
 5010 CONTINUE
C
C
C
 6000 CONTINUE
C
      RETURN
C
C
 9010 WRITE(BUFFER,9011) RWNAME(I),RTYPE,BLOWER,BUPPER,B(I)
 9011 FORMAT(1X,'FDAGGR: Row=',A8,' type=',A2,
     X ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9012)
 9012 FORMAT(1X,'FDAGGR: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9020 WRITE(BUFFER,9021) RWNAME(I),RTYPE,BLOWER,BUPPER,RHS0
 9021 FORMAT(1X,'FDAGGR: Row=',A8,' type=',A2,
     X ' BLO=',D10.3,' BUP=',D10.3,' RHS0=',D10.3)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9022)
 9022 FORMAT(1X,'FDAGGR: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9030 WRITE(BUFFER,9031) I,RWNAME(I),B(I)
 9031 FORMAT(1X,'FDAGGR: Constraint ',I6,' (name=',A8,
     X ') is violated, B=',D12.6)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9032)
 9032 FORMAT(1X,'FDAGGR: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9200 WRITE(BUFFER,9201)
 9201 FORMAT(1X,'FDAGGR: Please increase space for PRE_SOLVE ',
     X 'history list.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(6)
C
C
C *** LAST CARD OF (FDAGGR) ***
      END
