C*************************************************************
C     *** FDIDEN ... FIND VARIABLES OF IDENTICAL STRUCTURE ***
C*************************************************************
C
      SUBROUTINE FDIDEN(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X B,C,LOBND,UPBND,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X P,Q,PRLVAR,STAVAR,RWSTAT,CLNAME,RANGES,
     X MARKER,LENROW,RMTMP1)
C
C
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,MAXM,MAXN,MAXNZA,M,N,NSTRCT
      INTEGER*4 LNHIST,MXHIST
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
      DOUBLE PRECISION ACOEFF(MAXNZA),B(MAXM)
      DOUBLE PRECISION C(MAXN),LOBND(MAXN),UPBND(MAXN),RMTMP1(MAXM)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      DOUBLE PRECISION P(MAXM),Q(MAXM),PRLVAR(MAXN),RANGES(MAXM)
      CHARACTER*8 CLNAME(MAXN)
      INTEGER*2 STAVAR(MAXN),RWSTAT(MAXM)
      INTEGER*2 LENROW(MAXM),MARKER(MAXM)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 KSTAT,LROW,KROW,NFIXED,NFREE
      INTEGER*4 K,KBEG,KEND,K2,K2BEG,K2END,KBEG0,KEND0
      INTEGER*4 I,IKX,IR,IPOS,J,JCOL,J1,J2
      DOUBLE PRECISION ALPHA,BIG,BIGNEW,FSBTOL,SMALLA
      DOUBLE PRECISION DP,X0,BNDL1,BNDU1,BNDU2
      CHARACTER*100 BUFFER
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     LNHIST  Length of the PRE_SOLVE history list.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     C       Objective function coefficients.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     LOBND   Array of lower bounds.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     PRLVAR  Primal variables of the linear program.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     CLNAME  Array of column names.
C     RANGES  Array of constraint ranges.
C
C *** ON OUTPUT:
C
C
C
C
C *** WORK ARRAYS:
C     MARKER  Half-length integer work array of size MAXM.
C     LENROW  Integer work array of size MAXM.
C     RMTMP1  Nonzero elements of the analysed column.
C
C
C
C
C *** PURPOSE
C     This routine finds variables that have identical structure.
C     Different actions are possible if such variables are found:
C     - they may define a splitting of a hidden FREE variable (such
C       variables, if not treated in a special way, might cause
C       serious stability problems in an interior point algorithm);
C     - they may be aggregated (and one variable is FIXED on its
C       bound).
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,MYSTOP,DABS
C
C
C *** NOTES
C     This routine is given direct access to the matrix A.
C     It alters hidden data structures.
C
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  January 31, 1993
C     Last modified: June 9, 1995
C
C
C
C
C *** BODY OF (FDIDEN) ***
C
C
C
C     Initialize.
      BIG=1.0D+30
      BIGNEW=1.0D+20
      FSBTOL=1.0D-8
      SMALLA=1.0D-8
      NFIXED=0
      NFREE=0
C
C
C
C
C     Zero  LENROW  and RMTMP1 arrays.
      DO 100 I=1,M
         LENROW(I)=0
         RMTMP1(I)=0.0D0
  100 CONTINUE
C
C     Count nonzero elements in all rows of  A.
      DO 300 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 300
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 200 K=KBEG,KEND
            IR=RWNMBS(K)
            LENROW(IR)=LENROW(IR)+1
  200    CONTINUE
  300 CONTINUE
C
C
C
C
C
C
C     Main loop begins here.
C     Loop over all structural columns of  A.
      DO 1000 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 1000
         IF(STAVAR(J).LT.0) GO TO 1000
         IF(LENCOL(J).EQ.0) GO TO 1000
C
C     Save nonzero elements of column J in RMTMP1 array.
C     Determine the shortest row with an entry in column J.
C     Equality-type rows are prefered if ties are to be broken.
         KROW=0
         LROW=NSTRCT+1
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 380 K=KBEG,KEND
            IR=RWNMBS(K)
            RMTMP1(IR)=ACOEFF(K)
            IF(LENROW(IR)-LROW) 360,340,380
  340       IF(RWSTAT(IR).NE.1) GO TO 380
  360       LROW=LENROW(IR)
            KROW=IR
  380    CONTINUE
         IF(KROW.EQ.0) GO TO 940
C
C     Analyse all columns that have entries in row KROW.
C     Look for a column with identical sparsity structure as column J.
         IPOS=RWHEAD(KROW)
  400    IF(IPOS.EQ.0) GO TO 810
            JCOL=CLNMBS(IPOS)
            IF(LENCOL(JCOL).NE.LENCOL(J)) GO TO 800
            IF(STAVAR(JCOL).LT.0) GO TO 800
            IF(STAVAR(JCOL).GE.6) GO TO 800
            IF(JCOL.LE.J) GO TO 800
            IF(JCOL.GT.NSTRCT) GO TO 800
C
C     Here if two columns J and JCOL have the same length.
            K2BEG=CLPNTS(JCOL)
            K2END=K2BEG+LENCOL(JCOL)-1
C
C
C
C     Check if columns J and JCOL differ with the sign only.
            IR=RWNMBS(K2BEG)
            ALPHA=RMTMP1(IR)/ACOEFF(K2BEG)
            DO 500 K2=K2BEG+1,K2END
               IR=RWNMBS(K2)
               DP=DABS(ACOEFF(K)+ACOEFF(K2))
               DP=DABS(RMTMP1(IR)/ACOEFF(K2)-ALPHA)
               IF(DP.GE.SMALLA) GO TO 800
  500       CONTINUE
            IF(DABS(ALPHA+1.0D0).GE.SMALLA) GO TO 600
C
C     Here if two columns J and JCOL differ with the sign only.
            IF(MSGLEV.LE.2) GO TO 505
            WRITE(BUFFER,501) CLNAME(J),CLNAME(JCOL)
  501       FORMAT(1X,'FDIDEN: LP variables:   ',
     X       A8,' and ',A8,' differ with the sign only.')
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,502) J,C(J),STAVAR(J)
  502       FORMAT(1X,'FDIDEN: var=',I6,' Cj=',D14.6,' stavar=',I6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,503) JCOL,C(JCOL),STAVAR(JCOL)
  503       FORMAT(1X,'    and var=',I6,' Cj=',D14.6,' stavar=',I6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,504) KROW,RWSTAT(KROW),LROW
  504       FORMAT(1X,'        row=',I5,'  rwstat=',I5,'  len=',I5)
            CALL MYWRT(IOLOG,BUFFER)
  505       CONTINUE
C
C
C     Columns J and JCOL differ with the sign only.
C     If the objective coefficients C1 and C2 differ with the sign
C     only and variables have no UPPER bound, then columns J and JCOL
C     are splitting of some FREE variable. If at least one of variables
C     have an UPPER bound, then they may be aggregated and replaced
C     with a single variable.
            DP=C(J)+C(JCOL)
            IF(DABS(DP).GE.SMALLA) GO TO 600
C
C     Here if the objective coefficients C1 and C2 differ with the sign.
            KSTAT=STAVAR(J)
            BNDU1=BIG
            IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) BNDU1=UPBND(J)
            KSTAT=STAVAR(JCOL)
            BNDU2=BIG
            IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) BNDU2=UPBND(JCOL)
            IF(BNDU1.LE.BIGNEW) GO TO 520
            IF(BNDU2.LE.BIGNEW) GO TO 520
C
C     Hidden FREE variable is found.
            NFREE=NFREE+1
            IF(MSGLEV.LE.1) GO TO 512
            WRITE(BUFFER,511) CLNAME(J),CLNAME(JCOL)
  511       FORMAT(1X,'FDIDEN: LP variables:   ',
     X       A8,' and ',A8,' are split FREE variable.')
            CALL MYWRT(IOLOG,BUFFER)
  512       CONTINUE
            STAVAR(J)=-JCOL
            STAVAR(JCOL)=-J
            GO TO 810
C
C     J2 becomes an aggregate column and J1 becomes FIXED.
  520       IF(BNDU1.LE.BNDU2) THEN
               J1=J
               J2=JCOL
            ELSE
               J1=JCOL
               J2=J
            ENDIF
C
C     Determine bounds and status of the aggregate variable.
            BNDL1=-UPBND(J1)
            BNDU1=UPBND(J1)+UPBND(J2)
            KSTAT=STAVAR(J2)
            STAVAR(J2)=0
            UPBND(J2)=BIG
            IF(DABS(BNDL1).LE.FSBTOL) THEN
               BNDL1=0.0D0
            ELSE
               K2BEG=CLPNTS(J2)
               K2END=K2BEG+LENCOL(J2)-1
               DO 540 K2=K2BEG,K2END
                  IR=RWNMBS(K2)
                  B(IR)=B(IR)-BNDL1*ACOEFF(K2)
                  IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
  540          CONTINUE
               STAVAR(J2)=2
            ENDIF
            IF(KSTAT.LE.1) THEN
               LOBND(J2)=BNDL1
            ELSE
               STAVAR(J2)=2
               LOBND(J2)=LOBND(J2)+BNDL1
            ENDIF
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
            IF(LNHIST.GE.MXHIST) GO TO 9200
            LNHIST=LNHIST+1
            INHIST(LNHIST)=-J2
            DPHIST(LNHIST)=BNDL1
C
            IF(BNDU1.LE.BIGNEW) THEN
               STAVAR(J2)=STAVAR(J2)+1
               UPBND(J2)=BNDU1
C
C     Reinitialize bounds on shadow prices.
                  KBEG0=CLPNTS(J2)
                  KEND0=KBEG0+LENCOL(J2)-1
                  DO 560 IKX=KBEG0,KEND0
                     IR=RWNMBS(IKX)
                     P(IR)=-BIG
                     Q(IR)=BIG
                     IF(RANGES(IR).LE.BIGNEW) GO TO 560
                     IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                     IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
  560             CONTINUE
            ENDIF
            IF(MSGLEV.LE.1) GO TO 562
            WRITE(BUFFER,561) CLNAME(J2),CLNAME(J1)
  561       FORMAT(1X,'FDIDEN: Variable=',A8,
     X       ' becomes a diff. of variable=',A8,' and itself.')
            CALL MYWRT(IOLOG,BUFFER)
  562       CONTINUE
            GO TO 720
C
C
C
C     Check if columns J and JCOL are identical.
  600       IF(DABS(ALPHA-1.0D0).GE.SMALLA) GO TO 800
C
C     Here if two columns J and JCOL are identical.
            IF(MSGLEV.LE.2) GO TO 705
            WRITE(BUFFER,701) CLNAME(J),CLNAME(JCOL)
  701       FORMAT(1X,'FDIDEN: LP variables:   ',
     X       A8,' and ',A8,' are identical.')
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,702) J,C(J),STAVAR(J)
  702       FORMAT(1X,'FDIDEN: var=',I6,' Cj=',D14.6,' stavar=',I6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,703) JCOL,C(JCOL),STAVAR(JCOL)
  703       FORMAT(1X,'    and var=',I6,' Cj=',D14.6,' stavar=',I6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,704) KROW,RWSTAT(KROW),LROW
  704       FORMAT(1X,'        row=',I5,'  rwstat=',I5,'  len=',I5)
            CALL MYWRT(IOLOG,BUFFER)
  705       CONTINUE
C
C
C     Identical columns are found.
C     If the objective coefficients C1 and C2 are the same, then column
C     aggregation is possible. If they are different, however, then it
C     may be possible to fix one of variables.
            DP=C(J)-C(JCOL)
            IF(DABS(DP).LE.SMALLA) THEN
C
C     Here if the objective coefficients C1 and C2 are the same.
               KSTAT=STAVAR(J)
               BNDU1=BIG
               IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) BNDU1=UPBND(J)
               KSTAT=STAVAR(JCOL)
               BNDU2=BIG
               IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) BNDU2=UPBND(JCOL)
C
C     J2 becomes an aggregate column and J1 becomes FIXED.
               J1=JCOL
               J2=J
C
C     Determine an UPPER bound and status of the aggregate variable.
               STAVAR(J2)=0
               UPBND(J2)=BIG
               BNDL1=LOBND(J2)
               IF(DABS(BNDL1).LE.FSBTOL) THEN
                  BNDL1=0.0D0
               ELSE
                  STAVAR(J2)=2
               ENDIF
               BNDU1=BNDU1+BNDU2
               IF(BNDU1.LE.BIGNEW) THEN
                  STAVAR(J2)=STAVAR(J2)+1
                  UPBND(J2)=BNDU1
C
C     Reinitialize bounds on shadow prices.
                  KBEG0=CLPNTS(J2)
                  KEND0=KBEG0+LENCOL(J2)-1
                  DO 710 IKX=KBEG0,KEND0
                     IR=RWNMBS(IKX)
                     P(IR)=-BIG
                     Q(IR)=BIG
                     IF(RANGES(IR).LE.BIGNEW) GO TO 710
                     IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                     IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
  710             CONTINUE
               ENDIF
               IF(MSGLEV.LE.1) GO TO 712
               WRITE(BUFFER,711) CLNAME(J2),CLNAME(J1)
  711          FORMAT(1X,'FDIDEN: Variable=',A8,
     X          ' becomes a sum of variable=',A8,' and itself.')
               CALL MYWRT(IOLOG,BUFFER)
  712          CONTINUE
               GO TO 720
            ENDIF  
C
C     Here if the objective coefficients C1 and C2 are different.
            IF(DP.GE.SMALLA) THEN
               J1=J
               J2=JCOL
            ENDIF
            IF(DP.LE.-SMALLA) THEN
               J1=JCOL
               J2=J
            ENDIF
            KSTAT=STAVAR(J2)
            IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) GO TO 800
C
C     FIX variable J1 on its LOWER bound.
  720       X0=0.0D0
            IF(MSGLEV.LE.1) GO TO 722
            WRITE(BUFFER,721) J1,CLNAME(J1),X0
  721       FORMAT(1X,'FDIDEN: Variable ',I6,' (name=',A8,
     X       ') is being FIXED on X=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
  722       CONTINUE
            IF(STAVAR(J1).NE.1.AND.STAVAR(J1).NE.3) THEN
C
C     Reinitialize bounds on shadow prices.
C              WRITE(BUFFER,741) J1,CLNAME(J1),STAVAR(J1)
C 741          FORMAT(1X,'FDIDEN: Variable ',I6,' (name=',A8,
C    X          ' st=',I6,')')
C              CALL MYWRT(IOLOG,BUFFER)
               KBEG0=CLPNTS(J1)
               KEND0=KBEG0+LENCOL(J1)-1
               DO 740 IKX=KBEG0,KEND0
                  IR=RWNMBS(IKX)
                  P(IR)=-BIG
                  Q(IR)=BIG
                  IF(RANGES(IR).LE.BIGNEW) GO TO 740
                  IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                  IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
  740          CONTINUE
            ENDIF
            NFIXED=NFIXED+1
            PRLVAR(J1)=X0
            STAVAR(J1)=6
            IF(J1.EQ.J) GO TO 810
            IF(STAVAR(J).LT.0) GO TO 810
C
  800    IPOS=RWLINK(IPOS)
         GO TO 400
  810    CONTINUE
C
C
C     Restore zero value of  RMTMP1 array.
  940    DO 960 K=KBEG,KEND
            IR=RWNMBS(K)
            RMTMP1(IR)=0.0D0
  960    CONTINUE
C
C
C
C     End of main loop.
 1000 CONTINUE
C
C
C
C
C
C
C     Check if RMTMP1 array is zero.
      DO 1100 I=1,M
         IF(DABS(RMTMP1(I)).GE.SMALLA) THEN
            WRITE(BUFFER,1101) I,RMTMP1(I)
 1101       FORMAT(1X,'FDIDEN ERROR: RMTMP1(',I6,')=',D14.8)
            CALL ERRWRT(IOLOG,BUFFER)
            CALL MYSTOP(94)
         ENDIF
 1100 CONTINUE
C
C
C
C
C
C
C     Here if a successful run has been completed.
      IF(MSGLEV.LE.0) GO TO 1110
      WRITE(BUFFER,1105) NFIXED
 1105 FORMAT(1X,'FDIDEN: Variables eliminated:   ',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,1106) NFREE
 1106 FORMAT(1X,'        Split FREE variables:   ',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
 1110 CONTINUE
C
C
C
C
C
C     Zero  RWHEAD and LENROW arrays.
      IF(NFIXED.EQ.0) GO TO 1400
      DO 1200 I=1,M
         RWHEAD(I)=0
         LENROW(I)=0
 1200 CONTINUE
C
C     Set the row linked lists.
C     Count nonzero elements in all rows of  A.
      DO 1300 J=1,N
C
C     Omit  FIXED variables.
         IF(STAVAR(J).GE.6) GO TO 1300
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 1250 K=KBEG,KEND
            I=RWNMBS(K)
            RWLINK(K)=RWHEAD(I)
            CLNMBS(K)=J
            RWHEAD(I)=K
            LENROW(I)=LENROW(I)+1
 1250    CONTINUE
 1300 CONTINUE
 1400 CONTINUE
C
C
C
C
C
C
      RETURN
C
C
 9200 WRITE(BUFFER,9201)
 9201 FORMAT(1X,'FDIDEN: Please increase space for PRE_SOLVE ',
     X 'history list.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(6)
C
C
C
C *** LAST CARD OF (FDIDEN) ***
      END
