C********************************************************
C     **** FSATY ... FAST (sparse)Atransp * (dense)Y ****
C********************************************************
C
      SUBROUTINE FSATY(MAXM,MAXN,MAXNZA,Y,M,X,N,
     X ACOEFF,CLPNTS,RWNMBS,LENCOL,VUSED,IOLOG)
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,M,N,IOLOG
      DOUBLE PRECISION X(N),Y(M),ACOEFF(MAXNZA)
      INTEGER*4 CLPNTS(MAXN+1)
      INTEGER*2 RWNMBS(MAXNZA),LENCOL(MAXN)
      LOGICAL VUSED(MAXN)
C
C *** LOCAL VARIABLES
      INTEGER*4 IR,J,K,KBEG,KEND
C
C
C *** PURPOSE
C     This routine computes the product of a sparse matrix  Atransp
C     and a dense vector Y and saves the result in a dense vector X.
C     It is given direct access to matrix  A.
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix.
C     N       Number of columns of the LP constraint matrix.
C     Y       Dense vector of dimension M.
C     ACOEFF  Nonzero elements of matrix A.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     VUSED   An indicator if a variable is active in the optimization
C             process:
C             .TRUE.   active variable;
C             .FALSE.  FIXED variable.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     X       Dense vector of dimension N (X = Atransp * Y).
C
C
C *** SUBROUTINES CALLED:
C     NONE
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: May 17, 1995
C
C
C *** BODY OF (FSATY) ***
C
      DO 200 J=1,N
         X(J)=0.0
         IF(.NOT.VUSED(J)) GO TO 200
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 100 K=KBEG,KEND
            IR=RWNMBS(K)
            X(J)=X(J)+Y(IR)*ACOEFF(K)
  100    CONTINUE
  200 CONTINUE
      RETURN
C
C *** LAST CARD OF (FSATY) ***
      END
