C**************************************************************
C     ***  GETDIM ... DETERMINE CURRENT PROBLEM DIMENSIONS  ***
C**************************************************************
C
      SUBROUTINE GETDIM(IOLOG,
     X MAXM,MAXN,M,N,NSTRCT,
     X M1,N1,NZ1,B,C,LENCOL,STAVAR)
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MAXM,MAXN,M,N,NSTRCT,M1,N1,NZ1
      DOUBLE PRECISION B(MAXM),C(MAXN)
      INTEGER*2 LENCOL(MAXN),STAVAR(MAXN)
C
C *** LOCAL VARIABLES
      INTEGER*4 J,K
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     B       Right hand side of the linear program.
C     C       Objective function coefficients.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C
C *** ON OUTPUT:
C     M1      Number of constraints.
C     N1      Number of structural variables.
C     MZ1     Number of nonzeros of the LP constraint matrix.
C
C
C
C
C *** PURPOSE
C     This routine determines current dimension of the problem.
C
C *** SUBROUTINES CALLED
C
C *** NOTES
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior
C        to applying an interior point method, Technical Report
C        No 1994.3, Department of Management Studies, University
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland,
C        February 1994, revised in December 1994.
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (GETDIM) ***
C
C     Determine current dimensions of the problem.
      M1=M
      N1=0
      NZ1=0
      DO 100 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 100
         IF(STAVAR(J).LT.0) THEN
            K=-STAVAR(J)
            IF(J.GE.K) GO TO 100
         ENDIF
         N1=N1+1
         NZ1=NZ1+LENCOL(J)
  100 CONTINUE
C
      RETURN
C
C *** LAST CARD OF (GETDIM) ***
      END
