C***************************************************************
C     **** LDAAT ... LOAD A*Atransp INTO CHOLESKY FACTOR ****
C***************************************************************
C
      SUBROUTINE LDAAT(ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X THETA,STAVAR,MAXNZL,MAXM,MAXN,MAXNZA,M,
     X LCOEFF,LCLPTS,LRWNBS,LDIAG,DPWORK,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,MAXN,MAXNZA,M,IOLOG
      DOUBLE PRECISION ACOEFF(MAXNZA),THETA(MAXN)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM),RWLINK(MAXNZA)
      DOUBLE PRECISION LCOEFF(*),LDIAG(MAXM),DPWORK(MAXM)
      INTEGER*4 LCLPTS(MAXM+1)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      INTEGER*2 STAVAR(MAXN),LRWNBS(MAXNZL)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IR,J,JX,JCOL,JBEG,JEND,K,KX,KBEG,KEND
      DOUBLE PRECISION DP
      CHARACTER*100 BUFFER
C
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     ACOEFF  Nonzero entries of an  LP constraint matrix.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of entries of matrix A.
C     RWLINK  Row linked lists of entries of matrix A.
C     CLNMBS  Column numbers of nonzeros in a given row of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     THETA   Diagonal weight matrix.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicate the position of the original variable.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     LCOEFF  Off-diagonal nonzero coefficients of  A*THETA*Atransp
C             matrix (fill-in positions are zeroed).
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of Cholesky factor.
C
C     WORK ARRAYS:
C     DPWORK  Array for temporary storage of a single column
C             of  A*THETA*Atransp matrix.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT
C
C
C *** PURPOSE:
C     This routine constructs the  A*THETA*Atransp matrix
C     and places it in a data structure for its Cholesky factor.
C     All fill-in positions of this structure are zeroed.
C
C
C *** NOTES:
C     1. Matrices  ACOEFF, CLPNTS and RWNMBS handle the  LP
C        constraint matrix as a collection of columns.
C     2. Matrices  ACOEFF, RWHEAD, RWLINK and CLNMBS handle
C        matrix  A in a form of the row linked lists.
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapters  2 and  10.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 19, 1991
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (LDAAT) ***
C
C
C
C *** DEBUGGING
C     WRITE(BUFFER,51)
C  51 FORMAT(1X,'LDAAT: Constraint matrix')
C     CALL MYWRT(IOLOG,BUFFER)
C     DO 58 JCOL=1,N
C        KBEG=CLPNTS(JCOL)
C        KEND=KBEG+LENCOL(JCOL)-1
C        WRITE(BUFFER,55) JCOL
C  55    FORMAT(1X,'LDAAT: Column ',I5)
C        CALL MYWRT(IOLOG,BUFFER)
C        DO 57 K=KBEG,KEND
C           WRITE(BUFFER,56) K,RWNMBS(K),ACOEFF(K)
C  56       FORMAT(1X,'LDAAT: K=',I4,' row=',I4,' elt=',1PD20.12)
C           CALL MYWRT(IOLOG,BUFFER)
C  57 CONTINUE
C  58 CONTINUE
C     DO 70 JCOL=1,N
C        IF(STAVAR(JCOL).NE.6) THEN
C           WRITE(BUFFER,68) JCOL,THETA(JCOL)
C  68       FORMAT(1X,'LDAAT: THETA(',I4,')=',D12.5)
C           CALL MYWRT(IOLOG,BUFFER)
C        ENDIF
C  70 CONTINUE
C
C
C
C
C
C     Main loop begins here (loop over columns of A*THETA*Atransp).
      DO 1000 J=1,M
C
C
C     Zero all the elements of  DPWORK array in the sparsity
C     pattern of the appropriate column of  Cholesky factor.
         JBEG=LCLPTS(J)
         JEND=LCLPTS(J+1)-1
         DO 200 JX=JBEG,JEND
            IR=LRWNBS(JX)
            DPWORK(IR)=0.0
  200    CONTINUE
C
C
C     Scan row  J of matrix  A (column  J of  Atransp).
         LDIAG(J)=0.0
         K=RWHEAD(J)
  300    IF(K.EQ.0) GO TO 500
         JCOL=CLNMBS(K)
         DP=THETA(JCOL)*ACOEFF(K)
         LDIAG(J)=LDIAG(J)+DP*ACOEFF(K)
C
C     Initialize loop over nonzeros of a column. The whole column
C     have to be analysed if its nonzeros are not in increasing
C     order of row numbers.
C        KBEG=CLPNTS(JCOL)
C        KEND=KBEG+LENCOL(JCOL)-1
         KBEG=K+1
         KEND=CLPNTS(JCOL)+LENCOL(JCOL)-1
         DO 400 KX=KBEG,KEND
            IR=RWNMBS(KX)
            DPWORK(IR)=DPWORK(IR)+DP*ACOEFF(KX)
  400    CONTINUE
C
C
C     Continue the scanning of  J-th row of matrix  A.
  450    K=RWLINK(K)
         GO TO 300
C
C
C     Here when the column building is completed.
C     Copy the column to the data structures for Cholesky factor.
C     Note that fill-in positions contain zeros.
  500    CONTINUE
C
C
C *** DEBUGGING
C     WRITE(BUFFER,451) J
C 451 FORMAT(1X,'LDAAT: Cholesky matrix, column ',I5)
C     CALL MYWRT(IOLOG,BUFFER)
C     DO 457 JX=JBEG,JEND
C        IR=LRWNBS(JX)
C        WRITE(BUFFER,456) JX,LRWNBS(JX),DPWORK(IR)
C 456    FORMAT(1X,'LDAAT: JX=',I4,' rw=',I4,' elt=',1PD20.12)
C        CALL MYWRT(IOLOG,BUFFER)
C 457 CONTINUE
C
C
         DO 600 JX=JBEG,JEND
            IR=LRWNBS(JX)
            LCOEFF(JX)=DPWORK(IR)
  600    CONTINUE
C
C
C
C
C
C     End of main loop.
 1000 CONTINUE
C
C
C
      RETURN
C
C *** LAST CARD OF (LDAAT) ***
      END
