C************************************************
C     ***  MKSPAR ... MAKE MATRIX  A SPARSER  ***
C************************************************
C
      SUBROUTINE MKSPAR(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X M1,N1,NZ1,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,
     X B,RANGES,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X P,Q,STAVAR,RWSTAT,STAROW,RWNAME,
     X MARKER,LENROW,LIST,LACTIV,ACTIVE)
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,MAXM,MAXN,MAXNZA,M,N,NSTRCT,M1,N1,NZ1
      INTEGER*4 IROW(MAXN),IMTMP1(MAXM),INTMP1(MAXN)
      INTEGER*2 INTMP2(MAXN)
      DOUBLE PRECISION RELT(MAXN),ACOEFF(MAXNZA),B(MAXM),RANGES(MAXM)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      CHARACTER*8 RWNAME(MAXM)
      DOUBLE PRECISION P(MAXM),Q(MAXM)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      INTEGER*2 MARKER(MAXM),LENROW(MAXM)
      INTEGER*2 LIST(MAXM),LACTIV(MAXM),ACTIVE(MAXM)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 COLLEN,MNEW,NCOMN,FIRST,LAST,LENOK,NZELIM
      INTEGER*4 I,IPOS,JPIVOT,IR,IRUN,J,K,IKX,KOK,KOUT,NZEL0
      INTEGER*4 JSHORT,JLONG,KSHORT,KLONG,KSHBEG,KSHEND,LSHORT
      DOUBLE PRECISION DP,PIVOT,AELT,BELT
      DOUBLE PRECISION BIG,SMALLA,GROWTH
      CHARACTER*100 BUFFER
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     RANGES  Array of constraint ranges.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWNAME  Array of row names (increasing order sort).
C
C *** ON OUTPUT:
C
C
C
C
C *** WORK ARRAYS:
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     IMTMP1  Integer work array of size MAXM.
C     INTMP1  Integer work array of size MAXN
C     INTMP2  Half-length integer work array of size MAXN.
C     MARKER  Half-length integer work array of size MAXM.
C     LENROW  Half-length integer work array of size MAXM.
C     LIST    Half-length integer work array of size MAXM.
C     LACTIV  Half-length integer work array of size MAXM.
C     ACTIVE  Half-length integer work array of size MAXM.
C
C
C
C
C *** PURPOSE
C     This routine finds EQUALITY type LP constraints with
C     the sparsity structure being the subset of the other
C     constraint. Shorter row of A is then used to eliminate
C     nonzero entries from the longer one.
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,GETCOL,GETROW,DABS,EMPTYR,REORDA
C
C
C *** NOTES
C     This routine is given direct access to the matrix A.
C     It alters hidden data structures.
C
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: May 20, 1995
C
C
C
C
C *** BODY OF (MKSPAR) ***
C
C
C
C     Initialize.
      BIG=1.0D+30
      SMALLA=1.0D-10
      GROWTH=1.0D+3
      IRUN=1
      NZELIM=0
C
C
C
C
C     Count nonzero elements of all LP constraints.
C     Zero INTMP1 array (it will be used to store sparsity
C     pattern of the short EQUALITY type constraint).
      DO 20 I=1,M
         LENROW(I)=0
   20 CONTINUE
      DO 60 J=1,N
         INTMP1(J)=0
         IF(STAVAR(J).GE.6) GO TO 60
         KSHBEG=CLPNTS(J)
         KSHEND=KSHBEG+LENCOL(J)-1
         DO 40 K=KSHBEG,KSHEND
            I=RWNMBS(K)
            LENROW(I)=LENROW(I)+1
   40    CONTINUE
   60 CONTINUE
C
C
C
C     Build linked list of equality type  LP constraints with
C     the same lengths. Initialize ACTIVE array.
      LENOK=NSTRCT+1
  100 DO 120 J=1,NSTRCT+1
         INTMP2(J)=0
  120 CONTINUE
      DO 140 I=M,1,-1
         ACTIVE(I)=0
         IF(RWSTAT(I).NE.1) GO TO 140
         ACTIVE(I)=1
         K=LENROW(I)
         LIST(I)=INTMP2(K)
         INTMP2(K)=I
  140 CONTINUE
C
C     Build a list of equality type rows ordered with increasing
C     number of nonzero entries.
      LAST=0
      DO 180 J=1,LENOK
         I=INTMP2(J)
         IF(I.EQ.0) GO TO 180
  160    LAST=LAST+1
         LACTIV(LAST)=I
C        WRITE(BUFFER,161) I,RWSTAT(I),J
C 161    FORMAT(1X,'row=',I6,'  rwstat=',I6,'  length=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
         I=LIST(I)
         IF(I.GT.0) GO TO 160
  180 CONTINUE
C
C
C
C
C
C
C     Main loop begins here.
C     Look for EQUALITY type constraints with at least 2 entries.
      FIRST=1
  200 IF(LAST.LT.FIRST) GO TO 1000
      I=LACTIV(FIRST)
C
C     Pack the sparsity pattern of row I into INTMP1 array (omit
C     already eliminated entries). Save the number of the shortest
C     column as JSHORT and the number of the longest column as JLONG.
C     Save the largest elt of this row as PIVOT and the corresponding
C     column number as JPIVOT.
      K=RWHEAD(I)
      JSHORT=0
      LSHORT=M+1
      JLONG=0
      COLLEN=0
      JPIVOT=0
      PIVOT=0.0D0
  320 IF(K.EQ.0) GO TO 400
      J=CLNMBS(K)
      IF(J.LE.0) GO TO 340
      INTMP1(J)=K
      IF(LENCOL(J).LT.LSHORT) THEN
         JSHORT=J
         LSHORT=LENCOL(J)
      ENDIF
      IF(LENCOL(J).GT.COLLEN) THEN
         JLONG=J
         COLLEN=LENCOL(J)
      ENDIF
      IF(DABS(ACOEFF(K)).GT.PIVOT) THEN
         JPIVOT=J
         PIVOT=DABS(ACOEFF(K))
      ENDIF
  340 K=RWLINK(K)
      GO TO 320
C
C
  400 CONTINUE
C
C *** DEBUGGING
      IF(MSGLEV.LE.2) GO TO 410
      WRITE(BUFFER,401) 
  401 FORMAT(1X)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,402) I,LENROW(I),JPIVOT,PIVOT
  402 FORMAT(1X,'row=',I6,'  ln=',I6,'  Jpiv=',I6,'  piv=',D10.3)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,403) JSHORT,LSHORT,JLONG,COLLEN
  403 FORMAT(1X,'short=',I6,'  ln=',I6,'    long=',I6,'  ln=',I6)
      CALL MYWRT(IOLOG,BUFFER)
  410 CONTINUE
C
C
C     Scan short column. Check if any row crossing it contains row I.
C     There is no chance for elimination, if a short column has length 1.
C     Omit already emptied rows (by the unexpected cancellations).
      IF(LSHORT.EQ.1) GO TO 900
      IF(JSHORT.EQ.0) GO TO 900
      KSHBEG=CLPNTS(JSHORT)
      KSHEND=KSHBEG+LENCOL(JSHORT)-1
      DO 800 K=KSHBEG,KSHEND
         IR=RWNMBS(K)
         IF(IR.EQ.I) GO TO 800
         IF(IR.LE.0) GO TO 800
         IF(LENROW(IR).LT.LENROW(I)) GO TO 800
C
C     Scan row IR. Count the number of common entries with row I.
C     Pack the sparsity pattern of row IR into IROW array (omit
C     already eliminated entries).
         NCOMN=0
         KLONG=RWHEAD(IR)
  420    IF(KLONG.EQ.0) GO TO 500
         J=CLNMBS(KLONG)
         IF(J.LE.0) GO TO 440
         IROW(J)=KLONG
         IF(INTMP1(J).GT.0) NCOMN=NCOMN+1
  440    KLONG=RWLINK(KLONG)
         GO TO 420
C
C     Check if the sparsity pattern of row I is a subset of that
C     of row IR.
  500    CONTINUE
         IF(MSGLEV.LE.2) GO TO 502
         WRITE(BUFFER,501) IR,LENROW(IR),NCOMN
  501    FORMAT(1X,'row=',I6,'  len=',I6,'  NCOMN=',I6)
         CALL MYWRT(IOLOG,BUFFER)
  502    CONTINUE
         IF(NCOMN.LT.LENROW(I)) GO TO 800
C
C
C     Perform the elimination. Recall, that INTMP1 and IROW arrays
C     remember sparsity patterns of rows I and IR, respectively.
C     Compute Gaussian elementary operator that eliminates nonzero
C     entry from the longest column. If it is acceptable, then use it.
C     Otherwise use the PIVOT element (the largest entry of row I)
C     to define the operator.
         NZEL0=NZELIM
         KSHORT=INTMP1(JLONG)
         KLONG=IROW(JLONG)
         DP=ACOEFF(KLONG)/ACOEFF(KSHORT)
         IF(DABS(DP).LE.GROWTH) GO TO 600
         KSHORT=INTMP1(JPIVOT)
         KLONG=IROW(JPIVOT)
         DP=ACOEFF(KLONG)/ACOEFF(KSHORT)
C
C     Perform elimination:  ROW(ir) := ROW(ir) - DP * ROW(i)
  600    CONTINUE
         IF(MSGLEV.LE.2) GO TO 602
         WRITE(BUFFER,601) KSHORT,KLONG,DP
  601    FORMAT(1X,'KSHORT=',I6,' KLONG=',I6,'  M_entry=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
  602    CONTINUE
         IF(DABS(DP).GE.GROWTH) GO TO 800
C
C
C     Perform elimination:  ROW(ir) := ROW(ir) - DP * ROW(i)
C     Loop over nonzero entries of row I.
         KSHORT=RWHEAD(I)
  620    IF(KSHORT.EQ.0) GO TO 700
         J=CLNMBS(KSHORT)
C        WRITE(BUFFER,621) KSHORT,I,CLNMBS(KSHORT),ACOEFF(KSHORT)
C 621    FORMAT(1X,'Kshort=',I5,' rw=',I6,' cl=',I6,'  elt=',D10.3)
C        CALL MYWRT(IOLOG,BUFFER)
         IF(J.LE.0) GO TO 680
         KLONG=IROW(J)
C        WRITE(BUFFER,622) KLONG,RWNMBS(KLONG),CLNMBS(KLONG),
C    X    ACOEFF(KLONG)
C 622    FORMAT(1X,'Klong=',I6,' rw=',I6,' cl=',I6,'  elt=',D10.3)
C        CALL MYWRT(IOLOG,BUFFER)
C
C     Update nonzero entry of the longer row.
         AELT=ACOEFF(KLONG)-DP*ACOEFF(KSHORT)
         ACOEFF(KLONG)=AELT
         IF(DABS(AELT).LE.SMALLA) THEN
            NZELIM=NZELIM+1
            LENROW(IR)=LENROW(IR)-1
            CLNMBS(KLONG)=-CLNMBS(KLONG)
            RWNMBS(KLONG)=-RWNMBS(KLONG)
            IF(MSGLEV.LE.2) GO TO 642
            WRITE(BUFFER,641) IR
  641       FORMAT(1X,'MKSPAR: Entry removed from row=',I6)
            CALL MYWRT(IOLOG,BUFFER)
  642       CONTINUE
         ENDIF
  680    KSHORT=RWLINK(KSHORT)
         GO TO 620
C
C     Update RHS.
  700    BELT=B(IR)-DP*B(I)
         IF(DABS(BELT).LE.SMALLA) BELT=0.0D0
         B(IR)=BELT
C
C     Update bounds on shadow prices.
         IF(DP.LE.0.0D0) THEN
            P(I)=P(I)+DP*Q(IR)
            Q(I)=Q(I)+DP*P(IR)
         ELSE
            P(I)=P(I)+DP*P(IR)
            Q(I)=Q(I)+DP*Q(IR)
         ENDIF
C
C     Add row IR to the list of active rows.
         IF(RWSTAT(IR).NE.1) GO TO 800
         IF(ACTIVE(IR).EQ.1) GO TO 800
         IF(NZEL0.EQ.NZELIM) GO TO 800
         IF(LAST.EQ.M) THEN
C
C     Move the list to the beginning of LACTIV array.
            DO 720 IKX=1,LAST-FIRST+1
               LACTIV(IKX)=LACTIV(IKX+FIRST-1)
  720       CONTINUE
            LAST=LAST-FIRST+1
            FIRST=1
         ENDIF
         LAST=LAST+1
         ACTIVE(IR)=1
         LACTIV(LAST)=IR
C
C     Save length of the row used in the elimination.
         LENOK=LENROW(I)
C
  800 CONTINUE
C
C
C     Zero INTMP1 array.
  900 K=RWHEAD(I)
  920 IF(K.EQ.0) GO TO 960
      J=CLNMBS(K)
      IF(J.LE.0) GO TO 940
      INTMP1(J)=0
  940 K=RWLINK(K)
      GO TO 920
C
C     Eliminate row I from the list of active rows.
  960 CONTINUE
      ACTIVE(I)=0
      FIRST=FIRST+1
C
C
C
C
C
C
C     End of main loop.
      GO TO 200
 1000 CONTINUE
C
C
C
C
C
C
C     Check if MAKE_SPARSE heuristic produced considerable reduction
C     (at least 50%) of nonzero entries in A. If so, then repeat it
C     for all rows of length at most LENOK.
      IF(IRUN.EQ.2) GO TO 1020
      IF(NZELIM.LE.100) GO TO 1020
      IF(2*NZELIM.LE.NZ1) GO TO 1020
C
C     Repeat MAKE_SPARSE heuristic.
      IRUN=2
      IF(MSGLEV.LE.0) GO TO 1012
      WRITE(BUFFER,1011) NZELIM
 1011 FORMAT(1X,'MKSPAR: First pass, nonz. elim: ',I9)
      CALL MYWRT(IOLOG,BUFFER)
 1012 CONTINUE
      GO TO 100
C
C
C
C
C
C
C     Here if a successful run of the loop has been completed.
 1020 IF(MSGLEV.LE.0) GO TO 1030
      WRITE(BUFFER,1021) NZELIM
 1021 FORMAT(1X,'MKSPAR: Nonzeros eliminated:    ',I9)
      CALL MYWRT(IOLOG,BUFFER)
 1030 CONTINUE
C
C
C
C
C
C
      IF(NZELIM.GT.0) THEN
C
C     Reorder elements within each column of the  LP constraint
C     matrix in such a way that those of the active part of  A
C     are at the beginning of the lists. The column lengths will
C     later be decreased to forget inactive part of matrix  A.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD array.
         DO 1200 I=1,M
            RWHEAD(I)=0
 1200    CONTINUE
C
C     Reorder nonzero elements within each column.
         DO 1500 J=1,N
            IF(STAVAR(J).GE.6) GO TO 1500
            KSHBEG=CLPNTS(J)
            KSHEND=KSHBEG+LENCOL(J)-1
            KOK=0
            KOUT=0
            DO 1300 K=KSHBEG,KSHEND
               I=RWNMBS(K)
               IF(I.GT.0) THEN
                  KOK=KOK+1
                  IROW(KOK)=RWNMBS(K)
                  RELT(KOK)=ACOEFF(K)
               ELSE
                  IPOS=LENCOL(J)-KOUT
                  KOUT=KOUT+1
                  IROW(IPOS)=RWNMBS(K)
                  RELT(IPOS)=ACOEFF(K)
               ENDIF
 1300       CONTINUE
C
C     Save only active part of the column.
C     Set the row linked lists.
            KSHBEG=CLPNTS(J)-1
            DO 1400 IKX=1,KOK
               K=KSHBEG+IKX
               I=IROW(IKX)
               RWNMBS(K)=I
               CLNMBS(K)=J
               ACOEFF(K)=RELT(IKX)
               RWLINK(K)=RWHEAD(I)
               RWHEAD(I)=K
 1400       CONTINUE
            LENCOL(J)=KOK
 1500    CONTINUE
C
C     Determine the permutation that puts all empty and inactive
C     rows at the end of the list.
C
         IRUN=3
         IF(MSGLEV.LE.1) IRUN=4
         CALL EMPTYR(MAXM,M,MNEW,IRUN,
     X    RWHEAD,STAROW,MARKER,LENROW,IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the analysis of EMPTYR.
         IF(MNEW.LT.M) THEN
C
            CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X       CLPNTS,RWNMBS,
     X       RWHEAD,RWLINK,CLNMBS,LENCOL,
     X       MARKER,LENROW,IMTMP1,IROW,RELT,
     X       RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C     Reorder bounds on shadow prices P and Q.
         CALL REORDV(MAXM,M,
     X    MARKER,LENROW,P,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    MARKER,LENROW,Q,RELT,IOLOG)
C
C     Set the new number of rows of the constraint matrix.
            M=MNEW
C
         ENDIF
      ENDIF
C
C
C
C
C
C
C     Determine current dimensions of the problem.
      M1=M
      N1=0
      NZ1=0
      DO 1600 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 1600
         IF(STAVAR(J).LT.0) THEN
            K=-STAVAR(J)
            IF(J.GE.K) GO TO 1600
         ENDIF
         N1=N1+1
         NZ1=NZ1+LENCOL(J)
 1600 CONTINUE
C
C
C
C
C
      RETURN
C
C
C
C *** LAST CARD OF (MKSPAR) ***
      END
