C*****************************************************************
C     *** PCCHCK ... CHECK FEASIBILITY OF THE CURRENT COLUTION ***
C     ***            PREDICTOR-CORRECTOR PRIMAL-DUAL METHOD    ***
C*****************************************************************
C
      SUBROUTINE PCCHCK(MAXM,MAXN,M,N,IOLOG,
     X MSGLEV,PRMAXB,PRMAXU,DLMAXC,
     X VUSED,VBNDED,XIB,XIC,XIU)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,M,N,IOLOG,MSGLEV
      DOUBLE PRECISION PRMAXB,PRMAXU,DLMAXC
      LOGICAL VUSED(MAXN),VBNDED(MAXN)
      DOUBLE PRECISION XIB(M),XIC(N),XIU(N)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,J
      DOUBLE PRECISION DP
      CHARACTER*100 BUFFER
C
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     N       Number of columns of the LP constraint matrix.
C     VUSED   An indicator if a variable is active in the optimization
C             process:
C             .TRUE.   active variable;
C             .FALSE.  FIXED variable.
C     VBNDED  An indicator if a variable has an UPPER bound:
C             .TRUE.   UPPER bounded variable;
C             .FALSE.  UNBOUNDED variable;
C     XIB     Violation of primal constraints, i.e.  b - A * x
C     XIC     Violation of dual   constraints, i.e.  c - At*y - z + w
C     XIU     Violation of variable bounds, i.e.     UPBND - x - s
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  if 1, then output is printed.
C     ON OUTPUT:
C     ERRB    max{|xib(i)|:i=1,...,m}.
C     ERRU    max{|xiu(j)|:i=1,...,n}.
C     ERRC    max{|xic(j)|:j=1,...,n}.
C
C
C *** LOCAL VARIABLES DESCRIPTION
C
C
C *** SUBROUTINES CALLED:
C     DABS
C
C
C *** PURPOSE:
C     This routine checks the primal and dual feasibility
C     of the current solution.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio,
C                    Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: May 17, 1995
C
C
C *** BODY OF (PCCHCK) ***
C
C
C     PRMAXB=max{|xib(i)|:i=1,...,m}
C     XIB: = b - A * x
C
      PRMAXB=0d0
      DO 100 I=1,M
         DP=DABS(XIB(I))
         IF(DP.GT.PRMAXB) PRMAXB=DP
  100 CONTINUE
C
C
C     PRMAXU=max{|xiu(j)|:j=1,...,n}
C     XIU: = UPBND - x - s
C
C     DLMAXC=max{|xic(j)|:j=1,...,n}
C     XIC: = c - At*y - z + w
C
      PRMAXU=0d0
      DLMAXC=0d0
      DO 200 J=1,N
         IF(VUSED(J)) THEN
            DP=DABS(XIC(J))
            IF(DP.GT.DLMAXC) DLMAXC=DP
            IF(VBNDED(J)) THEN
               DP=DABS(XIU(J))
               IF(DP.GT.PRMAXU) PRMAXU=DP
            ENDIF
         ENDIF
  200 CONTINUE
C
C
      IF(MSGLEV.EQ.1) THEN
         WRITE(BUFFER,201) PRMAXB,PRMAXU,DLMAXC
  201    FORMAT(1X,'PCCHCK: ||A*x-b||=',1PD9.3,
     X    '   ||x+s-u||=',1PD9.3,'   ||At*y+z-w-c||=',1PD9.3)
         CALL MYWRT(0,BUFFER)
         CALL MYWRT(IOLOG,BUFFER)
      ENDIF
C
C
      RETURN
C
C
C *** LAST CARD OF (PCCHCK) ***
      END
