C******************************************************************
C     * PCELIM ... ELIMINATE COLS/ROWS APPROACHING OPTIMAL VALUES *
C     *            PREDICTOR-CORRECTOR PRIMAL-DUAL METHOD         *
C******************************************************************
C
      SUBROUTINE PCELIM(LORD,MAXM,MAXN,MAXNZA,MAXNZL,
     X M,N,NSTRCT,NFIX,MOUT,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X LDSQRT,LCLPTS,LRWNBS,LLINKS,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X INTMP1,IXCHNG,ISCHNG,IMTMP1,IMTMP2,RNTMP1,
     X PERM,INVP,HEADER,LINKFD,LINKBK,
     X PRFSBT,DLFSBT,XIB,XIU,XIC,XFIX,YFIX,
     X X,Y,S,Z,W,DELTAX,DELTAS,DELTAY,DELTAZ,DELTAW,YPROX,
     X VUSED,VBNDED,C,UPBND,P,Q,B,RANGES,
     X RSCALE,CSCALE,STAVAR,STAROW,RWSTAT,RWNAME,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 LORD,MAXM,MAXN,MAXNZA,MAXNZL,M,N,NSTRCT,NFIX,MOUT
      INTEGER*4 LIWORK,LRWORK,IOLOG
C
      DOUBLE PRECISION ACOEFF(MAXNZA)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
C
      INTEGER*4 INTMP1(MAXN),IROW(MAXN)
      INTEGER*2 IXCHNG(MAXN),ISCHNG(MAXN)
      DOUBLE PRECISION RNTMP1(MAXN),RELT(MAXN)
      INTEGER*4 IMTMP1(MAXM+1),IMTMP2(MAXM+1)
C
      INTEGER*2 PERM(MAXM),INVP(MAXM)
      INTEGER*2 HEADER(MAXM+1),LINKFD(MAXM+1),LINKBK(MAXM+1)
C
      LOGICAL VUSED(MAXN),VBNDED(MAXN)
      DOUBLE PRECISION P(MAXM),Q(MAXM)
      DOUBLE PRECISION C(MAXN),B(MAXM),UPBND(MAXN),RANGES(MAXM)
      DOUBLE PRECISION X(MAXN),S(MAXN),Y(MAXM),Z(MAXN),W(MAXN)
      DOUBLE PRECISION DELTAX(MAXN,LORD),DELTAS(MAXN,LORD)
      DOUBLE PRECISION DELTAY(MAXM,LORD),YPROX(MAXM)
      DOUBLE PRECISION DELTAZ(MAXN,LORD),DELTAW(MAXN,LORD)
      DOUBLE PRECISION PRFSBT,DLFSBT,XFIX,YFIX
      DOUBLE PRECISION XIB(MAXM),XIU(MAXN),XIC(MAXN)
      DOUBLE PRECISION RSCALE(MAXM),CSCALE(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      CHARACTER*8 RWNAME(MAXM)
C
C *** DATA STRUCTURES FOR CHOLESKY FACTOR
      DOUBLE PRECISION LDSQRT(MAXM)
      INTEGER*4 LCLPTS(MAXM+1),LLINKS(MAXNZL)
      INTEGER*2 LRWNBS(MAXNZL)
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(LIWORK),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(LRWORK)
C
C
C
C *** COMMON AREAS
C     An indicator if the elimination routine has been used.
      COMMON /ELMNTE/  IELIM
      INTEGER*4        IELIM
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4        KNZ
      INTEGER*4        I,IKX,IPOS,IR,IRUN,J,JCOL,MNEW
      INTEGER*4        K,KBEG,KEND,KOK,KOUT,KSTAT
      DOUBLE PRECISION DP
      CHARACTER*100    BUFFER
C
C
C *** COMMON ARREAS
C     Markers for linking rows.
      COMMON /ICGRAD/ MSPLIT(100000)
      INTEGER*2       MSPLIT
C
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C
C     LORD    The highest degree of computed derivatives of  x,s,y,z,w
C             (order of Mehrotra's method).
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     N       Number of columns of the LP constraint matrix.
C     NSTRCT  Number of structural variables.
C     NFIX    Number of variables FIXED on their optimal values.
C     MOUT    Number of LP constraints that are presumed to be
C             inactive at the optimum.
C
C     ACOEFF  Nonzero elements of matrix A.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C
C     LDSQRT  Square roots of the diagonal elements of Cholesky factor.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LLINKS  Linked lists for Cholesky factor.
C
C     INTMP1  Integer work array of size MAXN.
C     IXCHNG  An indicator of changes of variable X.
C     ISCHNG  An indicator of changes of variable S.
C     IMTMP1  Integer work array of size MAXM.
C     IMTMP2  Integer work array of size MAXM.
C     IROW  and  RELT are the arrays for temporary handling
C             of rows/columns of the constraint matrix. They
C             are primarily intended to handle sparse vectors
C             (in packed form) but may also be used for storing
C             dense ones.
C     RNTMP1  Double precision work array of size MAXN.
C     PRFSBT  Primal feasibility tolerance.
C     DLFSBT  Dual feasibility tolerance.
C     XIB     Current primal residual  A*x-b.
C     XIU     Current primal residual  x+s-upbnd.
C     XIC     Current dual   residual  At*y+z-w-c.
C     XFIX    Threshold value for fixing primal variables. As soon
C             as the primal variable is smaller than XFIX (and the 
C             appropriate dual slack variable is bounded away from
C             zero), the variable is presumed to approach a zero
C             optimal value. It is then fixed and eliminated from
C             the problem.
C     YFIX    Threshold value for eliminating LP constraints. As soon
C             as the dual variable is smaller than  YFIX (and the
C             appropriate primal slack variable is bounded away from
C             zero), the constraint is presumed to be inactive at the
C             optimum. It is then eliminated from the problem.
C
C     PERM    Permutation resulting from the elimination of inactive
C             constraints.
C     INVP    Inverse permutation.
C     HEADER  Header of the doubly linked lists.
C     LINKFD  Forward linked lists.
C     LINKBK  Backward linked lists.
C
C     VUSED   An indicator if a variable is active in the optimization
C             process:
C             .TRUE.   active variable;
C             .FALSE.  FIXED variable.
C     VBNDED  An indicator if a variable has an UPPER bound:
C             .TRUE.   UPPER bounded variable;
C             .FALSE.  UNBOUNDED variable;
C     C       Objective function coefficients.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     B       Right-hand-side of the linear program.
C     RANGES   Array of constraint ranges.
C
C     X       Primal variables of the linear program.
C     S       Primal slack variables of the linear program.
C     Y       Dual variables of the linear program.
C     Z       Dual slack variables of the linear program.
C     W       Dual slack variables of the linear program.
C     DELTAX(*,L)  L-th derivative of x(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     DELTAS(*,L)  L-th derivative of s(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     DELTAY(*,L)  L-th derivative of y(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     DELTAZ(*,L)  L-th derivative of z(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     DELTAW(*,L)  L-th derivative of w(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     YPROX   Dual proximal point.
C
C     RSCALE  Current row scaling factors.
C     CSCALE  Current column scaling factors.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C             7  (or larger) PRESUMED OPTIMAL variable i.e.: x = x0;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWSTAT  Array of row types (sort as before):
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  row type is objective or free.
C     RWNAME  Array of row names (increasing order sort).
C
C     IOLOG   Output unit number where log messages are to be written.
C
C
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to C array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C
C *** LOCAL VARIABLES DESCRIPTION
C
C
C
C
C *** SUBROUTINES CALLED:
C     SATY(FSATY),GETCOL,GETROW,MYWRT,MYSTOP,
C     EMPTYR,REORDA,REORDV,REORDI,SYMFCT,SYMREF
C
C
C *** PURPOSE:
C     This routine eliminates:
C      - variables presumed to approach their optimal values;
C      - constraints presumed to be inactive at the optimum.
C
C
C *** NOTES:
C
C
C
C *** WARNING:
C     This routine alters hidden data structures.
C     It then should be used with extreme care.
C
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). An efficient implementation of 
C        a higher order primal-dual interior point method for large 
C        sparse linear programs, Archives of Control Sciences 2, 
C        No 1-2, pp. 23-40. 
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio,
C                    Systems Research Institute,
C                    Polish Academy of Sciences,
C                    Newelska 6, 01-447 Warsaw, Poland.
C     Date written:  October 12, 1992
C     Last modified: June 9, 1995
C
C
C
C
C
C
C *** BODY OF (PCELIM) ***
C
C
C
C     Set an indicator if the elimination routine has been used.
      IELIM=1
C
C
C
C
C *** DEBUGGING
C     DO 40 J=1,N
C        IF(X(J).LE.XFIX) THEN
C           WRITE(BUFFER,41) J,STAVAR(J),X(J),Z(J)
C  41       FORMAT(1X,'J=',I4,' st=',I4,' X=',1PD9.2,' Z=',1PD9.2)
C           CALL MYWRT(IOLOG,BUFFER)
C        ENDIF
C        IF(STAVAR(J).EQ.1.OR.STAVAR(J).EQ.3) THEN
C           K=STAVAR(J)
C           IF(S(J).LE.XFIX) THEN
C              WRITE(BUFFER,42) J,K,S(J),W(J)
C  42          FORMAT(1X,'J=',I4,' st=',I4,' S=',1PD9.2,' W=',1PD9.2)
C              CALL MYWRT(IOLOG,BUFFER)
C           ENDIF
C        ENDIF
C  40 CONTINUE
C
C
C
C
C *** DEBUGGING
C     DO 80 J=1,N
C        IF(STAVAR(J).GE.0) GO TO 80
C
C     Here for a FREE variable.
C        JCOL=J
C        K=STAVAR(JCOL)
C        IF(JCOL.LT.-K) GO TO 80
C        CALL GETCOL(JCOL,RWORK,IWORK,RMAP,IMAP,
C    X    IROW,RELT,KNZ,MAXN,IOLOG)
C        CALL SDOT(Y,IROW,RELT,KNZ,DP)
C        WRITE(IOLOG,81) JCOL,KNZ,DP,XIC(JCOL),Z(JCOL),C(JCOL)
C  81    FORMAT(1X,'col=',I4,' nz=',I4,' Aty=',1PD9.2,
C    X    ' XIC=',1PD9.2,' Z=',1PD9.2,' C=',1PD9.2)
C        JCOL=-K
C        K=STAVAR(JCOL)
C        CALL GETCOL(JCOL,RWORK,IWORK,RMAP,IMAP,
C    X    IROW,RELT,KNZ,MAXN,IOLOG)
C        CALL SDOT(Y,IROW,RELT,KNZ,DP)
C        WRITE(IOLOG,82) JCOL,KNZ,DP,XIC(JCOL),Z(JCOL),C(JCOL)
C  82    FORMAT(1X,'col=',I4,' nz=',I4,' Aty=',1PD9.2,
C    X    ' XIC=',1PD9.2,' Z=',1PD9.2,' C=',1PD9.2)
C  80 CONTINUE
C
C
C
C
C
C
C     Loop over variables.
C     Fix all the variables that approach their bounds.
C     The following conditions have to be satisfied to eliminate
C     a variable:
C     1.  is satisfies bound constraint with sufficient accuracy;
C     2.  appropriate dual variable is bounded away from zero;
C     3.  it is already small  (i.e. X(j) < XFIX);
C     4.  it has continuously decreased in at least 5 subsequent
C          iters.
      NFIX=0
      DO 200 J=1,NSTRCT
         IF(.NOT.VUSED(J)) GO TO 200
C
         IF(STAVAR(J).LT.0) THEN
C
C     Here for a FREE variable.
            KSTAT=STAVAR(J)
            DP=X(J)-X(-KSTAT)
            IF(DABS(DP).LE.1.0D0) GO TO 200
         ENDIF
C
         IF(VBNDED(J)) THEN
C
C     Here for UPPER bounded variable.
C     First, check if  x + s = u.
            DP=DABS(XIU(J))
            IF(DP.GT.0.00001*UPBND(J)) GO TO 200
            IF(X(J).LE.XFIX.AND.X(J).LE.0.000001*UPBND(J)) THEN
               IF(Z(J)-W(J)+XIC(J).LE.0.5) GO TO 200
               IF(IXCHNG(J).GT.-3) GO TO 200
               IF(ISCHNG(J).LE.2) GO TO 200
C              WRITE(BUFFER,111) J,STAVAR(J),X(J),S(J),UPBND(J)
C 111          FORMAT(1X,'111, J=',I6,'  st=',I6,
C    X          '  X=',D12.5,'  S=',D12.5,'  Uj=',D12.5)
C              CALL MYWRT(IOLOG,BUFFER)
               X(J)=0.0
               S(J)=UPBND(J)
               GO TO 140
            ENDIF
            IF(S(J).LE.XFIX.AND.S(J).LE.0.000001*UPBND(J)) THEN
               IF(W(J)-Z(J)-XIC(J).LE.0.5) GO TO 200
               IF(ISCHNG(J).GT.-3) GO TO 200
               IF(IXCHNG(J).LE.2) GO TO 200
C              WRITE(BUFFER,112) J,STAVAR(J),X(J),S(J),UPBND(J)
C 112          FORMAT(1X,'112, J=',I6,'  st=',I6,
C    X          '  X=',D12.5,'  S=',D12.5,'  Uj=',D12.5)
C              CALL MYWRT(IOLOG,BUFFER)
               X(J)=UPBND(J)
               S(J)=0.0
               GO TO 140
            ENDIF
C
C     Here to eliminate useless (large) UPPER bounds.
            IF(UPBND(J).GE.1.0D+4.AND.S(J).GE.0.95*UPBND(J)) THEN
               IF(Z(J)-W(J)+XIC(J).LE.0.5) GO TO 200
               IF(DABS(XIU(J)).GE.1.0D-3) GO TO 200
               IF(W(J).GE.1.0D-4) GO TO 200
               IF(IXCHNG(J).GT.-2) GO TO 200
               IF(ISCHNG(J).LE.2) GO TO 200
C              WRITE(BUFFER,121) J,STAVAR(J),X(J),S(J),UPBND(J)
C 121          FORMAT(1X,'121, J=',I6,'  st=',I6,
C    X          '  X=',D12.5,'  S=',D12.5,'  Uj=',D12.5)
C              CALL MYWRT(IOLOG,BUFFER)
C
C     Do not alter Z.
               S(J)=0.0
               XIC(J)=XIC(J)-W(J)
               W(J)=0.0
               XIU(J)=0.0
               STAVAR(J)=2
               VBNDED(J)=.FALSE.
C
C     Zero all unused components of DELTAs.
               DO 120 I=1,LORD
                  DELTAS(J,I)=0.0
                  DELTAW(J,I)=0.0
  120          CONTINUE
               GO TO 200
            ENDIF
C
C     Here to eliminate useless (inactive) UPPER bounds.
            IF(UPBND(J).GE.1.0D+2.AND.DABS(XIU(J)).LE.YFIX) THEN
               IF(Z(J)+W(J).GE.YFIX) GO TO 200
               IF(DABS(XIC(J)).GE.YFIX) GO TO 200
               IF(S(J).LE.2.0D-1*UPBND(J)) GO TO 200
C              WRITE(BUFFER,131) J,STAVAR(J),X(J),S(J),UPBND(J)
C 131          FORMAT(1X,'131, J=',I6,'  st=',I6,
C    X          '  X=',D12.5,'  S=',D12.5,'  Uj=',D12.5)
C              CALL MYWRT(IOLOG,BUFFER)
C
C     Do not alter Z.
               S(J)=0.0
               XIC(J)=XIC(J)-W(J)
               W(J)=0.0
               XIU(J)=0.0
               STAVAR(J)=2
               VBNDED(J)=.FALSE.
C
C     Zero all unused components of DELTAs.
               DO 130 I=1,LORD
                  DELTAS(J,I)=0.0
                  DELTAW(J,I)=0.0
  130          CONTINUE
               GO TO 200
            ENDIF
         ELSE
C
C     Here for UNBOUNDED (or LOWER bounded) variable.
            IF(X(J).LE.XFIX) THEN
               IF(Z(J)+XIC(J).LE.0.2) GO TO 200
               IF(IXCHNG(J).GT.-3) GO TO 200
C              WRITE(BUFFER,141) J,STAVAR(J),X(J),S(J),UPBND(J)
C 141          FORMAT(1X,'141, J=',I6,'  st=',I6,
C    X          '  X=',D12.5,'  S=',D12.5,'  Uj=',D12.5)
C              CALL MYWRT(IOLOG,BUFFER)
               X(J)=0.0
               KSTAT=STAVAR(J)
               IF(KSTAT.LT.0) THEN
C
C     Here for a FREE variable. Change status of the split brother.
                  STAVAR(J)=0
                  STAVAR(-KSTAT)=0
               ENDIF
               GO TO 140
            ENDIF
         ENDIF
         GO TO 200
C
C     Eliminate variable J from the LP problem.
  140    NFIX=NFIX+1
         VUSED(J)=.FALSE.
         DP=X(J)
         XIU(J)=0.0D0
         XIC(J)=0.0D0
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
C        WRITE(BUFFER,161) J,X(J)
C 161    FORMAT(1X,'PCELIM: Fixing variable J=',I7,' X=',1PD16.8)
C        CALL MYWRT(IOLOG,BUFFER)
C        WRITE(BUFFER,162) J,STAVAR(J),IXCHNG(J),ISCHNG(J),UPBND(J)
C 162    FORMAT(1X,'J=',I6,'   st=',I6,
C    X    '   ixchng=',I3,'   ischng=',I3,'   Uj=',D12.4)
C        CALL MYWRT(IOLOG,BUFFER)
         STAVAR(J)=STAVAR(J)+7
         DO 160 K=KBEG,KEND
            I=RWNMBS(K)
            B(I)=B(I)-DP*ACOEFF(K)
  160    CONTINUE
C
C     Zero all unused components of DELTAs.
         DO 180 I=1,LORD
            DELTAX(J,I)=0.0
            DELTAZ(J,I)=0.0
            DELTAS(J,I)=0.0
            DELTAW(J,I)=0.0
  180    CONTINUE
C
C
C
C
C     End of the loop over variables.
  200 CONTINUE
C
C
C
C
C
C     Check if there were any new  FIXED variables. If so, then
C     they have been removed from the  LP problem formulation and,
C     consequently, should be removed from the row linked lists.
C     Compute primal residual XIB.
      IF(NFIX.GT.0) THEN
C
C
C     Here if there were FIXED variables.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD array.
         DO 340 I=1,M
            RWHEAD(I)=0
            XIB(I)=B(I)
  340    CONTINUE
C
C     Set the row linked lists.
         DO 380 J=1,N
C
C     Omit all  FIXED variables.
            IF(VUSED(J)) THEN
               KBEG=CLPNTS(J)
               KEND=KBEG+LENCOL(J)-1
               DO 360 K=KBEG,KEND
                  I=RWNMBS(K)
                  XIB(I)=XIB(I)-X(J)*ACOEFF(K)
                  RWLINK(K)=RWHEAD(I)
                  CLNMBS(K)=J
                  RWHEAD(I)=K
  360          CONTINUE
            ENDIF
  380    CONTINUE
C
C     Check if the eliminated rows were not violated.
         DO 400 I=1,M
            K=RWHEAD(I)
            IF(K.LE.0) THEN 
C              WRITE(BUFFER,401) I,RWSTAT(I),RWHEAD(I),B(I)
C 401          FORMAT(1X,'PCELIM: i=',I5,' st=',I2,' hd=',I6,' B=',D10.4)
C              CALL MYWRT(IOLOG,BUFFER)
C
               IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY constraint.
                  IF(DABS(B(I)).GT.PRFSBT) GO TO 9020
                  GO TO 400
               ENDIF
C
               IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
                  IF(B(I).GT.PRFSBT) GO TO 9020
                  GO TO 400
               ENDIF
C
               IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
                  IF(B(I).LT.-PRFSBT) GO TO 9020
                  GO TO 400
               ENDIF
            ENDIF
  400    CONTINUE
C
      ENDIF
C
C
C
C
C     XIB array now contains primal residuals. If the residual
C     (of some inequality-type  LP constraint) is nonzero,
C     then try to absorb it by the appropriate slack variable.
C     If it is not possible, then leave nonzero residual in XIB.
C     As the above analysis involves the loop over all slack
C     variables associated with the inequality constraints
C     it is also used to eliminate inactive constraints.
      DO 410 J=1,N
         INTMP1(J)=0
  410 CONTINUE
      MOUT=0
      DO 460 J=NSTRCT+1,N
         IF(.NOT.VUSED(J)) GO TO 460
C
C     Get column  J from the data structures.
         KBEG=CLPNTS(J)
         IR=RWNMBS(KBEG)
         IF(IR.GT.M) GO TO 460
         DP=(XIB(IR)+X(J)*ACOEFF(KBEG))/ACOEFF(KBEG)
C
         IF(RWLINK(KBEG).EQ.0) THEN
C
C     Single-element row with only a slack entry is found.
C     Remove the row from the LP problem.
C           WRITE(BUFFER,421) J,IR,DP
C 421       FORMAT(1X,'PCELIM: col=',I6,' IR=',I6,' slack=',D10.4,
C    X       ' is eliminated.')
C           CALL MYWRT(IOLOG,BUFFER)
C
C     Check if the slack variable is nonnegative.
            IF(DP.LE.-PRFSBT) GO TO 9030
C
C     Here to eliminate single-element LP constraint.
            X(J)=DP
            STAVAR(J)=14
            VUSED(J)=.FALSE.
            INTMP1(J)=INTMP1(J)+1
            MOUT=MOUT+1
            RWHEAD(IR)=-RWHEAD(IR)
C           Y(IR)=0.0
            XIC(J)=0.0
C
C     Zero all unused components of DELTAs.
            DO 420 I=1,LORD
               DELTAX(J,I)=0.0
               DELTAZ(J,I)=0.0
               DELTAS(J,I)=0.0
               DELTAW(J,I)=0.0
  420       CONTINUE
            GO TO 460
C
         ENDIF
C
C
C     Check if the slack variable can absorb infeasibility. If so,
C     then correct X(J). If not, then do not alter the old X(J).
         IF(DP.GE.0.1) THEN
C
C     Here if the slack variable absorbs infeasibility.
            X(J)=DP
            XIB(IR)=0.0
C
C
C     Check if the analysed constraint is to be eliminated.
            IF(X(J).GE.1.0) THEN
               IF(DABS(Y(IR)).GE.YFIX) GO TO 460
C
C     Here to eliminate inactive LP constraint.
               CALL GETROW(IR,RWORK,IWORK,RMAP,IMAP,
     X          IROW,RELT,K,MAXN,IOLOG)
               DP=0.0
               DO 430 IKX=1,K
                  IF(DABS(RELT(IKX)).GT.DP) DP=DABS(RELT(IKX))
  430          CONTINUE
               IF(DABS(Y(IR))*DP.GE.YFIX*100.0) GO TO 460
C              WRITE(BUFFER,442) J,IR,DP,DABS(Y(IR)),RSCALE(IR)
C 442          FORMAT(1X,'PCELIM: col=',I4,' IR=',I4,' ||r||=',D10.4,
C    X          ' |y|=',D10.4,' rscl=',D10.4)
C              CALL MYWRT(IOLOG,BUFFER)
               DO 440 IKX=1,K
                  JCOL=IROW(IKX)
                  INTMP1(JCOL)=INTMP1(JCOL)+1
C                 WRITE(BUFFER,443) IR,JCOL,RELT(IKX)
C 443             FORMAT(1X,'PCELIM: Rw=',I6,' cl=',I6,' elt=',D12.4)
C                 CALL MYWRT(IOLOG,BUFFER)
  440          CONTINUE
               MOUT=MOUT+1
               RWHEAD(IR)=-RWHEAD(IR)
               Y(IR)=0.0
C              WRITE(BUFFER,444) J,Z(J),XIC(J)
C 444          FORMAT(1X,'PCELIM: col=',I6,'  Z=',D12.4,'  XIC=',D12.4)
C              CALL MYWRT(IOLOG,BUFFER)
               XIC(J)=0.0
C
C     Zero all unused components of DELTAs.
               DO 450 I=1,LORD
                  DELTAX(J,I)=0.0
                  DELTAZ(J,I)=0.0
                  DELTAS(J,I)=0.0
                  DELTAW(J,I)=0.0
  450          CONTINUE
C
            ENDIF
         ENDIF
  460 CONTINUE
C
C
C
C
C     Determine the permutation that puts all empty and inactive
C     rows at the end of the list.
C
      IRUN=3
      CALL EMPTYR(MAXM,M,MNEW,IRUN,
     X RWHEAD,STAROW,PERM,INVP,IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the analysis of EMPTYR.
      IF(MNEW.LT.M) THEN
C
         CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X    CLPNTS,RWNMBS,
     X    RWHEAD,RWLINK,CLNMBS,LENCOL,
     X    PERM,INVP,IMTMP1,IMTMP2,RELT,
     X    RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C
C     Reorder  RSCALE, YPROX, P, Q, XIB, Y, LDSQRT and MSPLIT arrays.
C
         CALL REORDV(MAXM,M,
     X    PERM,INVP,RSCALE,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,YPROX,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,P,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,Q,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,XIB,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,Y,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,LDSQRT,RELT,IOLOG)
         CALL REORDI(MAXM,M,
     X    PERM,INVP,MSPLIT,IMTMP1(1),IOLOG)
C
C
C
C     Reorder elements within each column of the  LP constraint
C     matrix in such a way that those of the active part of  A
C     are at the beginning of the lists. The column lengths will
C     later be decreased to forget inactive part of matrix  A.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD array.
         DO 620 I=1,M
            RWHEAD(I)=0
  620    CONTINUE
         DO 700 J=1,N
            IF(.NOT.VUSED(J)) GO TO 700
            KBEG=CLPNTS(J)-1
            KOK=0
            KOUT=0
            DO 640 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=RWNMBS(K)
               IF(I.LE.MNEW) THEN
                  KOK=KOK+1
                  IROW(KOK)=RWNMBS(K)
                  RELT(KOK)=ACOEFF(K)
               ELSE
                  IPOS=LENCOL(J)-KOUT
                  KOUT=KOUT+1
                  IROW(IPOS)=RWNMBS(K)
                  RELT(IPOS)=ACOEFF(K)
               ENDIF
  640       CONTINUE
C
C     Set the row linked lists.
            DO 660 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=IROW(IKX)
               RWNMBS(K)=I
               ACOEFF(K)=RELT(IKX)
               RWLINK(K)=RWHEAD(I)
               RWHEAD(I)=K
  660       CONTINUE
  700    CONTINUE
C
C
C
C
C     Update the column lengths of the active part of the  LP
C     constraint matrix. Mark all eliminated slacks.
         DO 720 J=1,N
            LENCOL(J)=LENCOL(J)-INTMP1(J)
            IF(LENCOL(J).EQ.0.AND.J.GT.NSTRCT) THEN
               STAVAR(J)=14
               VUSED(J)=.FALSE.
            ENDIF
  720    CONTINUE
C
C
C     Prepare data structures for the new Cholesky matrix.
         IF(M-MNEW.GT.M/50.OR.NFIX.GT.NSTRCT/20) THEN
C
C     Repeat symbolic factorization.
            CALL SYMFCT(LLINKS,IROW,
     X       LCLPTS,LRWNBS,MAXNZL,MAXM,MAXN,MAXNZA,MNEW,
     X       HEADER,LINKFD,LINKBK,IMTMP1,IMTMP2,STAVAR,
     X       CLPNTS,RWNMBS,
     X       RWHEAD,RWLINK,CLNMBS,LENCOL,IOLOG)
C
         ELSE
C
C     Compress data structures for Cholesky matrix.
            CALL SYMREF(MAXNZL,MAXM,M,MNEW,
     X       LCLPTS,LRWNBS,PERM,INVP,IOLOG)
C
         ENDIF
C
C
C     Set the new number of rows of the constraint matrix.
         M=MNEW
C
C
C
C
C     Recompute the residual of the dual constraint.
C
C     XIC := c - At*y - z + w
C
C        CALL SATY(RWORK,IWORK,RMAP,IMAP,Y,M,XIC,N,
C    X    IROW,RELT,MAXN,IOLOG)
         CALL FSATY(MAXM,MAXN,MAXNZA,Y,M,XIC,N,
     X    RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),IWORK(IMAP(6)),
     X    VUSED,IOLOG)
         DO 780 J=1,N
            IF(VUSED(J)) THEN
               XIC(J)=C(J)-XIC(J)-Z(J)
               IF(VBNDED(J)) THEN
                  XIC(J)=XIC(J)+W(J)
                  IF(XIC(J).LT.0.0D0) THEN
                     Z(J)=Z(J)-XIC(J)
                     XIC(J)=0.0D0
                  ENDIF
               ENDIF
            ENDIF
  780    CONTINUE
      ENDIF
C
C
C
C     Write the  LP problem statistics.
      KNZ=0
      DO 880 J=1,N
         IF(VUSED(J)) THEN
            IF(STAVAR(J).LT.0) THEN
               K=-STAVAR(J)
               IF(J.GE.K) GO TO 880
            ENDIF
            KNZ=KNZ+LENCOL(J)
         ENDIF
  880 CONTINUE
C
      WRITE(BUFFER,891)
  891 FORMAT(1X,'PCELIM: New  LP problem statistics:')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,892) M
  892 FORMAT(1X,'        Constraints         ',I13)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,893) KNZ
  893 FORMAT(1X,'        Nonzero elts in  A  ',I13)
      CALL MYWRT(IOLOG,BUFFER)
C
      RETURN
C
C
C     Here if an error occurs.
 9020 WRITE(BUFFER,9021) I,RWNAME(I),B(I)
 9021 FORMAT(1X,'PCELIM: Constraint ',I6,' (name=',A8,
     X ') is violated, slack=',D12.6)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9022)
 9022 FORMAT(1X,'PCELIM: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9030 WRITE(BUFFER,9031) IR,RWNAME(IR),DP
 9031 FORMAT(1X,'PCELIM: Constraint ',I6,' (name=',A8,
     X ') is violated, slack=',D12.6)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9032)
 9032 FORMAT(1X,'PCELIM: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
C
C
C *** LAST CARD OF (PCELIM) ***
      END
