C*****************************************************************
C     ***  RCLSNG ... ELIMINATE FREE COLUMN SINGLETONS FROM A  ***
C     ***  If you love somebody, set them FREE                 ***
C*****************************************************************
C
      SUBROUTINE RCLSNG(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,RNTMP1,
     X B,RANGES,C,LOBND,UPBND,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X P,Q,PRLVAR,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,LENROW)
C
C
C
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real array that contains real  LP problem data.
C     IWORK   Integer array that contains integer  LP problem data.
C     RMAP    Map of RWORK array.
C     IMAP    Map of IWORK array.
C
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,MAXM,MAXN,MAXNZA
      INTEGER*4 M,N,NSTRCT,LNHIST,MXHIST
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
      INTEGER*4 IROW(MAXN),IMTMP1(MAXM),INTMP1(MAXN)
      INTEGER*2 INTMP2(MAXN)
      DOUBLE PRECISION RELT(MAXN),RNTMP1(MAXN)
      DOUBLE PRECISION ACOEFF(MAXNZA),B(MAXM),RANGES(MAXM)
      DOUBLE PRECISION C(MAXN),LOBND(MAXN),UPBND(MAXN)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      DOUBLE PRECISION P(MAXM),Q(MAXM),PRLVAR(MAXN)
      CHARACTER*8 RWNAME(MAXM),CLNAME(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      INTEGER*2 INVP(MAXM),PERM(MAXM),LENROW(MAXM)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 ROWLEN,NELIM,NFREE
      INTEGER*4 I,IKX,IPOS,IRUN,J,JCOL,K,KOK,KOUT
      INTEGER*4 KBEG,KEND,MNEW,SNGLHD
      DOUBLE PRECISION BIG,BIGNEW,DP,BNDJLO,BNDJUP,OLDBND
      DOUBLE PRECISION BLOWER,BUPPER,FSBTOL,SMALLA
      CHARACTER*100 BUFFER
      CHARACTER*2   RTYPE
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     LNHIST  Length of the PRE_SOLVE history list.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     RANGES  Array of constraint ranges.
C     C       Objective function coefficients.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     LOBND   Array of lower bounds.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     PRLVAR  Primal variables of the linear program.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWNAME  Array of row names (increasing order sort).
C     CLNAME  Array of column names (unordered).
C
C *** ON OUTPUT:
C
C
C
C
C *** WORK ARRAYS:
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     IMTMP1  Integer work array of size MAXM.
C     INTMP1  Integer work array of size MAXN
C     INTMP2  Half-length integer work array of size MAXN.
C     RNTMP1  Double precision work array of size MAXN.
C     PERM    Half-length integer work array of size MAXM.
C     INVP    Half-length integer work array of size MAXM.
C     LENROW  Half-length integer work array of size MAXM.
C
C
C
C
C *** PURPOSE
C     This routine looks for FREE (and implied FREE) singleton
C     columns. If such a variable is found, then the constraint
C     with an entry in it can be made FREE. It is thus removed
C     from the problem formulation. The information about each
C     such event is stored on a stack-type history list, which
C     makes possible recovering full PRIMAL solution in a 
C     POST_SOLVE processing.
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,DABS,EMPTYR,REORDA,REORDI,REORDV,MYSTOP
C
C
C *** NOTES
C     This routine is given direct access to the matrix A.
C     It alters hidden data structures.
C
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  January 12, 1994
C     Last modified: June 9, 1995
C
C
C
C
C *** BODY OF (RCLSNG) ***
C
C
C
C     Initialize.
      BIG=1.0D+30
      BIGNEW=1.0D+20
      FSBTOL=1.0D-7
      SMALLA=1.0D-8
      NELIM=0
      NFREE=0
C
      IF(MSGLEV.LE.3) GO TO 140
      DO 130 J=1,N
         IF(STAVAR(J).LT.6) GO TO 130
         WRITE(BUFFER,131) J,LENCOL(J),STAVAR(J),
     X    LOBND(J),UPBND(J),PRLVAR(J)
  131    FORMAT(1X,'col=',I6,' ln=',I4,' st=',I6,' LO=',D10.3,
     X    ' UP=',D10.3,' X=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
  130 CONTINUE
  140 CONTINUE
C
C
C
C     Prepare data structures used in a search for singleton columns.
C     INTMP1 array handles (dynamically changing) column lengths.
C     INTMP2 array handles linked list of singleton columns.
      SNGLHD=0
      DO 200 J=1,NSTRCT
         INTMP1(J)=LENCOL(J)
         IF(LENCOL(J).NE.1) GO TO 200
         IF(STAVAR(J).GE.6) GO TO 200
         IF(STAVAR(J).LT.0) THEN
            K=-STAVAR(J)
            IF(J.GE.K) GO TO 200
         ENDIF
         INTMP2(J)=SNGLHD
         SNGLHD=J
  200 CONTINUE
C
C
C
C
C
C
C     Main loop begins here.
C     Loop over all singleton columns.
 1000 CONTINUE
      IF(SNGLHD.EQ.0) GO TO 2100
C
C
C     Pick up a singleton column.
         J=SNGLHD
         SNGLHD=INTMP2(SNGLHD)
         IF(INTMP1(J).NE.1) GO TO 2000
         IF(STAVAR(J).GE.6) GO TO 2000
C
C     Look for still active element in a column.
         KBEG=CLPNTS(J)
         KEND=CLPNTS(J)+LENCOL(J)-1
         DO 1040 K=KBEG,KEND
            I=RWNMBS(K)
            IF(RWHEAD(I).GT.0) GO TO 1060
 1040    CONTINUE
         WRITE(BUFFER,1041) J
 1041    FORMAT(1X,'RCLSNG: Column ',I8,' has no entries.')
         CALL ERRWRT(IOLOG,BUFFER)
         CALL MYSTOP(93)
C
C     Save singleton position.
 1060    KBEG=K
C
C     Treat original FREE variable specially.
         IF(STAVAR(J).LT.0) THEN
            K=-STAVAR(J)
            IF(J.GE.K) GO TO 2000
            GO TO 1400
         ENDIF
C
C
C     Analyse if the singleton column can be made FREE.
C     Compute LOWER and UPPER limits of the LP constraint.
C     Loop over nonzero entries of row I. Omit column J.
         ROWLEN=0
         IPOS=RWHEAD(I)
         BLOWER=0.0D0
         BUPPER=0.0D0
 1100    IF(IPOS.EQ.0) GO TO 1140
            ROWLEN=ROWLEN+1
            JCOL=CLNMBS(IPOS)
            IF(JCOL.EQ.J) GO TO 1120
            K=STAVAR(JCOL)
            IF(K.GE.6) GO TO 1120
            BNDJUP=BIG
            IF(K.EQ.1.OR.K.EQ.3) BNDJUP=UPBND(JCOL)
            IF(ACOEFF(IPOS).LT.0.0D0) THEN
               BLOWER=BLOWER+BNDJUP*ACOEFF(IPOS)
            ELSE
               BUPPER=BUPPER+BNDJUP*ACOEFF(IPOS)
            ENDIF
 1120    IPOS=RWLINK(IPOS)
         GO TO 1100
C
 1140    CONTINUE
         IF(MSGLEV.LE.2) GO TO 1144
         WRITE(BUFFER,1141) I,RWNAME(I),ROWLEN,RWSTAT(I)
 1141    FORMAT(1X,'RCLSNG: Row ',I6,' (name=',A8,
     X    ') len=',I6,' RWSTAT=',I6)
         CALL MYWRT(IOLOG,BUFFER)
C        WRITE(BUFFER,1142) BLOWER,B(I),BUPPER
C1142    FORMAT(1X,'RCLSNG: blower=',D12.5,' Bi=',D12.5,
C    X    ' bupper=',D12.5)
C        CALL MYWRT(IOLOG,BUFFER)
 1144    CONTINUE
C
C     Compute implied LOWER and UPPER bounds of the variable J.
         K=STAVAR(J)
         OLDBND=BIG
         IF(K.EQ.1.OR.K.EQ.3) OLDBND=UPBND(J)
         IF(ACOEFF(KBEG).LT.0.0D0) THEN
            BNDJLO=(B(I)-BLOWER)/ACOEFF(KBEG)
            BNDJUP=(B(I)-BUPPER)/ACOEFF(KBEG)
         ELSE
            BNDJLO=(B(I)-BUPPER)/ACOEFF(KBEG)
            BNDJUP=(B(I)-BLOWER)/ACOEFF(KBEG)
         ENDIF
         IF(MSGLEV.LE.2) GO TO 1164
         WRITE(BUFFER,1161) OLDBND
 1161    FORMAT(38X,' oldUPj=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,1162) J,STAVAR(J),BNDJLO,BNDJUP
 1162    FORMAT(1X,'cl=',I6,' st=',I6,' newLOj=',D10.3,
     X    ' newUPj=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
 1164    CONTINUE
C
C     Check if the implied bounds are at least as tight as 
C     the original ones.
         IF(BNDJLO.LE.-FSBTOL) GO TO 2000
         IF(BNDJUP.GE.BIGNEW) BNDJUP=BIGNEW
         IF(BNDJUP.GT.OLDBND+FSBTOL) GO TO 2000
C
C
C     Singleton FREE column found.
 1400    CONTINUE
         IF(MSGLEV.LE.1) GO TO 1404
         WRITE(BUFFER,1401) I,RWNAME(I)
 1401    FORMAT(1X,'RCLSNG: Row      ',I6,' (name=',A8,
     X    ') is eliminated (it became FREE).')
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,1402) J,CLNAME(J)
 1402    FORMAT(1X,'RCLSNG: Variable ',I6,' (name=',A8,
     X    ') is a sinleton FREE one.')
         CALL MYWRT(IOLOG,BUFFER)
 1404    CONTINUE
C        IF(MSGLEV.LE.2) GO TO 1406
C        WRITE(BUFFER,1405) I,RWNAME(I),RWSTAT(I)
C1405    FORMAT(1X,'RCLSNG: Row ',I6,' (name=',A8,') RWSTAT=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
C1406    CONTINUE
C
C
C     Eliminate row I from matrix A. Update objective function,
C     column lengths and the linked list of singleton columns.
         NELIM=NELIM+1
         NFREE=NFREE+1
         IPOS=RWHEAD(I)
         RWHEAD(I)=-RWHEAD(I)
         DP=C(J)/ACOEFF(KBEG)
 1500    IF(IPOS.EQ.0) GO TO 1540
            JCOL=CLNMBS(IPOS)
            IF(JCOL.EQ.J) GO TO 1520
            IF(JCOL.GT.NSTRCT) GO TO 1520
C           WRITE(BUFFER,1501) JCOL,C(JCOL),-DP*ACOEFF(IPOS)
C1501       FORMAT(1X,'RCLSNG: j=',I6,' obj=',D10.3,' corr.=',D10.3)
C           CALL MYWRT(IOLOG,BUFFER)
            C(JCOL)=C(JCOL)-DP*ACOEFF(IPOS)
            INTMP1(JCOL)=INTMP1(JCOL)-1
            IF(INTMP1(JCOL).EQ.1) THEN
               IF(STAVAR(JCOL).LT.0) THEN
                  IF(JCOL.GE.-STAVAR(JCOL)) GO TO 1520
               ENDIF
               INTMP2(JCOL)=SNGLHD
               SNGLHD=JCOL
            ENDIF
 1520    IPOS=RWLINK(IPOS)
         GO TO 1500
C
C
C     Remove column J from the active part of the LP constraint
C     matrix. If it was original FREE variable, then remove also
C     its brother. Save position of the pivot element.
 1540    INTMP1(J)=KBEG
         K=STAVAR(J)
         STAVAR(J)=5
         IF(K.LT.0) THEN
            K=-K
            STAVAR(K)=4
            PRLVAR(K)=0.0D0
         ENDIF
C
C     Fix the slack variable if there is any.
C     Eliminate slack column from the row linked list.
C     It is required for compatibility with SFSM library as its
C     CRASH routine reconstructs all logical part of matrix A.
         IF(RWSTAT(I).GE.2) THEN
            KOK=-RWHEAD(I)
            JCOL=CLNMBS(KOK)
            PRLVAR(JCOL)=0.0D0
            STAVAR(JCOL)=14
            RWHEAD(I)=-RWLINK(KOK)
C           WRITE(BUFFER,1541) I,RWSTAT(I),JCOL
C1541       FORMAT(1X,'RCLSNG: rw=',I6,' rw_st=',I6,' slack=',I6)
C           CALL MYWRT(IOLOG,BUFFER)
         ENDIF
C
C     Save the new FREE variable in a PRE_SOLVE history list.
         IF(LNHIST.GE.MXHIST) GO TO 9200
         LNHIST=LNHIST+1
         INHIST(LNHIST)=J
         DPHIST(LNHIST)=ACOEFF(KBEG)
C
C
C
C     End of main loop.
 2000 GO TO 1000
 2100 CONTINUE
C
C
C
C
C
C
C
C     Determine the permutation that puts all empty and inactive
C     rows at the end of the list.
C
      IRUN=3
      IF(MSGLEV.LE.1) IRUN=4
      CALL EMPTYR(MAXM,M,MNEW,IRUN,
     X RWHEAD,STAROW,PERM,INVP,IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the analysis of EMPTYR.
      IF(MNEW.LT.M) THEN
C
         CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X    CLPNTS,RWNMBS,
     X    RWHEAD,RWLINK,CLNMBS,LENCOL,
     X    PERM,INVP,IMTMP1,IROW,RELT,
     X    RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C     Reorder bounds on shadow prices P and Q.
         CALL REORDV(MAXM,M,
     X    PERM,INVP,P,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,Q,RELT,IOLOG)
C
C     Reorder elements within each column of the  LP constraint
C     matrix in such a way that those of the active part of  A
C     are at the beginning of the lists. The column lengths will
C     later be decreased to forget inactive part of matrix  A.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD and LENROW arrays.
         DO 5200 I=1,M
            RWHEAD(I)=0
            LENROW(I)=0
 5200    CONTINUE
C
C     Reorder nonzero elements within each column.
         DO 5500 J=1,N
            IF(STAVAR(J).GE.6) GO TO 5500
            KBEG=CLPNTS(J)-1
            KOK=0
            KOUT=0
C
C     Put the pivot element just behind the active part of every
C     FREE (or implied FREE) singleton column. Observe that we put
C     it at the last position so as 5300 loop could do the rest.
            IF(STAVAR(J).EQ.5) THEN 
               IPOS=INTMP1(J)
               I=RWNMBS(IPOS)
               DP=ACOEFF(IPOS)
               KEND=KBEG+LENCOL(J)
               RWNMBS(IPOS)=RWNMBS(KEND)
               ACOEFF(IPOS)=ACOEFF(KEND)
               RWNMBS(KEND)=I
               ACOEFF(KEND)=DP
            ENDIF
C
            DO 5300 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=RWNMBS(K)
               IF(I.LE.MNEW) THEN
                  KOK=KOK+1
                  IROW(KOK)=RWNMBS(K)
                  RELT(KOK)=ACOEFF(K)
               ELSE
                  IPOS=LENCOL(J)-KOUT
                  KOUT=KOUT+1
                  IROW(IPOS)=RWNMBS(K)
                  RELT(IPOS)=ACOEFF(K)
               ENDIF
 5300       CONTINUE
C
C     Set the row linked lists.
C     Count nonzero elements in all rows of  A.
            DO 5400 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=IROW(IKX)
               RWNMBS(K)=I
               ACOEFF(K)=RELT(IKX)
               RWLINK(K)=RWHEAD(I)
               RWHEAD(I)=K
               LENROW(I)=LENROW(I)+1
 5400       CONTINUE
            LENCOL(J)=KOK
C
C     Set status for eliminated variables.
            IF(STAVAR(J).EQ.4) THEN
               IF(J.LE.NSTRCT) THEN
                  STAVAR(J)=15
               ELSE
                  STAVAR(J)=14
               ENDIF
            ENDIF
            IF(STAVAR(J).EQ.5) STAVAR(J)=15
 5500    CONTINUE
C
C     Set the new number of rows of the constraint matrix.
         M=MNEW
C
      ENDIF
C
C
C
C
C
C
C
C     Here if a successful run of the loop has been completed.
      IF(MSGLEV.LE.0) GO TO 5010
      WRITE(BUFFER,5001) NELIM
 5001 FORMAT(1X,'RCLSNG: Constraints eliminated: ',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,5002) NFREE
 5002 FORMAT(1X,'        Singleton FREE variables:',I8)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
 5010 CONTINUE
C
C
C
      RETURN
C
C
 9010 WRITE(BUFFER,9011) RWNAME(I),RTYPE,BLOWER,BUPPER,B(I)
 9011 FORMAT(1X,'RCLSNG: Row=',A8,' type=',A2,
     X ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9012)
 9012 FORMAT(1X,'RCLSNG: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9030 WRITE(BUFFER,9031) I,RWNAME(I),B(I)
 9031 FORMAT(1X,'RCLSNG: Constraint ',I6,' (name=',A8,
     X ') is violated, B=',D12.6)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9032)
 9032 FORMAT(1X,'RCLSNG: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9200 WRITE(BUFFER,9201)
 9201 FORMAT(1X,'RCLSNG: Please increase space for PRE_SOLVE ',
     X 'history list.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(6)
C
C
C
C
C
C *** LAST CARD OF (RCLSNG) ***
      END
