C********************************************************************
C     ******* RDRHS ... READ THE RHS SECTION OF THE MPS FILE *******
C********************************************************************
C
      SUBROUTINE RDRHS(MAXM,M,LINE,
     X HDRWCD,LNKRW,HDCLCD,LNKCL,
     X NAMEB,RHS,RWNAME,SECT,INMPS,IOLOG)
C
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,M,LINE,INMPS,IOLOG
      CHARACTER*8 NAMEB,RWNAME(MAXM)
      INTEGER*2 HDRWCD(M+1),LNKRW(M+1)
      INTEGER*2 HDCLCD(M+1),LNKCL(M+1)
      DOUBLE PRECISION RHS(MAXM)
      CHARACTER SECT
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 INDEX
      DOUBLE PRECISION VAL1,VAL2
      CHARACTER*8 NAME0,NAMRW1,NAMRW2
      CHARACTER*100 BUFFER
C
C
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     MAXM    Maximum number of constraints.
C     M       Current number of constraints.
C     LINE    Current number of the line read from the  MPS file.
C     NAMEB   The name of the right hand side section chosen.
C     RWNAME  Array of row names.
C     HDRWCD  Header to the linked list of rows with the same codes.
C     LNKRW   Linked list of rows with the same codes.
C     HDCLCD  Header to the linked list of columns with the same codes.
C     LNKCL   Linked list of columns with the same codes.
C     IOLOG   Output unit number where log messages are to be written.
C     INMPS   Input unit number where the input MPS file is read from.
C
C     ON OUTPUT:
C     RHS     The right hand side vector.
C     SECT    Indicator of the section that follows  RHS one.
C
C
C
C *** SUBROUTINES CALLED
C     LKINDX,MYSTOP
C
C
C
C *** PURPOSE
C     This routine reads the  RHS section of the  MPS file.
C     (It can also be used to read the  RANGES section).
C
C
C
C *** NOTES
C
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). An efficient implementation of 
C        a higher order primal-dual interior point method for large 
C        sparse linear programs, Archives of Control Sciences 2, 
C        No 1-2, pp. 23-40. 
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: June 9, 1995
C
C
C
C *** BODY OF (RDRHS) ***
C
C
C
C
C     Main loop begins here.
  200 LINE=LINE+1
      INDEX=1
      READ(INMPS,201,ERR=9000) SECT,NAME0,NAMRW1,VAL1,NAMRW2,VAL2
  201 FORMAT(A1,3X,A8,2X,A8,2X,D12.0,3X,A8,2X,D12.0)
C
C     Check if the line belongs to the same section.
      IF(SECT.NE.' ') GO TO 300
C
C     First record met defines default section name.
      IF(NAMEB.EQ.'        ') THEN
         NAMEB=NAME0
      ENDIF
      IF(NAME0.NE.NAMEB) GO TO 9000
C
C
C     Find the position of the nonzero element.
C 250 CALL LKINDX(RWNAME,M,NAMRW1,INDEX)
  250 CALL LKCODE(RWNAME,M,NAMRW1,INDEX,HDRWCD,LNKRW,IOLOG)
      IF(INDEX.EQ.0) GO TO 9010
C
C     Save the  RHS coefficient.
      RHS(INDEX)=VAL1
C     WRITE(BUFFER,251) INDEX,RWNAME(INDEX),VAL1
C 251 FORMAT(1X,'RDRHS: rw=',I6,'  rwname=',A8,'  elt=',D14.6)
C     CALL MYWRT(IOLOG,BUFFER)
C
C     Check if there is another nonzero read in the analysed line.
      IF(NAMRW2.NE.'        ') THEN
         NAMRW1=NAMRW2
         VAL1=VAL2
         NAMRW2='        '
         GO TO 250
      ELSE
         GO TO 200
      ENDIF
C
C
C
  300 CONTINUE
      RETURN
C
C
C
C     Here if an error occurs.
 9000 WRITE(BUFFER,9001) LINE
 9001 FORMAT(1X,'RDRHS ERROR: Unexpected characters found',
     X ' at line',I10,' of the MPS file.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(83)
C
 9010 WRITE(BUFFER,9011) LINE
 9011 FORMAT(1X,'RDRHS ERROR: Unknown row was found',
     X ' at line',I10,' of the MPS file.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(83)
C
C
C
C *** LAST CARD OF (RDRHS) ***
      END
