C**********************************************************
C     ****    REORDA ... REORDERING ROWS OF  A     ****
C**********************************************************
C
      SUBROUTINE REORDA(MAXM,MAXN,MAXNZA,M,N,
     X CLPNTS,RWNMBS,RWHEAD,RWLINK,CLNMBS,LENCOL,
     X PERM,INVP,TEMP1,TEMP2,DPWORK,
     X RWNAME,STAROW,RWSTAT,RANGES,RHS,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,M,N,IOLOG
      INTEGER*4 TEMP1(MAXM),TEMP2(MAXM)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      CHARACTER*8 RWNAME(MAXM)
      DOUBLE PRECISION RANGES(MAXM),RHS(MAXM),DPWORK(MAXM)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 PERM(MAXM),INVP(MAXM)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      INTEGER*2 STAROW(MAXM),RWSTAT(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IDUMMY,IROW,K,SVROW
      CHARACTER*8 SVNAME
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     N       Number of columns of the LP constraint matrix.
C     PERM    Permutation resulting from the minimum degree ordering.
C     INVP    Inverse permutation.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     RWNAME  Array of row names.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWSTAT  Array of row types (sort as before):
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  row type is objective or free.
C     RANGES  Array of constraint ranges.
C     RHS     LP right-hand-side.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     Reordered  LP constraint matrix.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     RWNAME  Array of row names.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWSTAT  Array of row types (sort as before):
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  row type is objective or free.
C     RANGES  Array of constraint ranges.
C
C     WORK ARRAYS:
C     TEMP1   Integer work array.
C     TEMP2   Integer work array.
C     DPWORK  Double precision work array.
C
C
C *** SUBROUTINES CALLED:
C     NONE
C
C
C *** PURPOSE:
C     1. This routine reorders the rows of the  LP constraint
C        matrix according to a given permutation (as e.g. the one
C        resulting from the minimum degree ordering or the one
C        that removes empty rows from A).
C     2. It also reorders RWNAME, STAROW, RWSTAT, RHS and RANGES arrays.
C
C
C *** NOTES:
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter 2.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  October 26, 1992
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (REORDA) ***
C
C
C
C *** DEBUGGING
C     WRITE(IOLOG,51) 
C  51 FORMAT(1X/1X,'REORDA: LP rows before reordering')
C     DO 53 IROW=1,M
C        WRITE(IOLOG,52) IROW,RWNAME(IROW)
C  52    FORMAT(1X,'REORDA: row= ',I5,'   name=',A8)
C  53 CONTINUE
C
C
C
C
C     Save row headers in TEMP1 array.
      DO 100 I=1,M
         TEMP1(I)=RWHEAD(I)
  100 CONTINUE
C
C     Reorder the rows of  A.
      DO 300 I=1,M
         IROW=INVP(I)
C
C     Row  I of  A becomes the row with index  IROW.
         RWHEAD(IROW)=TEMP1(I)
C
C     Modify the whole ancient row  I.
         K=RWHEAD(IROW)
  200    IF(K.EQ.0) GO TO 300
         RWNMBS(K)=IROW
         K=RWLINK(K)
         GO TO 200
  300 CONTINUE
C
C
C
C     Reorder  STAROW and  RWSTAT arrays.
      DO 400 I=1,M
         TEMP1(I)=STAROW(I)
         TEMP2(I)=RWSTAT(I)
  400 CONTINUE
      DO 500 I=1,M
         IROW=INVP(I)
         STAROW(IROW)=TEMP1(I)
         RWSTAT(IROW)=TEMP2(I)
  500 CONTINUE
C
C
C     Reorder  RWNAME array.
C     An in-place ordering algorithm is implemented to avoid
C     the need of using character work array.
      DO 680 IDUMMY=1,M
         I=IDUMMY
         IROW=INVP(I)
         IF(IROW.EQ.I) GO TO 640
         IF(IROW.LE.0) GO TO 680
C
C     Start the loop over the permutation chain.
C     First, save the contents of  IROW cell.
         SVNAME=RWNAME(IROW)
         SVROW=IROW
C
C     Place the contents of cell  I into its new position
C     and mark row  I as the one that has already been used.
  600    RWNAME(IROW)=RWNAME(I)
         INVP(I)=-INVP(I)
C
C     Loop over the permutation chain.
         IROW=I
         I=PERM(I)
         IF(I.NE.SVROW) GO TO 600
C
C     End up the loop over the permutation chain.
         RWNAME(IROW)=SVNAME
  640    INVP(I)=-INVP(I)
  680 CONTINUE
C
C
C     Reorder  RANGES array.
C     Restore the positive sign of  INVP array.
      DO 700 I=1,M
         INVP(I)=-INVP(I)
         DPWORK(I)=RANGES(I)
  700 CONTINUE
      DO 750 I=1,M
         IROW=INVP(I)
         RANGES(IROW)=DPWORK(I)
  750 CONTINUE
C
C
C     Reorder  RHS array.
      DO 800 I=1,M
         DPWORK(I)=RHS(I)
  800 CONTINUE
      DO 850 I=1,M
         IROW=INVP(I)
         RHS(IROW)=DPWORK(I)
  850 CONTINUE
C
C
C
C *** DEBUGGING
C     WRITE(IOLOG,951) 
C 951 FORMAT(1X/1X,'REORDA: LP rows after reordering')
C     DO 953 IROW=1,M
C        WRITE(IOLOG,952) IROW,RWNAME(IROW)
C 952    FORMAT(1X,'REORDA: row= ',I5,'   name=',A8)
C 953 CONTINUE
C
C
      RETURN
C
C *** LAST CARD OF (REORDA) ***
      END
