C*****************************************************************
C     ***  RRWSNG ... ELIMINATE ROW SINGLETONS FROM MATRIX  A  ***
C*****************************************************************
C
      SUBROUTINE RRWSNG(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,IMTMP1,IMTMP2,
     X B,RANGES,LOBND,UPBND,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X P,Q,PRLVAR,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X MARKER,LENROW,HEADER,LINKFD,LINKBK)
C
C
C
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real array that contains real  LP problem data.
C     IWORK   Integer array that contains integer  LP problem data.
C     RMAP    Map of RWORK array.
C     IMAP    Map of IWORK array.
C
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,MAXM,MAXN,MAXNZA,M,N,NSTRCT
      INTEGER*4 LNHIST,MXHIST
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
      INTEGER*4 IROW(MAXN),IMTMP1(MAXM+1),IMTMP2(MAXM+1)
      DOUBLE PRECISION RELT(MAXN)
      DOUBLE PRECISION ACOEFF(MAXNZA),B(MAXM),RANGES(MAXM)
      DOUBLE PRECISION LOBND(MAXN),UPBND(MAXN)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      DOUBLE PRECISION P(MAXM),Q(MAXM),PRLVAR(MAXN)
      CHARACTER*8 RWNAME(MAXM),CLNAME(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      INTEGER*2 LENROW(MAXM),MARKER(MAXM)
      INTEGER*2 HEADER(MAXN),LINKFD(MAXM),LINKBK(MAXM)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IKX,IR,IPOS,J,JCOL,K,KBEG,KEND,KOK,KOUT,KSTAT
      INTEGER*4 NEQELM,NNEELM,MNEW,SNGLHD
      DOUBLE PRECISION BIG,BIGNEW,SMALLA
      DOUBLE PRECISION BL,BU,BNDJLO,BNDJUP,FSBTOL,X0
      CHARACTER*100 BUFFER
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     LNHIST  Length of the PRE_SOLVE history list.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     RANGES  Array of constraint ranges.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     LOBND   Array of lower bounds.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     PRLVAR  Primal variables of the linear program.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWNAME  Array of row names (increasing order sort).
C     CLNAME  Array of column names (unordered).
C
C *** ON OUTPUT:
C
C
C
C
C *** WORK ARRAYS:
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     IMTMP1  Integer work array of size MAXM.
C     IMTMP2  Integer work array of size MAXM
C     MARKER  Integer work array of size MAXM.
C     LENROW  Integer work array of size MAXM.
C     HEADER  Header of the doubly linked lists.
C     LINKFD  Forward linked lists.
C     LINKBK  Backward linked lists.
C
C
C
C *** LOCAL VARIABLES DESCRIPTION:
C
C
C
C *** PURPOSE
C     This routine eliminates row singletons.
C     Variable with an entry in a singleton EQUALITY row is FIXED.
C     If a variable that has an entry in a singleton INEQUALITY
C     row, then a new BOUND is defined for it.
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,GETCOL,DABS,EMPTYR,REORDA,REORDV,MYSTOP
C
C
C *** NOTES
C     This routine is given direct access to the matrix A.
C     It alters hidden data structures.
C
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  February 21, 1993
C     Last modified: June 9, 1995
C
C
C
C
C *** BODY OF (RRWSNG) ***
C
C
C
C     Initialize.
      BIG=1.0D+30
      BIGNEW=1.0D+20
      FSBTOL=1.0D-8
      SMALLA=1.0D-8
C
      IF(MSGLEV.LE.3) GO TO 140
      DO 130 J=1,N
         IF(STAVAR(J).LT.6) GO TO 130
         WRITE(BUFFER,131) J,LENCOL(J),STAVAR(J),
     X    LOBND(J),UPBND(J),PRLVAR(J)
  131    FORMAT(1X,'col=',I6,' ln=',I4,' st=',I6,' LO=',D10.3,
     X    ' UP=',D10.3,' X=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
  130 CONTINUE
  140 CONTINUE
C
C
C
C
C     Compute row lengths and save them in  LENROW array.
C     Mark all rows as active in a search process.
C     MARKER(i)=0 means the row has been eliminated.
      DO 200 I=1,M
         LENROW(I)=0
         MARKER(I)=1
  200 CONTINUE
C
C     Loop over all structural columns of  A.
C     Omit FIXED variables and aggregate split FREE variables.
      DO 300 J=1,NSTRCT
         KSTAT=STAVAR(J)
         IF(KSTAT.GE.6) GO TO 300
         IF(KSTAT.LT.0) THEN
            IF(J.GT.-KSTAT) GO TO 300
         ENDIF
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 250 K=KBEG,KEND
            I=RWNMBS(K)
            LENROW(I)=LENROW(I)+1
  250    CONTINUE
  300 CONTINUE
C
C
C
C
C     Prepare data structures used in a search for singleton rows.
C     IMTMP1 array handles (dynamically changing) row lengths.
C     IMTMP2 array handles linked list of singleton rows.
C     Store equality-type rows at the beginning of the list.
      SNGLHD=0
      DO 400 I=1,M
         IMTMP1(I)=LENROW(I)
         IF(LENROW(I).NE.1) GO TO 400
         IF(RWSTAT(I).EQ.1) GO TO 400
C        WRITE(BUFFER,401) I,RWNAME(I),RWSTAT(I)
C 401    FORMAT(1X,'RRWSNG: rw=',I6,' (name=',A8,
C    X    ') rwstat=',I3,' has len=1')
C        CALL MYWRT(IOLOG,BUFFER)
         IMTMP2(I)=SNGLHD
         SNGLHD=I
  400 CONTINUE
      DO 500 I=1,M
         IF(LENROW(I).NE.1) GO TO 500
         IF(RWSTAT(I).NE.1) GO TO 500
C        WRITE(BUFFER,501) I,RWNAME(I),RWSTAT(I)
C 501    FORMAT(1X,'RRWSNG: rw=',I6,' (name=',A8,
C    X    ') rwstat=',I3,' has len=1')
C        CALL MYWRT(IOLOG,BUFFER)
         IMTMP2(I)=SNGLHD
         SNGLHD=I
  500 CONTINUE
C
C
C
C
C
C
C
C     Main loop begins here.
C     Loop over all singleton rows.
      NEQELM=0
      NNEELM=0
 1000 CONTINUE
      IF(SNGLHD.EQ.0) GO TO 2100
C
C
C     Pick up a singleton row.
         I=SNGLHD
         SNGLHD=IMTMP2(SNGLHD)
         IF(IMTMP1(I).NE.1) GO TO 2000
         IF(MARKER(I).EQ.0) GO TO 2000
C *********************
C        IPOS=RWHEAD(I)
C        IF(RWSTAT(I).GE.2) IPOS=RWLINK(IPOS)
C        KOK=0
C 720    IF(IPOS.LE.0) GO TO 760
C           J=CLNMBS(IPOS)
C           KSTAT=STAVAR(J)
C           IF(KSTAT.GE.6) GO TO 740
C           IF(KSTAT.LT.0) THEN
C              IF(J.GT.-KSTAT) GO TO 740
C           ENDIF
C           KOK=KOK+1
C           WRITE(BUFFER,721) I,J,STAVAR(J)
C 721       FORMAT(1X,'rw=',I6,', pvt cand, cl=',I6,' st=',I3)
C           CALL MYWRT(IOLOG,BUFFER)
C 740    IPOS=RWLINK(IPOS)
C        GO TO 720
C 760    IF(KOK.NE.1) THEN
C           WRITE(BUFFER,761) KOK
C 761       FORMAT(1X,'No of pivot candidates ',I8)
C           CALL MYWRT(0,BUFFER)
C           CALL MYWRT(IOLOG,BUFFER)
C        ENDIF
C *********************
C
C     Look for a pivot element in a row.
         IPOS=RWHEAD(I)
         IF(RWSTAT(I).GE.2) IPOS=RWLINK(IPOS)
C        IF(RWSTAT(I).GE.2) THEN
C           WRITE(BUFFER,1001) I,CLNMBS(IPOS)
C1001       FORMAT(1X,'Inequality row=',I6,', slack=',I6)
C           CALL MYWRT(IOLOG,BUFFER)
C           IPOS=RWLINK(IPOS)
C        ENDIF
 1020    IF(IPOS.LE.0) GO TO 1060
            J=CLNMBS(IPOS)
            KSTAT=STAVAR(J)
            IF(KSTAT.GE.6) GO TO 1040
            IF(KSTAT.LT.0) THEN
               IF(J.GT.-KSTAT) GO TO 1040
            ENDIF
C           WRITE(BUFFER,1021) I,J
C1021       FORMAT(1X,'Singleton rw=',I6,' has pivot in cl=',I6)
C           CALL MYWRT(IOLOG,BUFFER)
            GO TO 1080
 1040    IPOS=RWLINK(IPOS)
         GO TO 1020
 1060    WRITE(BUFFER,1061) I
 1061    FORMAT(1X,'RRWSNG: Row ',I8,' has no entries.')
         CALL ERRWRT(IOLOG,BUFFER)
         CALL MYSTOP(92)
C
C     Here when pivot element has been found.
 1080    KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         BNDJLO=0.0D0
         IF(KSTAT.LT.0) BNDJLO=-BIG
         BNDJUP=BIG
         IF(KSTAT.EQ.1.OR.KSTAT.EQ.3) BNDJUP=UPBND(J)
         IF(MSGLEV.LE.1) GO TO 1085
         WRITE(BUFFER,1081) J,STAVAR(J),BNDJLO,BNDJUP
 1081    FORMAT(1X,'pvt, col=',I6,' st=',I7,' Lj=',D10.3,' Uj=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
         IF(RANGES(I).LE.BIGNEW) THEN
            WRITE(BUFFER,1082) I,RWNAME(I),RWSTAT(I),RANGES(I)
 1082       FORMAT(1X,' rw=',I6,' nm=',A8,' rwst=',I2,' range=',D10.3)
            CALL MYWRT(IOLOG,BUFFER)
         ENDIF
 1085    CONTINUE
C
C
C
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY type constraint.
C     Check if the eliminated variable is feasible.
            X0=B(I)/ACOEFF(IPOS)
            IF(KSTAT.LT.0) GO TO 1800
            IF(X0.LE.-FSBTOL) GO TO 9010
            IF(X0.GE.BNDJUP+FSBTOL) GO TO 9010
C
C     Fix variable J on X0.
            GO TO 1800
C
         ENDIF
C
C
C
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
            IF(ACOEFF(IPOS).LE.0.0D0) THEN
               BU=B(I)/ACOEFF(IPOS)
               BL=-BIG
               IF(RANGES(I).LE.BIGNEW) THEN
                  BL=(B(I)+RANGES(I))/ACOEFF(IPOS)
                  GO TO 1200
               ENDIF
               GO TO 1400
            ELSE
               BL=B(I)/ACOEFF(IPOS)
               BU=BIG
               IF(RANGES(I).LE.BIGNEW) THEN
                  BU=(B(I)+RANGES(I))/ACOEFF(IPOS)
               ENDIF
               GO TO 1200
            ENDIF
C
         ENDIF
C
C
C
         IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
            IF(ACOEFF(IPOS).GE.0.0D0) THEN
               BU=B(I)/ACOEFF(IPOS)
               BL=-BIG
               IF(RANGES(I).LE.BIGNEW) THEN
                  BL=(B(I)-RANGES(I))/ACOEFF(IPOS)
                  GO TO 1200
               ENDIF
               GO TO 1400
            ELSE
               BL=B(I)/ACOEFF(IPOS)
               BU=BIG
               IF(RANGES(I).LE.BIGNEW) THEN
                  BU=(B(I)-RANGES(I))/ACOEFF(IPOS)
               ENDIF
               GO TO 1200
            ENDIF
C
         ENDIF
C
         GO TO 2000
C
C
C
C     Here if the new LOWER bound has been defined.
 1200    CONTINUE
         IF(MSGLEV.LE.1) GO TO 1202
         WRITE(BUFFER,1201) J,CLNAME(J),BL
 1201    FORMAT(1X,'RRWSNG: Variable ',I6,' (name=',A8,
     X    ') has new LOWER bound=',D14.6)
         CALL MYWRT(IOLOG,BUFFER)
 1202    CONTINUE
C        IF(KSTAT.LT.0) THEN
C           WRITE(BUFFER,1203) J
C1203       FORMAT(1X,'RRWSNG: LO bnd on a FREE variable ',I6)
C           CALL MYWRT(IOLOG,BUFFER)
C        ENDIF
C
C     Check if it is tighter then the old LOWER bound.
C     If so, then update the bound.
 1220    IF(BL.LE.BNDJLO+FSBTOL) GO TO 1400
         DO 1240 K=KBEG,KEND
            IR=RWNMBS(K)
            IF(MARKER(IR).EQ.0) GO TO 1240
            B(IR)=B(IR)-BL*ACOEFF(K)
            IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 1240    CONTINUE
         STAVAR(J)=2
         LOBND(J)=LOBND(J)+BL
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
         IF(LNHIST.GE.MXHIST) GO TO 9200
         LNHIST=LNHIST+1
         INHIST(LNHIST)=-J
         DPHIST(LNHIST)=BL
C
C     Catch up bound on a FREE variable (FIX its split brother).
         IF(KSTAT.LT.0) THEN
            JCOL=-KSTAT
            X0=0.0D0
            IF(MSGLEV.LE.1) GO TO 1243
            WRITE(BUFFER,1241) JCOL,CLNAME(JCOL),X0
 1241       FORMAT(1X,'RRWSNG: Variable ',I6,' (name=',A8,
     X       ') is being FIXED on X=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,1242)
 1242       FORMAT(1X,'RRWSNG: FREE variable eliminated !!! ')
            CALL MYWRT(0,BUFFER)
            CALL MYWRT(IOLOG,BUFFER)
 1243       CONTINUE
            PRLVAR(JCOL)=X0
            STAVAR(JCOL)=6
         ENDIF
C
         IF(BNDJUP.LE.BIGNEW) STAVAR(J)=3
         IF(BL.GE.BNDJUP+FSBTOL) GO TO 9210
         UPBND(J)=UPBND(J)-BL
         IF(BL.LE.BNDJUP-FSBTOL) THEN
            BU=BU-BL
            BNDJUP=BNDJUP-BL
            GO TO 1400
         ENDIF
         X0=0.0D0
         GO TO 1800
C
C
C
C     Here if the new UPPER bound has been defined.
 1400    IF(BU.GE.BIGNEW) GO TO 1960
         IF(MSGLEV.LE.1) GO TO 1402
         WRITE(BUFFER,1401) J,CLNAME(J),BU
 1401    FORMAT(1X,'RRWSNG: Variable ',I6,' (name=',A8,
     X    ') has new UPPER bound=',D14.6)
         CALL MYWRT(IOLOG,BUFFER)
 1402    CONTINUE
         IF(KSTAT.LT.0) THEN
C           WRITE(BUFFER,1403) J
C1403       FORMAT(1X,'RRWSNG: UP bnd on a FREE variable ',I6)
C           CALL MYWRT(IOLOG,BUFFER)
C
C     UP bnd on x1 can be handled as LO bnd on x2.
            J=-KSTAT
            KSTAT=STAVAR(J)
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            BL=-BU
            GO TO 1220
         ENDIF
C
C     Check if it is tighter then the old UPPER bound.
C     If so, then update the bound.
         IF(BU.GE.BNDJUP-FSBTOL) GO TO 1960
         IF(BU.LE.BNDJLO-FSBTOL) GO TO 9220
         UPBND(J)=BU
         STAVAR(J)=3
         IF(BU.GE.FSBTOL) GO TO 1960
         X0=0.0D0
C
C
C
C     Fix variable J on X0 and eliminate it.
C     Update row lengths and the linked list of singleton rows.
 1800    CONTINUE
         IF(MSGLEV.LE.1) GO TO 1803
         WRITE(BUFFER,1801) J,CLNAME(J),X0
 1801    FORMAT(1X,'RRWSNG: Variable ',I6,' (name=',A8,
     X    ') is being FIXED on X=',D14.6)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,1802) I,RWNAME(I)
 1802    FORMAT(1X,'RRWSNG: Row      ',I6,' (name=',A8,
     X    ') is eliminated.')
         CALL MYWRT(IOLOG,BUFFER)
 1803    CONTINUE
C
         PRLVAR(J)=X0
         STAVAR(J)=6
         DO 1840 K=KBEG,KEND
            IR=RWNMBS(K)
C           IF(MARKER(IR).EQ.0) GO TO 1840
            B(IR)=B(IR)-X0*ACOEFF(K)
            IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
            IMTMP1(IR)=IMTMP1(IR)-1
            IF(IMTMP1(IR).EQ.1) THEN
               IMTMP2(IR)=SNGLHD
               SNGLHD=IR
            ENDIF
            IF(IR.EQ.I) GO TO 1840
C
C     Check if an empty row has not been created.
            IF(IMTMP1(IR).EQ.0) THEN
               KOK=RWHEAD(IR)
C              WRITE(BUFFER,1805) I,IR,KOK
C1805          FORMAT(1X,'1805 row=',I6,', empty rw=',I6,' KOK=',I6)
C              CALL MYWRT(IOLOG,BUFFER)
               IF(KOK.LE.0) GO TO 1840
               IF(RWSTAT(IR).EQ.1) THEN
                  NEQELM=NEQELM+1
               ELSE
                  NNEELM=NNEELM+1
                  JCOL=CLNMBS(KOK)
                  PRLVAR(JCOL)=0.0D0
                  STAVAR(JCOL)=14
               ENDIF
               IF(MSGLEV.LE.1) GO TO 1842
               WRITE(BUFFER,1841) IR,RWNAME(IR)
 1841          FORMAT(1X,'RRWSNG: Row      ',I6,' (name=',A8,
     X          ') is eliminated.')
               CALL MYWRT(IOLOG,BUFFER)
 1842          CONTINUE
               RWHEAD(IR)=-RWHEAD(IR)
               MARKER(IR)=0
            ENDIF
 1840    CONTINUE
C
C
C     Check if the eliminated variable is a FREE one.
C     If so, then FIX its split brother on a LOWER bound.
         IF(KSTAT.LT.0) THEN
            JCOL=-KSTAT
            X0=0.0D0
            IF(MSGLEV.LE.1) GO TO 1903
            WRITE(BUFFER,1901) JCOL,CLNAME(JCOL),X0
 1901       FORMAT(1X,'RRWSNG: Variable ',I6,' (name=',A8,
     X       ') is being FIXED on X=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,1902)
 1902       FORMAT(1X,'RRWSNG: FREE variable eliminated !!! ')
            CALL MYWRT(0,BUFFER)
            CALL MYWRT(IOLOG,BUFFER)
 1903       CONTINUE
            PRLVAR(JCOL)=X0
            STAVAR(JCOL)=6
         ENDIF
C
C
C     Eliminate row I.
 1960    K=RWHEAD(I)
         IF(RWSTAT(I).EQ.1) THEN
            NEQELM=NEQELM+1
         ELSE
            NNEELM=NNEELM+1
            JCOL=CLNMBS(K)
            PRLVAR(JCOL)=0.0D0
            STAVAR(JCOL)=14
         ENDIF
         RWHEAD(I)=-RWHEAD(I)
         MARKER(I)=0
C
C
C
C
C
C
C
C     End of main loop.
 2000 GO TO 1000
 2100 CONTINUE
C
C
C
C
C
C
C     Here if a successful run of the loop has been completed.
      IF(MSGLEV.LE.0) GO TO 5010
      WRITE(BUFFER,5001) NEQELM
 5001 FORMAT(1X,'RRWSNG: Equalities eliminated:  ',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,5002) NNEELM
 5002 FORMAT(1X,'        Inequalities eliminated:',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
 5010 CONTINUE
C
C
C
C
C
C
C     Determine the permutation that puts all empty and inactive
C     rows at the end of the list.
C
      IR=3
      IF(MSGLEV.LE.1) IR=4
      CALL EMPTYR(MAXM,M,MNEW,IR,
     X RWHEAD,STAROW,LENROW,MARKER,IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the analysis of EMPTYR.
      IF(MNEW.LT.M) THEN
C
         CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X    CLPNTS,RWNMBS,
     X    RWHEAD,RWLINK,CLNMBS,LENCOL,
     X    LENROW,MARKER,IMTMP1,IROW,RELT,
     X    RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C     Reorder bounds on shadow prices P and Q.
         CALL REORDV(MAXM,M,
     X    LENROW,MARKER,P,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    LENROW,MARKER,Q,RELT,IOLOG)
C
C     Reorder elements within each column of the  LP constraint
C     matrix in such a way that those of the active part of  A
C     are at the beginning of the lists. The column lengths will
C     later be decreased to forget inactive part of matrix  A.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD and LENROW arrays.
         DO 5200 I=1,M
            RWHEAD(I)=0
            LENROW(I)=0
 5200    CONTINUE
C
C     Reorder nonzero elements within each column.
         DO 5500 J=1,N
            IF(STAVAR(J).GE.6) GO TO 5500
            KBEG=CLPNTS(J)-1
            KOK=0
            KOUT=0
C
            DO 5300 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=RWNMBS(K)
               IF(I.LE.MNEW) THEN
                  KOK=KOK+1
                  IROW(KOK)=RWNMBS(K)
                  RELT(KOK)=ACOEFF(K)
               ELSE
                  IPOS=LENCOL(J)-KOUT
                  KOUT=KOUT+1
                  IROW(IPOS)=RWNMBS(K)
                  RELT(IPOS)=ACOEFF(K)
               ENDIF
 5300       CONTINUE
C
C     Set the row linked lists.
C     Count nonzero elements in all rows of  A.
            DO 5400 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=IROW(IKX)
               RWNMBS(K)=I
               ACOEFF(K)=RELT(IKX)
               RWLINK(K)=RWHEAD(I)
               RWHEAD(I)=K
               LENROW(I)=LENROW(I)+1
 5400       CONTINUE
            LENCOL(J)=KOK
 5500    CONTINUE
C
C     Set the new number of rows of the constraint matrix.
         M=MNEW
C
      ENDIF
C
C
C
C
C     Check if there are inequality type rows to be eliminated.
C     Check if the eliminated rows were not violated.
      DO 5800 I=1,M
         IF(DABS(B(I)).LE.FSBTOL) B(I)=0.0D0
         K=RWHEAD(I)
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY constraint.
            IF(K.NE.0) GO TO 5800
            IF(DABS(B(I)).GT.FSBTOL) GO TO 9020
            GO TO 5800
         ENDIF
         KOK=RWLINK(K)
         IF(KOK.GT.0) GO TO 5800
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
            IF(B(I).GT.FSBTOL) GO TO 9020
            RWHEAD(I)=-RWHEAD(I)
            J=CLNMBS(K)
            STAVAR(J)=14
            GO TO 5800
         ENDIF
         IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
            IF(B(I).LT.-FSBTOL) GO TO 9020
            RWHEAD(I)=-RWHEAD(I)
            J=CLNMBS(K)
            STAVAR(J)=14
            GO TO 5800
         ENDIF
 5800 CONTINUE
C
C
C
C
C
C
      RETURN
C
C
C
C     Here if an error occurs.
 9010 WRITE(BUFFER,9011) J,CLNAME(J),X0
 9011 FORMAT(1X,'RRWSNG: Var. ',I6,' (name=',A8,
     X ') is beyond its bounds, X=',D12.6)
      CALL ERRWRT(IOLOG,BUFFER)
C     WRITE(BUFFER,9013) I,B(I),ACOEFF(IPOS)
C9013 FORMAT(1X,'RRWSNG: rw=',I6,' Bi=', D10.3,' elt=',D10.3)
C     CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9012)
 9012 FORMAT(1X,'RRWSNG: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9020 WRITE(BUFFER,9021) I,RWNAME(I),B(I)
 9021 FORMAT(1X,'RRWSNG: Constraint ',I6,' (name=',A8,
     X ') is violated, B=',D12.6)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9022)
 9022 FORMAT(1X,'RRWSNG: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
C     WRITE(BUFFER,9023) I,RWSTAT(I),RWHEAD(I),LENROW(I)
C9023 FORMAT(1X,'RRWSNG: Constraint ',I6,' rwstat=',I6,
C    X ' rwhead=',I8,' length=',I6)
C     CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9200 WRITE(BUFFER,9201)
 9201 FORMAT(1X,'RRWSNG: Please increase space for PRE_SOLVE ',
     X 'history list.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(6)
C
 9210 WRITE(BUFFER,9211) BL,BNDJUP,CLNAME(J)
 9211 FORMAT(1X,'RRWSNG: LO bound ',D12.6,' exceeds ',
     X 'UP one ',D12.6,' (var=',A8,').')
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9212)
 9212 FORMAT(1X,'RRWSNG: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9220 WRITE(BUFFER,9221) BNDJLO,BU,CLNAME(J)
 9221 FORMAT(1X,'RRWSNG: LO bound ',D12.6,' exceeds ',
     X 'UP one ',D12.6,' (var=',A8,').')
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9222)
 9222 FORMAT(1X,'RRWSNG: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
C
C
C *** LAST CARD OF (RRWSNG) ***
      END
