C******************************************************
C     ****    SCLCOL ... SCALE COLUMNS OF  A     ****
C******************************************************
C
      SUBROUTINE SCLCOL(MAXN,MAXNZA,N,
     X CLPNTS,LENCOL,ACOEFF,
     X CSCALE,OSCALE,C,UPBND,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXN,MAXNZA,N,IOLOG
      INTEGER*4 CLPNTS(MAXN+1)
      INTEGER*2 LENCOL(MAXN)
      DOUBLE PRECISION ACOEFF(MAXNZA)
      DOUBLE PRECISION C(MAXN),UPBND(MAXN),CSCALE(MAXN),OSCALE
C
C
C *** LOCAL VARIABLES
      INTEGER*4 JCOL,K,KBEG,KEND
      DOUBLE PRECISION DP
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     N       Number of columns of the LP constraint matrix.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     ACOEFF  Array of non zero elements for each column.
C     CSCALE  Column scaling factors.
C     OSCALE  Objective row scaling factor.
C     C       Objective function coefficients.
C     UPBND   Array of upper bounds.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     Scaled  LP constraint matrix.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     ACOEFF  Array of non zero elements for each column.
C     C       Objective function coefficients.
C     UPBND   Array of upper bounds.
C
C
C
C *** SUBROUTINES CALLED:
C     NONE
C
C
C *** PURPOSE:
C     1. This routine scales columns of A.
C        Each column of  A is divided by a given scaling factor. 
C     2. C and UPBND arrays are modified accordingly.
C     3. Additionally, it scales the objective row.
C
C
C *** NOTES:
C
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). An efficient implementation of 
C        a higher order primal-dual interior point method for large 
C        sparse linear programs, Archives of Control Sciences 2, 
C        No 1-2, pp. 23-40. 
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 29, 1992
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (SCLCOL) ***
C
C
C
C
C *** DEBUGGING
C     WRITE(IOLOG,51) 
C  51 FORMAT(1X/1X,'SCLCOL: Scaling factors:')
C     DO 53 JCOL=1,N
C        WRITE(IOLOG,52) JCOL,CSCALE(JCOL)
C  52    FORMAT(1X,'SCLCOL: col= ',I5,'   CSCALE=',D10.3)
C  53 CONTINUE
C
C
C
C     Main loop begins here.
C     Divide column  j of A by CSCALE(j).
      DO 500 JCOL=1,N
         DP=CSCALE(JCOL)
         KBEG=CLPNTS(JCOL)
         KEND=KBEG+LENCOL(JCOL)-1
C
C     Modify the whole column  JCOL.
         DO 200 K=KBEG,KEND
            ACOEFF(K)=ACOEFF(K)/DP
  200    CONTINUE
C
         C(JCOL)=C(JCOL)/DP
         UPBND(JCOL)=UPBND(JCOL)*DP
C
C     End of main loop.
  500 CONTINUE
C
C
C
      DO 600 JCOL=1,N
         C(JCOL)=C(JCOL)/OSCALE
  600 CONTINUE
C
C
      RETURN
C
C *** LAST CARD OF (SCLCOL) ***
      END
