C******************************************************
C     ****    SCLROW ... SCALE ROWS OF  A     ****
C******************************************************
C
      SUBROUTINE SCLROW(MAXM,MAXNZA,M,NSTRCT,
     X RWHEAD,RWLINK,CLNMBS,ACOEFF,
     X RSCALE,RANGES,RHS,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXNZA,M,NSTRCT,IOLOG
      INTEGER*4 RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 CLNMBS(MAXNZA)
      DOUBLE PRECISION ACOEFF(MAXNZA)
      DOUBLE PRECISION RANGES(MAXM),RHS(MAXM),RSCALE(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROW,K
      DOUBLE PRECISION DP
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix.
C     NSTRCT  Number of structural variables (excluding slacks, surplus
C             and artificials).
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     ACOEFF  Nonzero elements of matrix A.
C     RSCALE  Current row scaling factors.
C     RANGES  Array of constraint ranges.
C     RHS     LP right-hand-side.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     Scaled  LP constraint matrix.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     ACOEFF  Nonzero elements of matrix A.
C     RANGES  Array of constraint ranges.
C     RHS     LP right-hand-side.
C
C
C
C *** SUBROUTINES CALLED:
C     NONE
C
C
C *** PURPOSE:
C     1. This routine scales rows of A.
C        Each row of  A is divided by a given scaling factor. 
C     2. RHS and RANGES arrays are modified accordingly.
C
C
C *** NOTES:
C
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). An efficient implementation of 
C        a higher order primal-dual interior point method for large 
C        sparse linear programs, Archives of Control Sciences 2, 
C        No 1-2, pp. 23-40. 
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 29, 1992
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (SCLROW) ***
C
C
C
C
C *** DEBUGGING
C     WRITE(IOLOG,51) 
C  51 FORMAT(1X/1X,'SCLROW: Scaling factors:')
C     DO 53 IROW=1,M
C        WRITE(IOLOG,52) IROW,RSCALE(IROW)
C  52    FORMAT(1X,'SCLROW: row= ',I5,'   RSCALE=',D10.3)
C  53 CONTINUE
C
C
C
C     Main loop begins here.
C     Divide row  i of A by RSCALE(i).
      DO 500 IROW=1,M
         DP=RSCALE(IROW)
C
C     Modify the whole row  IROW.
         K=RWHEAD(IROW)
  100    IF(K.EQ.0) GO TO 400
         IF(CLNMBS(K).LE.NSTRCT) GO TO 200
         K=RWLINK(K)
         GO TO 100
  200    IF(K.EQ.0) GO TO 400
         ACOEFF(K)=ACOEFF(K)/DP
         K=RWLINK(K)
         GO TO 200
C
  400    RHS(IROW)=RHS(IROW)/DP
         RANGES(IROW)=RANGES(IROW)/DP
C
C     End of main loop.
  500 CONTINUE
C
C
C
      RETURN
C
C *** LAST CARD OF (SCLROW) ***
      END
