C**************************************************************
C     **** SDOT ... SPARSE INNER PRODUCT OF TWO VECTORS ****
C**************************************************************
C
      SUBROUTINE SDOT(X,IROW,RELT,KNZ,PROD)
C
C *** PARAMETERS
      INTEGER*4 KNZ,IROW(KNZ)
      DOUBLE PRECISION X(*),RELT(KNZ),PROD
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IKX
C
C *** PURPOSE
C     This routine computes the scalar product of a dense vector  X
C     and a sparse vector  Y (packed in  IROW and  RELT arrays).
C
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     X       The first (dense) vector.
C     IROW    Row numbers of nonzeros in a sparse vector  Y.
C     RELT    Nonzero entries a sparse vector  Y.
C     KNZ     Number of nonzero entries in vector  Y.
C     ON OUTPUT:
C     PROD    Scalar product of vectors  X and  Y.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C
C     and            Dominique Tachat, LAMSADE,
C                    University of Paris Dauphine,
C                    Place du Marechal de Lattre de Tassigny,
C                    75775 Paris Cedex 16, France.
C
C     Last modified: March 21, 1992
C
C
C
C
C *** BODY OF (SDOT) ***
C
      PROD=0.
      DO 100 I=1,KNZ
         IKX=IROW(I)
         PROD=PROD+X(IKX)*RELT(I)
  100 CONTINUE
      RETURN
C
C *** LAST CARD OF (SDOT) ***
      END
