C****************************************************************
C     ** SETMAP ... SET MAPS OF THE HIDDEN DATA STRUCTURES  **
C****************************************************************
C
      SUBROUTINE setmap(MAXM,MAXN,MAXNZA,
     X IMAP,RMAP,LIWORK,LRWORK,INTSZ,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,LIWORK,LRWORK,INTSZ,IOLOG
      INTEGER*4 IMAP(*),RMAP(*)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 INTHLF,IMEMR,RMEMR
      CHARACTER*100 BUFFER
C
C
C *** PARAMETER DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     LIWORK  Size of IWORK array.
C     LRWORK  Size of RWORK array.
C     INTSZ   A size of INTEGER used to store indices.
C             2 means that half-length integers are used,
C             4 means that full-length integers are used.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to COBJ array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C *** SUBROUTINES CALLED:
C     MYSTOP
C
C *** PURPOSE:
C     This routine sets up the maps of the hidden data structures.
C
C
C *** NOTES:
C
C     ARRAY SIZES:
C     ARRAY        NO. OF ENTRIES       TYPE
C     ------       --------------       ----
C     CLPNTS       MAXN + 8             INTEGER *4
C     RWNMBS       MAXNZA               INTEGER *2 (or *4)
C     RWHEAD       MAXM                 INTEGER *4
C     RWLINK       MAXNZA               INTEGER *4
C     CLNMBS       MAXNZA               INTEGER *2 (or *4)
C     LENCOL       MAXN                 INTEGER *2 (or *4)
C
C     COEFFA       MAXNZA               DOUBLE PRECISION
C     COBJ         MAXN                 DOUBLE PRECISION
C     RHS          MAXM                 DOUBLE PRECISION
C
C     FORMULAS FOR DETERMINING HOW MUCH SPACE IS NEEDED:
C
C     LIWORK = MAXM + 2*MAXN + (1+2/INTHLF)*MAXNZA + 9
C     LRWORK = MAXM +   MAXN +              MAXNZA + 1
C
C
C
C *** REFERENCES:
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C
C     and            Dominique Tachat, LAMSADE,
C                    University of Paris Dauphine,
C                    Place du Marechal de Lattre de Tassigny,
C                    75775 Paris Cedex 16, France.
C
C     Date written:  April 11, 1991
C     Last modified: June 9, 1995
C
C
C
C
C
C *** BODY OF (SETMAP) ***
C
C
C
C     Set the parameter controling the length of representation
C     of the half-length INTEGER data. This applies to all arrays
C     handling indicices like row/col numbers or column lengths.
C     Row linked lists are always stored as INTEGER*4 as they are
C     expected to address more than 32000 nonzero elements.
C     INTHLF = 1  means that INTEGER *4 arrays will be used;
C     INTHLF = 2  means that INTEGER *2 arrays will be used;
      INTHLF=4/INTSZ
C     WRITE(0,81) INTSZ, INTHLF
C  81 FORMAT(1X,'SETMAP:    INTSZ=',I6,'  INTHLF=',I6)
C
C
C
C     Check if the problem size does not exceed size limits.
      IF(INTSZ.EQ.2) THEN
         IF(MAXM.GE.32767.OR.MAXN.GE.32767) THEN
            WRITE(BUFFER,91)
   91       FORMAT(1X,'SETMAP ERROR: MAXM, MAXN cannot exceed 32767.')
            CALL ERRWRT(IOLOG,BUFFER)
            CALL MYSTOP(10)
         ENDIF
      ENDIF
C
C
C
C     Set up IMAP, the map of the hidden INTEGER data.
      IMAP(1)=1
      IMAP(2)=IMAP(1)+MAXN+8
      IMAP(3)=IMAP(2)+MAXNZA/INTHLF
      IMAP(4)=IMAP(3)+MAXM
      IMAP(5)=IMAP(4)+MAXNZA
      IMAP(6)=IMAP(5)+MAXNZA/INTHLF
      IMAP(7)=IMAP(6)+MAXN/INTHLF
C
C     Set up RMAP, the map of the hidden REAL data.
      RMAP(1)=1
      RMAP(2)=RMAP(1)+MAXNZA
      RMAP(3)=RMAP(2)+MAXN
      RMAP(4)=RMAP(3)+MAXM
C
C
C
C     Figure out how much space is left.
      IMEMR=LIWORK-IMAP(7)
      RMEMR=LRWORK-RMAP(4)
C
      WRITE(BUFFER,101) LIWORK,LRWORK
  101 FORMAT(1X,'SETMAP: Available memory:',
     X 3X,'INTEGER:',I9,5X,'REAL:',I9)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,102)
  102 FORMAT(1X)
      CALL MYWRT(IOLOG,BUFFER)
C
      IF(IMEMR.GE.0.AND.RMEMR.GE.0) GO TO 200
      WRITE(BUFFER,103)
  103 FORMAT(1X,'SETMAP ERROR: Please increase work arrays:')
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,104) IMAP(7),RMAP(4)
  104 FORMAT(29X,'INTEGER:',I9,5X,'REAL:',I9)
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(9)
C
  200 IF(IMEMR.GT.0) THEN
         WRITE(BUFFER,201) LIWORK,IMAP(7)
  201    FORMAT(9X,'INTEGER memory can be reduced from',I9,' to ',I9)
         CALL MYWRT(IOLOG,BUFFER)
      ENDIF
      IF(RMEMR.GT.0) THEN
         WRITE(BUFFER,202) LRWORK,RMAP(4)
  202    FORMAT(9X,'REAL    memory can be reduced from',I9,' to ',I9)
         CALL MYWRT(IOLOG,BUFFER)
      ENDIF
C
C
      RETURN
C
C *** LAST CARD OF (SETMAP) ***
      END
