C*****************************************************************
C     *** SOLVLT ... SOLVE EQUATION WITH THE TRANSPOSE OF L ***
C*****************************************************************
C
      SUBROUTINE SOLVLT(LCOEFF,LCLPTS,LRWNBS,
     X MAXNZL,MAXM,M,COLUMN,RHS,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,M,IOLOG
      DOUBLE PRECISION LCOEFF(*),COLUMN(MAXM),RHS(MAXM)
      INTEGER*4 LCLPTS(MAXM+1)
C
C *** The following array can be half-length integer.
      INTEGER*2 LRWNBS(MAXNZL)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROW,IX,JCOL,KBEG,KEND,K,LENCOL
      DOUBLE PRECISION DP
C
C
C *** FUNCTIONS
      DOUBLE PRECISION DDOT
C
C
C *** COMMON ARREAS
C     Cholesky factorization parameters.
      COMMON /CHFCT/   CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN,IDNSRW
      DOUBLE PRECISION CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN
      INTEGER*4        IDNSRW
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     RHS     Right-hand-side of the equation.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     COLUMN  Solution of the equation  (Ltransp) * X = RHS.
C
C     WORK ARRAYS:
C     NONE
C
C
C *** SUBROUTINES CALLED:
C     DDOT
C
C
C *** PURPOSE:
C     This routine solves equation with the transpose
C     of the Cholesky factor L.
C
C
C *** NOTES:
C     The lower right corner of the Cholesky factor is stored
C     as a dense matrix (double addressing is thus avoided).
C     IDNSRW (from CHFACT common block) is a number of the first
C     row of a dense window.
C
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter  6.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  April 15, 1991
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (SOLVLT) ***
C
C
C
C     Solve the equation  Ltransp * X = RHS
C     (save  X in  COLUMN array).
C
C     Begin in dense mode.
      IX=1
      DO 300 JCOL=M,IDNSRW,-1
         KBEG=LCLPTS(JCOL)
         LENCOL=LCLPTS(JCOL+1)-KBEG
C        CALL DDOT(LCOEFF(KBEG),COLUMN(JCOL+1),LENCOL,DP)
         DP = ddot(LENCOL,LCOEFF(KBEG),ix,COLUMN(JCOL+1),ix)
         COLUMN(JCOL)=RHS(JCOL)-DP
  300 CONTINUE
C
C     Switch to sparse mode.
      DO 500 JCOL=IDNSRW-1,1,-1
         COLUMN(JCOL)=RHS(JCOL)
         KBEG=LCLPTS(JCOL)
         KEND=LCLPTS(JCOL+1)-1
         IF(KBEG.GT.KEND) GO TO 500
         DO 400 K=KBEG,KEND
            IROW=LRWNBS(K)
            COLUMN(JCOL)=COLUMN(JCOL)-COLUMN(IROW)*LCOEFF(K)
  400    CONTINUE
  500 CONTINUE
C
C
C
      RETURN
C
C
C *** LAST CARD OF (SOLVLT) ***
      END
