C**********************************************************
C     ****    SYMREF ... SYMBOLIC REFACTORIZATION    ****
C**********************************************************
C
      SUBROUTINE SYMREF(MAXNZL,MAXM,M,MNEW,
     X LCLPTS,LRWNBS,PERM,INVP,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,M,MNEW,IOLOG
      INTEGER*4 LCLPTS(MAXM+1)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 LRWNBS(MAXNZL),PERM(MAXM),INVP(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IROW,IR,INEWRW,JCOL,K,KBEG,KEND,KX
      INTEGER*4 LENOFL,LENOK,LENROW
      DOUBLE PRECISION A0,A1,A2,DFLOPS
      CHARACTER*100 BUFFER
C
C
C *** COMMON ARREAS
C     Cholesky factorization parameters.
      COMMON /CHFCT/   CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN,IDNSRW
      DOUBLE PRECISION CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN
      INTEGER*4        IDNSRW
C
C     Additional Cholesky fact. parameters (interface to HYBRID).
      COMMON /CHHYB/   RO,FLOPS,IREG,NZCHL,RTCD
      DOUBLE PRECISION RO,FLOPS
      INTEGER*4        IREG,NZCHL,RTCD
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix before
C             reduction (and the dimension of  A*Atransp).
C     MNEW    Number of rows of the LP constraint matrix after
C             reduction (and the dimension of  A*Atransp).
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     PERM    Permutation resulting from the elimination of inactive
C             constraints.
C     INV     Inverse permutation.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C
C
C     Additionally, through the CHHYB common block SYMREF returns:
C     NZCHL   Overestimate of the number of nonzeros
C             of Cholesky factor (important if RTCD = 0).
C     RTCD    Return code from the symbolic factorization:
C             0 SYMREF failure;
C             1 SYMREF success.
C     FLOPS   Flops required to compute the decomposition.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT,MYSTOP
C
C
C *** PURPOSE:
C     This routine implements a compresion of the static data 
C     structures used to handle the Cholesky factor after removing
C     some rows from the LP constraint matrix (and, consequently, 
C     from the Cholesky matrix of A*THETA*Atransp).
C
C
C *** NOTES:
C     1. This routine assumes that nonzeros of columns of L are
C        in an increasing order and maintains such an order in a 
C        reduced matrix. 
C     2. The lower right corner of the Cholesky factor is stored
C        as a dense matrix (double addressing is thus avoided).
C        IDNSRW (from CHFACT common block) is a number of the
C        first row of a dense window.
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter 10.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: June 9, 1995
C
C
C
C *** BODY OF (SYMREF) ***
C
C     Set the return code (for successful run).
      RTCD=1
C
C
C     Set the parameters controlling the progress of building
C     the sparsity pattern of the Cholesky factor.
C     LENOFL  is a current length of the Cholesky factor.
C     FLOPS   is a cost of the numerical phase of the factorization.
C     LENROW  is the length of a given row of Cholesky factor.
C     INEWRW  is the number of a new row of Cholesky factor.
      LENOFL=0
      FLOPS=0.0D0
      LENROW=0
      INEWRW=0
C
C
C
C
C
C     Main loop begins here (loop over rows of Cholesky factor).
C     Compress all rows of the Cholesky matrix.
      DO 500 IROW=1,M
C
C     Omit inactive row.
         IF(INVP(IROW).GT.MNEW) GO TO 500
         INEWRW=INEWRW+1
         KBEG=LCLPTS(IROW)
         KEND=LCLPTS(IROW+1)-1
         LCLPTS(INEWRW)=LENOFL+1
C
C
C     Analyse old row IROW. Save only those nonzero entries which 
C     refer to still active LP constraints.
         LENROW=0
         DO 400 K=KBEG,KEND
            IR=LRWNBS(K)
            JCOL=INVP(IR)
            IF(JCOL.LE.MNEW) THEN
               LENROW=LENROW+1
               LENOFL=LENOFL+1
               LRWNBS(LENOFL)=JCOL
            ENDIF
  400    CONTINUE
C
C
C     Here if pivot row sparsity pattern is determined.
C     Update  FLOPS.
      FLOPS=FLOPS+DBLE(LENROW)*DBLE(LENROW)
C
C
C *** DEBUGGING
C     WRITE(BUFFER,401) IROW,LENROW
C 401 FORMAT(1X,'SYMREF: row ',I6,' of  L has length=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C
C
C     End of main loop.
  500 CONTINUE
      LCLPTS(INEWRW+1)=LENOFL+1
      LCLPTS(M+1)=LENOFL+1
C
C
C
C
C
C
C     Write problem statistics.
      KX=LENOFL
      A1=LENOFL*200.0
      A2=MNEW*MNEW-MNEW
      IF(MNEW.GT.1) THEN
         A1=A1/A2
      ELSE
         A1=0.0
      ENDIF
      WRITE(BUFFER,501) KX,A1
  501 FORMAT(1X,'SYMREF: Sparse matrix L has ',I13,
     X  ' subdiagonal elts (density=',F5.1,'%).')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,503) FLOPS
  503 FORMAT(1X,'        Decomposition flops',1PD14.6)
      CALL MYWRT(IOLOG,BUFFER)
C
C
C
C
C
C     Check if it is useful to make a switch to the full code
C     near the end of factorization.
C     IDNSRW indicates the first 'dense' row of Cholesky matrix
C     i.e., a row with the density at least  DENSE.
      IDNSRW=MNEW+1
      DO 600 I=MNEW,1,-1
         LENROW=LCLPTS(I+1)-LCLPTS(I)
         A0=(MNEW-I)*DENSE+0.5
         LENOK=A0
         IF(LENROW.GE.LENOK) THEN
            FLOPS=FLOPS-DBLE(LENROW)*DBLE(LENROW)
            IDNSRW=I
         ELSE
            GO TO 700
         ENDIF
  600 CONTINUE
C
C
C     Check if the dense window is not too small.
  700 IF(MNEW-IDNSRW.LE.10) THEN
         IDNSRW=MNEW+1
         GO TO 1000
      ENDIF
      WRITE(BUFFER,701) MNEW-IDNSRW
  701 FORMAT(1X,'SYMREF: Dense window found  ',I13)
      CALL MYWRT(IOLOG,BUFFER)
C
C
C     Expand the lower right triangle of the sparse Cholesky
C     factor to a dense matrix.
      IROW=IDNSRW
      LENOFL=LCLPTS(IDNSRW)-1
      LENROW=MNEW-IDNSRW
      LENROW=LENROW*(LENROW+1)/2
      IF(LENOFL+LENROW.GT.MAXNZL) GO TO 9000
      DO 900 IROW=IDNSRW,MNEW
         LCLPTS(IROW)=LENOFL+1
         DO 800 IR=IROW+1,MNEW
            LENOFL=LENOFL+1
            LRWNBS(LENOFL)=IR
  800    CONTINUE
  900 CONTINUE
      LCLPTS(MNEW+1)=LENOFL+1
C
C
C
C
C     Write final problem statistics.
      DFLOPS=MNEW-IDNSRW
      DFLOPS=2.0*DFLOPS+1.0
      DFLOPS=DBLE(LENROW)*DFLOPS/3.
      FLOPS=FLOPS+DFLOPS
      KX=LENOFL
      A1=LENOFL*200.0
      IF(MNEW.GT.1) THEN
         A1=A1/A2
      ELSE
         A1=0.0
      ENDIF
      WRITE(BUFFER,901) KX,A1
  901 FORMAT(1X,'SYMREF: Final matrix  L has ',I13,
     X  ' subdiagonal elts (density=',F5.1,'%).')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,903) FLOPS,DFLOPS
  903 FORMAT(1X,'        Decomposition flops',1PD14.6,
     X  ' (',1PD12.6,' in dense mode).')
      CALL MYWRT(IOLOG,BUFFER)
C
C
C
C
 1000 CONTINUE
      NZCHL=LENOFL
      RETURN
C
C
C
C
C     Here to write error message.
 9000 WRITE(BUFFER,9001) LENOFL+LENROW
 9001 FORMAT(1X,'SYMREF ERROR: Matrix L overflow ',I10)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9002) MAXNZL
 9002 FORMAT(1X,'    space was provided for only ',I10,' nonzeros.')
      CALL ERRWRT(IOLOG,BUFFER)
C
      IF(IREG.EQ.-1) THEN
C *** Here for HOPDM: STOP the program.
         CALL MYSTOP(7)
      ENDIF
      IF(IREG.GE.0) THEN
C *** Here for HYBRID: Do not STOP the program.
         LENROW=MNEW-IROW+1
         LENROW=LENROW*(LENROW+1)/2
         NZCHL=LENOFL+LENROW
         RTCD=0
         RETURN
      ENDIF
C
C
C
C *** LAST CARD OF (SYMREF) ***
      END
