      SUBROUTINE TIMEPF( JOB, NOUT, IDATIM)
      INTEGER*4          JOB, NOUT
      INTEGER*4          IDATIM(9)
C
C     TIMEPF - Set the Current Date, Time and Elapsed Time
C
C*****Purpose:
C     Subroutine  TIMEPF gets the current date and time by calling
C     the user or Fortran supplied routines  GETDAT and  GETTIM,
C     increases the elapsed time specified in the array  IDATIM
C     by the difference between the current time and the time
C     specified in  IDATIM and prints the current date, time and
C     elapsed time on the output file number  NOUT if  JOB=1.
C     Additionally, if  JOB=0, the elapsed time in  IDATIM is set
C     to zero, whereas if  JOB is neither 0 nor 1, the elapsed time
C     is not changed. Usually  TIMEPF will be called first with
C     JOB=0, and then with  JOB=1 to output the time elapsed since
C     the first call.
C
C*****Remark:
C     The IBM Professional Fortran subroutines from the library
C     file  PROFORT.LIB
C        CALL  GETDAT( IYEAR, IMONTH, IDAY)
C        CALL  GETTIM( IHOUR, IMINUTE, ISECOND, IHUNDREDSECOND)
C     return the current date and time in their  INTEGER*2 arguments.
C     For other compilers you must provide your own versions of
C     GETDAT and  GETTIM. Note that the last argument  HUNDREDSECOND
C     is not used here.
C
C*****Parameters:
C     JOB  is an integer input variable.
C     NOUT  is an integer input variable that specifies a 
C        non-negative output file number.
C     IDATIM  is an integer array of length 9 which on output
C        stores the current date, time and elapsed time as
C        follows:
C        IDATIM(1)  -  year;
C        IDATIM(2)  -  month;
C        IDATIM(3)  -  day;
C        IDATIM(4)  -  hours;
C        IDATIM(5)  -  minutes;
C        IDATIM(6)  -  seconds;
C        IDATIM(7)  -  hours of elapsed time;
C        IDATIM(8)  -  minutes of elapsed time;
C        IDATIM(9)  -  seconds of elapsed time.
C        On input  IDATIM(I) must be set as above for I=4 to 9
C        if  JOB=1, and for I=7 to 9 if  JOB is neither 0 nor 1;
C        for  JOB=0  IDATIM is arbitrary.
C*****Subprograms called:
C     Fortran-supplied  -  FLOAT, GETDAT, GETTIM, IDINT.
      INTEGER*4          IDINT
      REAL               FLOAT
C
C*****History:
C     Written by Krzysztof C. Kiwiel, Systems Research Institute,
C     Polish Academy of Sciences, Newelska 6, 01-447 Warsaw.
C     Date last modified: January 14, 1987.
C
C*****Body of subroutine TIMEPF:
      INTEGER*4          I
      DOUBLE PRECISION   DMINUT, DSECND
      INTEGER*2          JDATIM(7)
C     Get the current date and time. JDATIM is set as  IDATIM.
      CALL GETDAT( JDATIM(1), JDATIM(2), JDATIM(3))
      CALL GETTIM( JDATIM(4), JDATIM(5), JDATIM(6), JDATIM(7))
      IF ( JOB.NE.0 ) GO TO 10
C        Zero the elapsed time.
         IDATIM(7)=0
         IDATIM(8)=0
         IDATIM(9)=0
   10 IF ( JOB.NE.1 ) GO TO 20
C        Calculate the elapsed time in seconds.
         DSECND=60.0*(60.0D+0*FLOAT(IDATIM(7)+JDATIM(4)-IDATIM(4))
     *          +FLOAT(IDATIM(8)+JDATIM(5)-IDATIM(5)))
     *          +FLOAT(IDATIM(9)+JDATIM(6)-IDATIM(6))
C        Account for passing midnight.
         IF ( JDATIM(4).LT.IDATIM(4) ) DSECND=DSECND+86400.0D+0
         DMINUT=IDINT( DSECND/60.0)
         IDATIM(9)=DSECND-60.0*DMINUT
         IDATIM(7)=IDINT( DMINUT/60.0)
         IDATIM(8)=DMINUT-60.0D+0*FLOAT( IDATIM(7))
   20 CONTINUE
C     Save the current date and time.
      DO 30 I=1,6
         IDATIM(I)=JDATIM(I)
   30 CONTINUE
C     Print the date, time and elapsed time.
      IF (NOUT.LT.0) RETURN
      WRITE(NOUT,301) IDATIM
  301 FORMAT(' DATE..',I4,'-',I2.2,'-',I2.2,
     *       '    TIME..',I2,':',I2.2,':',I2.2,
     *       '    ELAPSED TIME..',I2,':',I2.2,':',I2.2)
      RETURN
C*****Last card of subroutine TIMEPF**********************************
      END
