C**************************************************************
C     ****  XGTROW ... GET THE  I-th ROW OF MATRIX  A  ****
C**************************************************************
C
      SUBROUTINE XGTROW(I,ACOEFF,
     X RWHEAD,RWLINK,CLNMBS,
     X JCOL,RELT,ROWLEN,IOLOG)
C
C
C
C *** PARAMETERS
      INTEGER*4 I,ROWLEN,IOLOG
      INTEGER*4 RWHEAD(*),RWLINK(*),JCOL(*)
      INTEGER*2 CLNMBS(*)
      DOUBLE PRECISION ACOEFF(*),RELT(*)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 K
C
C
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     I       Index of the row to be created.
C     ACOEFF  Array of non zero elements for each column.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     JCOL    Column indices of nonzero entries of the row to be created.
C     RELT    Nonzero entries of the row to be created.
C     ROWLEN  Number of nonzero entries of the row to be created.
C
C
C
C *** SUBROUTINES CALLED
C     NONE
C
C
C *** NOTES
C
C
C
C *** REFERENCES:
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  June 10, 1992
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (XGTROW) ***
C
      ROWLEN=0
      K=RWHEAD(I)
  100 IF(K.LE.0) GO TO 200
      ROWLEN=ROWLEN+1
      JCOL(ROWLEN)=CLNMBS(K)
      RELT(ROWLEN)=ACOEFF(K)
      K=RWLINK(K)
      GO TO 100
C
  200 CONTINUE
      RETURN
C
C *** LAST CARD OF (XGTROW) ***
      END
