C*******************************************************************
C     * PCSTEP ... COMPUTE THE LARGEST FIISBLE STEP IN A DIRECTION *
C*******************************************************************
C
      SUBROUTINE PCSTEP(N,JMIN,XMIN,X,DELTA,IOLOG)
C
C *** PARAMETERS
      INTEGER*4 N,JMIN,IOLOG
      DOUBLE PRECISION XMIN,X(N),DELTA(N)
C
C *** LOCAL VARIABLES
      INTEGER*4        J
      DOUBLE PRECISION STEPJ
      CHARACTER*100    BUFFER
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     N      Dimension of arrays X and DELTA.
C     X      Current strictly feasible point.
C     DELTA  Direction to make step along.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     JMIN   Subscript of J such that X(J)/DELTA(J)=XMIN.
C     XMIN   Stepsize.
C
C *** SUBROUTINES CALLED:
C     NONE
C
C *** PURPOSE:
C     This routine computes the largest feasible step in direction
C     DELTA that does not violate the nonnegativity of variables.
C
C *** NOTES:
C     This routine exploits the fact that  DELTA(j)>=0 for all j
C     such that stavar(j)>=6 (FIXED variables).
C     Be sure to set DELTAs associated with FIXED variables to the
C     nonnegative values if calling this routine from HOPDM library.
C
C *** HISTORY:
C     Written by:    Jacek Gondzio,
C                    Systems Research Institute,
C                    Polish Academy of Sciences,
C                    Newelska 6, 01-447 Warsaw, Poland.
C     Last modified: May 15, 1995
C
C
C *** BODY OF (PCSTEP) ***
C
C
      JMIN=0
      XMIN=1.0D+20
      DO 100 J=1,N
         IF(DELTA(J).GE.0.0D0) GO TO 100
         STEPJ=-X(J)/DELTA(J)
         IF(STEPJ.GE.XMIN) GO TO 100
         JMIN=J
         XMIN=STEPJ
  100 CONTINUE
C
      RETURN
C
C
C *** LAST CARD OF (PCSTEP) ***
      END
